#include "date.h"

#include <time.h>
#include <stdio.h>
#include <string.h>

int LOCALTZ;

Date::Date(int dd,int mm,int yy,int hh,int mi,int ss,long tzoffset_) {
  strcpy(timezone,"UTC");
  type=DT_DATETIME;
  relative=0;

  week=0;
  day=dd; month=mm; year=yy;
  hour=hh; minute=mi; second=ss;
  tzoffset=tzoffset_;
  strcpy(buffer,"");
  ok=1;
}

int Date::OK() {
  return ok;
}

Date::Date(time_t t,int tzoffset_) {
  struct tm* ltm;
  
  strcpy(timezone,"Etc/UTC");
  type=DT_DATETIME;
  tzoffset=tzoffset_;
  relative=0;
  week=0;

  // convert to UTC!
  time_t tt=t-tzoffset;
  //ltm=gmtime(&tt);
  ltm=localtime(&tt);

  // printf("daylight 1: %d",ltm->tm_isdst);
  
  year=ltm->tm_year+1900;
  month=ltm->tm_mon+1;
  day=ltm->tm_mday;
  hour=ltm->tm_hour;
  minute=ltm->tm_min;
  second=ltm->tm_sec;
  wday=ltm->tm_wday;
  
  strcpy(buffer,"");

  ok=1;
}

void Date::Relative(int x) { 
  relative=x; 
}
int Date::Relative() { return relative; }

void Date::Type(DATETYPE x) { type=x; }
Date::DATETYPE Date::Type() { return type; }

void Date::Timezone(char* x) { strcpy(timezone,x); }
char* Date::Timezone() { return timezone; }

void Date::TZOffset(long tzoffset_) { tzoffset=tzoffset_;  MkTime(); }
long Date::TZOffset() { return tzoffset; }
time_t Date::Time() { return MkTime(); }

void Date::SetDate(char *ds,int absolute) {
  int result;
  int tzoffset=0;

  // printf("SetDate(%s)\n",ds);

  ok=1;
  day=0; month=0; year=0; hour=0; minute=0; second=0;

  if (strstr(ds,"+")) {
    if (sscanf(strstr(ds,"+"),"%d",&tzoffset)!=1)
      tzoffset=0;
  }

  // printf("Date::SetDate: tzoffset=%d\n",tzoffset);

  if ((!strstr(ds,"null")) && (strcmp(ds,""))) {
    result=sscanf(ds,"%04d%02d%02dT%02d%02d%02dZ",&year,&month,&day,&hour,&minute,&second);
    if (result!=6) {
      result=sscanf(ds,"%2d.%2d.%04d %2d:%02d:%02d",&day,&month,&year,&hour,&minute,&second);
      if (result!=6) {
	result=sscanf(ds,"%04d-%02d-%02d %02d:%02d:%02d",&year,&month,&day,&hour,&minute,&second);
	if (result!=6) {
	  day=0; month=0; year=0; hour=0; minute=0; second=0;
	  result=sscanf(ds,"%2d.%2d.%04d",&day,&month,&year);
	  if (result!=3) {
	    result=sscanf(ds,"%04d-%02d-%02d",&year,&month,&day);
	    if (result!=3) {
	      day=0; month=0; year=0; hour=0; minute=0; second=0;
	      result=sscanf(ds,"%d:%02d:%02d",&hour,&minute,&second);
	      if (result!=3) {
		day=0; month=0; year=0; hour=0; minute=0; second=0;
		result=sscanf(ds,"%d %d:%02d:%02d",&day,&hour,&minute,&second);
		// printf("DSPARSE: %d | %d %d:%02d:%02d\n",result,day,hour,minute,second);
		if (result!=4) {
		  day=0; month=0; year=0; hour=0; minute=0; second=0;
		  result=sscanf(ds,"%d %02d:%02d",&day,&hour,&minute);
		  if (result!=3) {
		    day=0; month=0; year=0; hour=0; minute=0; second=0;
		    result=sscanf(ds,"%02d:%02d",&hour,&minute);
		    if (result!=2) {
		      day=0; month=0; year=0; hour=0; minute=0; second=0;
		      result=sscanf(ds,"%d",&day);
		      if (result!=1) {
			printf("Date::Date: invalid date format '%s'",ds);
			ok=0;
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  strcpy(buffer,"");

  if (!absolute) year=0;

  // normalize
  if (ok) {
    minute=minute+(second/60);
    second=second % 60;
    hour=hour+(minute/60);
    minute=minute % 60;
    day=day + (hour / 24 );
    hour=hour % 24;
    if (absolute) {
      time_t t=MkTime(-tzoffset-LOCALTZ);
      settime(t,1,0);
    }
  }

  if ((year>0) && ((day<1) || (day>31) || (month<1) || (month>12))) {
    printf("Date::Date: illegal date '%s'",ds);
    ok=0;
  }

}

Date::Date(char *ds) {
  strcpy(timezone,"UTC");
  type=DT_DATETIME;
  relative=0;
  week=0;

  SetDate(ds,1);
}

Date::Date(const char *ds) {
  char *dds;

  strcpy(timezone,"UTC");
  type=DT_DATETIME;
  relative=0;
  week=0;

  dds=new char [strlen(ds)+1];
  strcpy(dds,ds);
  SetDate(dds,1);
  delete dds;
}

void Date::settime(time_t t,int absolute,int local) {
  struct tm *ltm;

  ok=1;
  if (absolute) {
    relative=0;
    // ltm=gmtime(&t);
    ltm=localtime(&t);
    year=ltm->tm_year+1900;
    month=ltm->tm_mon+1;
    day=ltm->tm_mday;
    hour=ltm->tm_hour;
    minute=ltm->tm_min;
    second=ltm->tm_sec;
    wday=ltm->tm_wday;
    // printf("  settime: absolute: %s\n",ToChar());
  } else {
    relative=1;
    time_t dt=t;
    year=0;
    month=0;
    wday=0;
    second=dt % 60;
    dt=(t - second)/60;
    minute=dt % 60;
    dt=(dt - minute)/60;
    hour=dt % 24;
    day=(dt - hour)/24;
    // printf("  settime: relative: %s\n",ToChar());
  }
}

void Date::Day(int dd) { day=dd; }
int Date::Day() { return day; }
void Date::WeekDay(int dd) { wday=dd; }
int Date::WeekDay() { return wday; }
void Date::Week(int ww) { week=ww; }
int Date::Week() { return week; }
void Date::Month(int mm) { month=mm; }
int Date::Month() { return month; }
void Date::Year(int yy) { year=yy; }
int Date::Year() { return year; }
void Date::Hour(int hh) { hour=hh; }
int Date::Hour() { return hour; }
void Date::Minute(int mi) { minute=mi; }
int Date::Minute() { return minute; }
void Date::Second(int ss) { second=ss; }
int Date::Second() { return second; }

time_t Date::MkTime(int tzoffset) const {
  struct tm ltm;
  time_t t;
  // printf("MkTime(%d)\n",tzoffset);
  if (ok) {
    ltm.tm_sec=0;
    ltm.tm_min=0;
    ltm.tm_hour=0;
    ltm.tm_mday=0;
    ltm.tm_mon=0;
    ltm.tm_year=0;
    ltm.tm_isdst=0;
    
    if (year>=1900) {
      ltm.tm_mday=day;
      ltm.tm_mon=month-1;
      ltm.tm_year=year-1900;
      if (hour>=0) {
	ltm.tm_sec=second;
	ltm.tm_min=minute;
	ltm.tm_hour=hour+tzoffset;
      }    
      t=mktime(&ltm);
    } else {
      t=((day*24+hour+tzoffset)*60+minute)*60+second;
    }
  } else
    t=0;
  
  return t;
}

static Date yesterday;
static Date today;
static Date tomorrow;
static Date now;
static Date nulldate;
static Date nodate;
static Date dummy;

Date &Date::operator + (Date b) {
  time_t t1,t2,dt;
  dummy.Timezone(Timezone());

  if ((this->Year()>=1900) && (b.Year()>=1900)) {
    printf("Date::operator +: can not add absolute dates");
    dummy.settime(0,0);
    return dummy;
  }

  t1=MkTime();
  t2=b.MkTime();

  dt=0;
  int s=1;

  if ((Relative()>=0) && (b.Relative()>=0)) {
    dt=t1+t2;
    s=1;
  }

  if ((Relative() <0) && (b.Relative()>=0)) {
    if (t2>=t1) {
      dt=t2-t1;
      s=1;
    } else {
      dt=t1-t2;
      s=-1;
    }
  }

  if ((Relative()>=0) && (b.Relative() <0)) {
    if (t1>=t2) {
      dt=t1-t2;
      s=1;
    } else {
      dt=t2-t1;
      s=-1;
    }
  }

  if ((Relative() <0) && (b.Relative() <0)) {
    dt=t1+t2;
    s=-1;
  }

  if ((this->Year()>=1900) || (b.Year()>=1900)) {
    dummy.settime(dt);
    if (s<0) dummy.Relative(s);
  } else {
    dummy.settime(dt,0);
    if (s<0) dummy.Relative(s);
  }

  return dummy;
}

Date &Date::operator + (long b) {
  time_t t1,t2,dt;
  dummy.Timezone(Timezone());

  t1=MkTime();
  t2=b;

  dt=0;
  int s=1;

  if ((Relative()>=0) && (t2>=0)) {
    dt=t1+t2;
    s=1;
  }

  if ((Relative() <0) && (t2>=0)) {
    if (t2>=t1) {
      dt=t2-t1;
      s=1;
    } else {
      dt=t1-t2;
      s=-1;
    }
  }

  if ((Relative()>=0) && (t2<0)) {
    if (t1>=t2) {
      dt=t1-t2;
      s=1;
    } else {
      dt=t2-t1;
      s=-1;
    }
  }

  if ((Relative() <0) && (t2<0)) {
    dt=t1+t2;
    s=-1;
  }

  if (this->Year()>=1900) {
    dummy.settime(dt);
    if (s<0) dummy.Relative(s);
  } else {
    dummy.settime(dt,0);
    if (s<0) dummy.Relative(s);
  }

  return dummy;
}

Date &Date::operator - (Date b) {
  time_t t1,t2,dt;
  dummy.Timezone(Timezone());

  //  printf("Date: [%s] - [%s]\n",ToChar(),b.ToChar());

  t1=MkTime();
  t2=b.MkTime();
  dt=t1-t2;

  //  printf("Date: [%u] - [%u] : %u\n",t1,t2,dt);

  if ((this->Year()>=1900) && (b.Year()>=1900))
    dummy.settime(dt,0);
  else if ((this->Year()<1900) && (b.Year()<1900))
    dummy.settime(dt,0);
  else
    dummy.settime(dt);

  return dummy;
}

Date &Date::operator - (long b) {
  time_t t1,t2,dt;
  dummy.Timezone(Timezone());

  t1=MkTime();
  t2=b;
  dt=t1-t2;
  if (this->Year()>=1900)
    dummy.settime(dt);
  else
    dummy.settime(dt,0);

  return dummy;
}

Date &Date::operator * (float x) {
  time_t t1,dt;
  dummy.Timezone(Timezone());

  if (this->Year()>=1900){
    printf("Date::operator *: can not stretch absolute date");
    dummy.settime(0,0);
    return dummy;
  }

  t1=MkTime();
  dt=(time_t)((float)t1*x);
  dummy.settime(dt,0);

  //  printf("%s * %d = %s\n",ToChar(),x,dummy.ToChar());

  return dummy;
}

Date &Date::operator * (int x) {
  time_t t1,dt;
  dummy.Timezone(Timezone());

  if (this->Year()>=1900){
    printf("Date::operator *: can not stretch absolute date");
    dummy.settime(0,0);
    return dummy;
  }

  t1=MkTime();
  dt=(time_t)((int)t1*x);
  dummy.settime(dt,0);

  //  printf("%s * %d = %s\n",ToChar(),x,dummy.ToChar());

  return dummy;
}

Date &Date::operator * (unsigned int x) {
  time_t t1,dt;
  dummy.Timezone(Timezone());

  if (this->Year()>=1900){
    printf("Date::operator *: can not stretch absolute date");
    dummy.settime(0,0);
    return dummy;
  }

  t1=MkTime();
  dt=(time_t)((unsigned int)t1*x);
  dummy.settime(dt,0);

  //  printf("%s * %d = %s\n",ToChar(),x,dummy.ToChar());

  return dummy;
}

Date &Date::operator = (Date b) {
  Week(b.Week());
  WeekDay(b.WeekDay());
  Day(b.Day());
  Month(b.Month());
  Year(b.Year());
  Hour(b.Hour());
  Minute(b.Minute());
  Second(b.Second());

  Type(b.Type());
  Timezone(b.Timezone());

  Relative(b.Relative());

  return *this;
}

int Date::operator <  (Date b) {
  return (this->MkTime() <  b.MkTime());
}

int Date::operator <= (Date b) {
  return (this->MkTime() <= b.MkTime());
}

int Date::operator == (Date b) {
  return (this->MkTime() == b.MkTime());
}

//int Date::operator != (Date b) {
//  return (this->MkTime() != b.MkTime());
//}

int operator != (const Date & a,const Date & b) {
  return (a.MkTime() != b.MkTime());
}

int Date::operator >= (Date b) {
  return (this->MkTime() >= b.MkTime());
}

int Date::operator >  (Date b) {
  return (this->MkTime() >  b.MkTime());
}

char *Date::ToChar() {
  strcpy(buffer,"");
  if (year>=1900) {
    if (hour>=0)
      sprintf(buffer,"%02d:%02d:%02d",hour,minute,second);
    if (year>=0)
      sprintf(buffer,"%02d.%02d.%04d",day,month,year);
    if ((year>=1900) && (hour>=0)) {
      if ((hour==0) && (minute==0) && (second==0))
	sprintf(buffer,"%02d.%02d.%04d",day,month,year);
      else
	sprintf(buffer,"%02d.%02d.%04d %02d:%02d:%02d",day,month,year,hour,minute,second);
    }
  } else {
    if ((hour>=0) && (day>=0))
      sprintf(buffer,"%d %02d:%02d:%02d",day,hour,minute,second);
    else if (hour>=0)
      sprintf(buffer,"%d %02d:%02d:%02d",day,hour,minute,second);
    else
      sprintf(buffer,"");
  }

  return buffer;
}

char *Date::ISOChar(int tzoffset) {
  Date dtz;
  if ((Hour()>=0) && (tzoffset)) {
    time_t t=MkTime(tzoffset);
    dtz.settime(t,1,tzoffset);
  } else
    dtz=*this;

  strcpy(buffer,"");
  if (dtz.Year()>=1900) {
    if (dtz.Hour()>=0)
      sprintf(buffer,"%02d:%02d:%02d zulu",dtz.Hour(),dtz.Minute(),dtz.Second());
    if (dtz.Year()>=0)
      sprintf(buffer,"%04d-%02d-%02d",dtz.Year(),dtz.Month(),dtz.Day());
    if ((dtz.Year()>=1900) && (dtz.Hour()>=0)) {
      //      if ((dtz.Hour()==0) && (dtz.Minute()==0) && (dtz.Second()==0))
      //	sprintf(buffer,"%04d-%02d-%02d",dtz.Year(),dtz.Month(),dtz.Day());
      //      else
	sprintf(buffer,"%04d-%02d-%02d %02d:%02d:%02d zulu",
		dtz.Year(),dtz.Month(),dtz.Day(),dtz.Hour(),dtz.Minute(),dtz.Second());
    }
  } else {
    if ((dtz.Hour()>=0) && (dtz.Day()>=0))
      sprintf(buffer,"%d %02d:%02d:%02d zulu",dtz.Day(),dtz.Hour(),dtz.Minute(),dtz.Second());
    else if (dtz.Hour()>=0)
      sprintf(buffer,"%d %02d:%02d:%02d",dtz.Day(),dtz.Hour(),dtz.Minute(),dtz.Second());
    else
      sprintf(buffer,"");
  }

  return buffer;
}

char *Date::ToDT() {
  strcpy(buffer,"");
  
  if (!relative) {
    if (year>=1900) {
      if (Type()==DT_DATETIME) {
	if (hour>=0)
	  sprintf(buffer,"%04d%02d%02dT%02d%02d%02dZ",year,month,day,hour,minute,second);
	else
	  sprintf(buffer,"%04d%02d%02dT000000Z",year,month,day);
      } else {
	sprintf(buffer,"%04d%02d%02d",year,month,day);
      }
    } else
      sprintf(buffer,"");
  } else {
    if (relative<0) strcpy(buffer,"-");
    strcat(buffer,"P");
    if (week>0) {
      char ds[32];
      sprintf(ds,"%dW",week);
      strcat(buffer,ds);
    } else {
      if (day>0) {
	char ds[32];
	sprintf(ds,"%dD",day);
	strcat(buffer,ds);
      }
      if ((hour>0) || (minute>0) || (second>0))
	strcat(buffer,"T");
      if (hour>0) {
	char ds[32];
	sprintf(ds,"%dH",hour);
	strcat(buffer,ds);
      }
      if (minute>0) {
	char ds[32];
	sprintf(ds,"%dM",minute);
	strcat(buffer,ds);
      }
      if (second>0) {
	char ds[32];
	sprintf(ds,"%dS",second);
	strcat(buffer,ds);
      }
    }
  }

  return buffer;
}

Date Now() {
  time_t t;

  time(&t);
  now.settime(t,1,1);

  return now;
}

Date Today() {
  today=Now();
  today.Hour(0);
  today.Minute(0);
  today.Second(0);

  return today;
}

Date Yesterday() {
  Date d(1,0,0,0,0,0);
  yesterday=Today()-d;
  return yesterday;
}

Date Tomorrow() {
  Date d(1,0,0,0,0,0);
  tomorrow=Today()+d;
  return tomorrow;
}

Date NullDate() {
  Date d(0,0,0,0,0,0);
  nulldate=d;
  return nulldate;
}

const Date &NoDate() {
  Date d(-1,-1,-1,-1,-1,-1);
  nodate=d;
  return nodate;
}

ostream &operator << (ostream &ostr,Date d) {
  return (ostr << d.Day() << "." << d.Month() << "." << d.Year() << " "
	  << d.Hour() << ":" << d.Minute() << ":" << d.Second());
}
