#ifndef DATE_
#define DATE_

#include <time.h>
#include <ostream.h>
#include <istream.h>

extern int LOCALTZ;

class Date {
 public:
  enum DATETYPE { DT_DATE, DT_DATETIME };
 private:
  char timezone[512];
  int week;
  int wday,day,month,year;
  int hour,minute,second;
  int tzoffset;
  char buffer[128];
  int ok;
  int relative;

  time_t thistime;
  DATETYPE type;

 public:
  time_t MkTime(int tzoffset=0) const;

  int OK();        // check for correct format
  Date(char *ds);
  Date(const char *ds);
  Date(time_t t,int tzoffset_=0);
  Date(int dd=0,int mm=0,int yy=-1,int hh=-1,int mi=0,int ss=0,long tzoffset_=0);

  void Relative(int x); // -1 before, 0=absolute date, 1 after
  int Relative();

  void Type(DATETYPE x);
  DATETYPE Type();

  void Timezone(char* x);
  char* Timezone();

  void TZOffset(long tzoffset_);
  long TZOffset();
  time_t Time();

  void SetDate(char *ds,int absolute=1);
  void settime(time_t t,int absoulute=1,int local=0);

  void Day(int dd);
  int Day();
  void WeekDay(int dd);
  int WeekDay();
  void Week(int ww);
  int Week();
  void Month(int mm);
  int Month();
  void Year(int yy);
  int Year();
  void Hour(int hh);
  int Hour();
  void Minute(int mi);
  int Minute();
  void Second(int ss);
  int Second();

  Date &operator + (Date b);
  Date &operator + (long b);
  Date &operator - (Date b);
  Date &operator - (long b);
  Date &operator * (float x);
  Date &operator * (int x);
  Date &operator * (unsigned int x);
  Date &operator = (Date b);

  int operator <  (Date b);
  int operator <= (Date b);
  int operator == (Date b);
  // int operator != (Date b);
  int operator >= (Date b);
  int operator >  (Date b);

  friend ostream &operator << (ostream &ostr,Date d);

  char *ToChar();
  char *ISOChar(int tzoffset=0);
  char *ToDT();
};

int operator != (const Date & a,const Date & b);

extern Date Yesterday();
extern Date Today();
extern Date Tomorrow();

extern Date Now();
extern Date NullDate();
extern const Date &NoDate();

#endif
