#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "db-interface.h"

#if defined(_WIN32) || defined(_WIN64)
  #include <windows.h>
#endif

#include <postgresql/libpq-fe.h>

Database::Database() {
  result=NULL;
  dbglevel=0;
}

int Database::Connect(char *host,char *dbname,const char *uname,const char *passwd) {
  if (dbglevel>=2) printf("Database::Connect(%s,%s,*)\n",host,uname);

  char conninfo[4096];
  sprintf(conninfo,"host='%s' dbname='%s' user='%s' password='%s'",host,dbname,uname,passwd);
  db=PQconnectdb(conninfo);
  if (PQstatus(db) == CONNECTION_BAD) {
    if (dbglevel>=1) printf("Database::Connect: %s\n",Error());
    PQfinish(db);
    return 0;
  }
  if (dbglevel>=2) printf("Database::Connect: OK\n");
  return 1;
}

int Database::Disconnect() {
  if (dbglevel>=2)printf("%s","Database::Disconnect()");
  PQfinish(db);
  return 0;
}

void Database::DebugLevel(int l) {
  dbglevel=l;
}

int Database::Errno() {
  if (PQresultStatus(result)>2) {
    if (dbglevel>=1) printf("Database::Errno: %d\n",PQresultStatus(result));
  }
  return PQresultStatus(result);
}

char *Database::Error() {
  if (dbglevel>=1) printf("Database::Error: %s\n", PQerrorMessage(db));
  return PQerrorMessage(db);
}

int Database::SQL(char *sql) {
  if (dbglevel>=3) printf("Database::SQL(%s)\n",sql);

  PQclear(result);
  result = NULL;
  rowid  = 0;

  result = PQexec(db,sql);
  int ec=Errno();

  if ((ec==PGRES_EMPTY_QUERY) || (ec==PGRES_COMMAND_OK) || (ec==PGRES_TUPLES_OK))
    return 1;

  if (dbglevel>=1) printf("Database::SQL: %s\n",Error());
  return 0;
}

int Database::NextRow() {
  if (dbglevel>=2) printf("%s","Database::NextRow");

  if (result!=NULL)
    if ((0<PQntuples(result)) && (rowid+1<PQntuples(result))) {
      rowid++;
      return 1;
    }

  if (dbglevel>=1) printf("Database::NextRow: %s\n",Error());
  return 0;
}

int Database::NotNull(int rc) {
  if (result!=NULL) {
    if ((0<PQntuples(result)) && (rowid<PQntuples(result))) {
      if ((0<PQnfields(result)) && (rc<PQnfields(result))) {
	return !(PQgetisnull(result,rowid,rc));
      }
    }
  }
  return 0;
}

unsigned int Database::GetUInt(int rc) {
  char *vb=GetString(rc);
  unsigned int v;
  if (vb!=NULL) {
    sscanf(vb,"%u",&v);
    return v;
  }

  if (dbglevel>=1) printf("Database::GetUInt: %s\n",Error());
  return RAND_MAX;
}

long int Database::GetLong(int rc) {
  char *vb=GetString(rc);
  long v;
  if (vb!=NULL) {
    sscanf(vb,"%ld",&v);
    return v;
  }

  if (dbglevel>=1) printf("Database::GetLong: %s\n",Error());
  return RAND_MAX;
}

int Database::GetInt(int rc) {
  char *vb=GetString(rc);
  int v;
  if (vb!=NULL) {
    sscanf(vb,"%d",&v);
    return v;
  }
  
  if (dbglevel>=1) printf("Database::GetInt: %s\n",Error());
  return RAND_MAX;
}

float Database::GetFloat(int rc) {
  char *vb=GetString(rc);
  float v;
  if (vb!=NULL) {
    sscanf(vb,"%f",&v);
    return v;
  }

  if (dbglevel>=1) printf("Database::GetFloat: %s\n",Error());
  return (float)1e38;
}

double Database::GetDouble(int rc) {
  char *vb=GetString(rc);
  double v;
  if (vb!=NULL) {
    sscanf(vb,"%lf",&v);
    return v;
  }

  if (dbglevel>=1) printf("Database::GetDouble: %s\n",Error());
  return 1e38;
}

char Database::GetChar(int rc) {
  char *vb=GetString(rc);
  char v;
  if (vb!=NULL) {
    v=vb[0];
    return v;
  }

  if (dbglevel>=1) printf("Database::GetChar: %s\n",Error());
  return (char)255;
}

char *Database::GetString(int rc) {
  if (result!=NULL) {
    if ((0<PQntuples(result)) && (rowid<PQntuples(result))) {
      if ((0<PQnfields(result)) && (rc<PQnfields(result))) {
	return PQgetvalue(result,rowid,rc);
      }
    }
  }

  if (dbglevel>=1) printf("Database::GetString: %s\n",Error());
  return NULL;
}

Database::~Database() {
  if (dbglevel>=2) printf("Database::~Database\n");
}
