#include <stdio.h>
#include <stdlib.h>

#include <base/globalsdef.h>
#include <base/test.h>
#include <base/util/StringBuffer.h>

USE_NAMESPACE

#include <string>
#include <memory>
#include <vector>
#include <utility>
#include <sstream>
#include <iomanip>
#include <fstream>
#include <iostream>
#include <algorithm>

#include <client/DMTClientConfig.h>
#include <client/RawFileSyncSource.h>
#include <spds/SyncSource.h>

#include <spds/DefaultConfigFactory.h>

#include <client/SyncClient.h>

#define APPLICATION_URI "http://asterix:8080/funambol/ds"
#define SOURCE_NAME "cal"
#define CLIENT_DEVICE_ID "1"
#define WSOURCE_NAME "e01.ics"

int main(int argc, char** argv) {

    // 1.Create the empty config.
    DMTClientConfig config(APPLICATION_URI);

    // 2.Read config from the storage.
    if ( !config.read() ) {
        // 2.1.Config not found -> generate a default config

        AccessConfig* ac = DefaultConfigFactory::getAccessConfig();
        // Set some client defined defaults here
        // ....
        config.setAccessConfig(*ac);
        delete ac;

        DeviceConfig* dc = DefaultConfigFactory::getDeviceConfig();
        // Set some client defined defaults here
        dc->setDevID( CLIENT_DEVICE_ID );
        dc->setMan ("Funambol");
        // dc->setVer ("1.0");
        // etc....
        config.setDeviceConfig(*dc);
        delete dc;

        SyncSourceConfig* sc = DefaultConfigFactory::getSyncSourceConfig(SOURCE_NAME);
        // Set some defaults for this SyncSource here
        sc->setEncoding("plain/text");
        sc->setType    ("text");
        sc->setURI     ("briefcase");
        // etc....
        config.setSyncSourceConfig(*sc);
        delete sc;
    }

    // 3.Create the SyncSource passing its name and its config.
    // TestSyncSource source(WSOURCE_NAME, config.getSyncSourceConfig(SOURCE_NAME));

    RawFileSyncSource source(WSOURCE_NAME, config.getSyncSourceConfig(SOURCE_NAME));
    SyncSource* ssArray[] = { &source, NULL };

    // 4.Create the SyncClient passing the config.
    SyncClient sampleClient;

    // 5.Start the sync!
    if( sampleClient.sync(config, ssArray) ) {
        LOG.error("Error in sync.");
    }

    // 6.Print sync results.
    StringBuffer res;
	sampleClient.getSyncReport()->toString(res);
    printf("\n%s", res.c_str());

    // 7.Save config to the storage.
    config.save();

}
