%{

#ifdef __cplusplus
extern "C" {
#endif
  #include <stdlib.h>
  #include <stdio.h>
  #include <ctype.h>
  #include <string.h>
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <unistd.h>
  #include <time.h>
#ifdef __cplusplus
}
#endif

#include <parser.h>
static unsigned long l=0;

#include <vCalendar.h>
#include <vCard.h>
#include <List.h>

static vCalendar *ics=NULL;
static vEvent    *event=NULL;
static vCard *vcf=NULL;
static vTimezone *tz=NULL;
static TZ_DEF *tzdef=NULL;
static int boolval=0;
static vRRule *rrule=NULL;
static vRRule::FTYPE ftype=vRRule::F_NO_FREQUENCY;
static vAttendee* attendee=NULL;
static Date date;
static vAlarm *valarm;
static List<int> numberlist;
static int wdnum=0;
static WEEKDAY wd;
static List<OrderedWeekday> weekdaylist;
static int token_initialized=0;

char PUID[1024];

void InitParser() {
  
}

%}

%start ROOT
%token CHAR NUMBER STRING EOLSTRING IDENT TIMESTAMP
%token COLON SEMICOLON EQUALS COMMA BEGIN END
%token CSTERMINATOR

%token TRUE FALSE FREE BUSY OOF BUSY_UNAVAILABLE BUSY_TENTATIVE
%token PUBLIC PRIVATE

%token VTIMEZONE TZID TZNAME STANDARD DAYLIGHT TZOFFSETFROM TZOFFSETTO BEGIN_TZDEF END_TZDEF

%token X_MICROSOFT_CDO_ALLDAYEVENT
%token X_KDE_LIBKCAL

%token VALUE DATE_TIME DATE

%token VCALENDAR VEVENT VTODO VJOURNAL VFREEBUSY
%token PRODID VERSION METHOD PUBLISH
%token CN
%token CLASS STATUS
%token CYTPYE INDIVIDUAL GROUP RESOURCE ROOM UNKOWN
%token DELEGATED_FROM DELEGATED_TO

%token EVENT_BEGIN EVENT_END

%token RRULE RDATE EXDATE
%token ALARM_BEGIN ALARM_END ACTION DISPLAY AUDIO EMAIL PROCEDURE X_MAPI_WHEN

%token TRIGGER LAST_TRIGGERED DURATION RELATED START MINUS PLUS DUR_P DUR_T DUR_H DUR_M DUR_S DUR_D DUR_W REPEAT
%token ATTACH FMTTYPE

%token ORGANIZER ATTENDEE
%token ROLE CHAIR REQ_PARTICIPANT OPT_PARTICIPANT NON_PARTICIPANT
%token PARTSTAT NEEDS_ACTION ACCEPTED DECLINED TENTATIVE DELEGATED

%token RSVP
%token T_UID REV SEQUENCE SUMMARY LOCATION DESCRIPTION CATEGORIES
%token DTSTAMP DTSTART DTEND LAST_MODIFIED CREATED
%token X_MICROSOFT_CDO_BUSYSTATUS PRIORITY
%token X_MOZ_LASTACK X_MOZ_GENERATION

%token TRANSP X_MAPI_LABEL X_MICROSOFT_CDO_IMPORTANCE

%token CARD_BEGIN CARD_END TEL
%token N FN NICKNAME X_FILE_AS ORG TITLE NOTE BDAY FBURL PHOTO
%token ENCODING EIGHT_BIT BASE64 QUOTED_PRINTABLE BINARY
%token ADR LABEL URL EMAIL
%token WORK HOME CELL CAR PAGER PREF AGENT POSTAL
%token VOICE FAX TLX ISDN INTERNET

%token FREQ SECONDLY MINUTELY HOURLY DAILY WEEKLY MONTHLY YEARLY 
%token COUNT UNTIL INTERVAL BYMONTH BYYEARDAY BYMONTHDAY BYWEEKNO BYDAY BYHOUR BYMINUTE 
%token BYSECOND WKST BYSETPOS MO TU WE TH FR SA SU 

%token TENTATIVE CONFIRMED CANCELLED
%token OPAQUE TRANSPARENT

%%

ROOT: { if (ics) { delete ics; ics=NULL; } } ICS_CAL | { if (vcf) { delete vcf; vcf=NULL; } } VCF_CARD;

ICS_CAL: BEGIN COLON VCALENDAR {
  ics=new vCalendar();
  strcpy(PUID,"");
} CAL_HEAD CAL_BODY END COLON VCALENDAR;

CAL_HEAD: | CAL_HEAD_PROP CAL_HEAD;

CAL_HEAD_PROP: CAL_PRODID | CAL_VERSION | CAL_METHOD | CAL_TZDEF;

CAL_PRODID: PRODID COLON { 
  char buffer[8192]; 
  if (ScanEOL(buffer,8192)>0) {
    ics->ProductID(buffer);
  } 
};

CAL_VERSION: VERSION COLON { 
  char buffer[8192]; 
  if (ScanEOL(buffer,8192)>0) { 
    ics->Version(buffer);
  } 
};

CAL_METHOD: METHOD COLON PUBLISH;

CAL_TZDEF: BEGIN_TZDEF { 
  tz=new vTimezone(); 
} TZD_OPTIONS TZZ END_TZDEF { 
  ics->AddTimezone(tz);
};

TZZ: TZ1 | TZ2;
TZ1: TZD_STD TZD_DAYLIGHT;
TZ2: TZD_DAYLIGHT TZD_STD;

TZD_OPTIONS: | TZD_OPTION TZD_OPTIONS;

TZD_OPTION: TZD_NAME | LAST_MODIFIED COLON TIMESTAMP;

TZD_NAME: TZID COLON  { 
  char buffer[8192]; 
  if (ScanEOL(buffer,8192)>0) {
    tz->Name(buffer);
  }
};

TZD_STD:  BEGIN COLON STANDARD { tzdef=new TZ_DEF(); } TZD_RULE END COLON STANDARD { tz->SetStandard(tzdef); };
TZD_DAYLIGHT: BEGIN COLON DAYLIGHT { tzdef=new TZ_DEF(); } TZD_RULE END COLON DAYLIGHT { tz->SetDaylight(tzdef); };
TZD_RULE: | TZD_RULE_ITEM TZD_RULE;
TZD_RULE_ITEM: TZD_START | TZD_OFFSETFROM | TZD_OFFSETTO | TZD_RRULE | TZD_NAMING;
TZD_START: | DTSTART COLON TIMESTAMP { tzdef->DTStart(Date(ts.dd,ts.mm,ts.yy,ts.hh,ts.mi,ts.ss)); };
TZD_OFFSETFROM: | TZOFFSETFROM COLON NUMBER { 
  int num; 
  sscanf(number,"%d",&num); 
  tzdef->OffsetFrom((num/100)*3600+(num % 100)*60); 
};
TZD_OFFSETTO:   | TZOFFSETTO COLON NUMBER {
  int num; 
  sscanf(number,"%d",&num); 
  tzdef->OffsetTo((num/100)*3600+(num % 100)*60); 
};
TZD_RRULE: | REPRULE { tzdef->RRule(rrule); };
TZD_NAMING: TZNAME COLON IDENT;

//---------------------------------------------------------------------------------------------------------------------

CAL_BODY: | EVENT CAL_BODY;

EVENT: EVENT_BEGIN { 
  event=new vEvent(NULL,ics->Timezones()); 
} EVENT_BODY EVENT_END { 
  ics->AddEvent(event); 
};

EVENT_BODY: | EVENT_PROPERTY EVENT_BODY;

EVENT_PROPERTY: EVENT_DTSTAMP | EVENT_DTSTART | EVENT_DTEND | EVENT_MODIFIED | EVENT_CREATED | EVENT_EXDATE | EVENT_ALLDAY | 
                REPRULE { event->RRule(rrule); } | EVENT_ORGANIZER | EVENT_ATTENDEE | EVENT_UID | EVENT_SEQUENCE | 
                EVENT_SUMMARY | EVENT_LOCATION | EVENT_CATEGORIES | EVENT_CLASS | EVENT_STATUS | EVENT_BUSYSTATE |
                EVENT_DESCRIPTION | EVENT_PRIORITY | EVENT_IMPORTANCE | EVENT_TRANSP | EVENT_MAPILABEL | EVENT_ALARM |
                MOZ_EVENT_LASTACK | MOZ_EVENT_GENERATION | EVENT_X_KDE_LIBKCAL;

EVENT_ALLDAY: X_MICROSOFT_CDO_ALLDAYEVENT COLON BOOLEAN { 
  event->AllDay(boolval); 
};

EVENT_MAPILABEL: X_MAPI_LABEL COLON NUMBER { 
  int x; sscanf(number,"%d",&x); event->MAPILabel(x); 
};

EVENT_IMPORTANCE: X_MICROSOFT_CDO_IMPORTANCE COLON NUMBER { 
  int x; sscanf(number,"%d",&x); event->CDOImportance(x); 
};

EVENT_CATEGORIES: CATEGORIES COLON { 
  char buffer[8192]; if (ScanEOL(buffer,8192)>0) { event->Category(buffer); } 
};

EVENT_CLASS: CLASS COLON { 
  char buffer[8192]; if (ScanEOL(buffer,8192)>0) { event->Class(buffer); } 
};

EVENT_X_KDE_LIBKCAL: X_KDE_LIBKCAL COLON {
  char buffer[8192]; if (ScanEOL(buffer,8192)>0) { } 
};

EVENT_STATUS: STATUS COLON EVENT_STATES;

EVENT_STATES: TENTATIVE { event->Status(vEvent::S_TENTATIVE); } | 
              CONFIRMED { event->Status(vEvent::S_CONFIRMED); } | 
              CANCELLED { event->Status(vEvent::S_CANCELLED); };

EVENT_TRANSP: TRANSP COLON EVENT_TRANSPARENCY;
EVENT_TRANSPARENCY: OPAQUE { event->Transparency(vEvent::T_OPAQUE); } | 
                    TRANSPARENT { event->Transparency(vEvent::T_TRANSPARENT); };

// ORGANIZER;CN=schmei:MAILTO:n.schmeisser@fzd.de

EVENT_ORGANIZER: ORGANIZER { attendee=new vAttendee("ORGANIZER"); } EV_ATT_OPT COLON
{
  char buffer[8192];
  if (ScanEOL(buffer,8192)>0) { attendee->Name(buffer); } 
  attendee->Role(vAttendee::R_ORGANIZER);
  event->Organizer(attendee);
};

EVENT_ATTENDEE: ATTENDEE { attendee=new vAttendee("ATTENDEE"); } EV_ATT_OPT COLON { 
  char buffer[8192]; 
  if (ScanEOL(buffer,8192)>0) { attendee->Name(buffer); } 
  event->AddAttendee(attendee);
};

EV_ATT_OPT: SEMICOLON EV_ATT_OPTION EV_ATT_OPT | ;
EV_ATT_OPTION: EV_ATT_ROLE | EV_ATT_PARTSTAT | EV_ATT_RSVP | EV_ATT_CN;

EV_ATT_ROLE: ROLE EQUALS P_ROLE;
P_ROLE: CHAIR { attendee->Role(vAttendee::R_CHAIR); } | 
        REQ_PARTICIPANT { attendee->Role(vAttendee::R_REQ_PARTICIPANT); } | 
        OPT_PARTICIPANT { attendee->Role(vAttendee::R_OPT_PARTICIPANT); } | 
        NON_PARTICIPANT { attendee->Role(vAttendee::R_NON_PARTICIPANT); } ;

EV_ATT_PARTSTAT: PARTSTAT EQUALS P_STAT;
P_STAT: NEEDS_ACTION { attendee->Partstat(vAttendee::PS_NEEDS_ACTION); } | 
        ACCEPTED { attendee->Partstat(vAttendee::PS_ACCEPTED); } | 
        DECLINED { attendee->Partstat(vAttendee::PS_DECLINED); } | 
        TENTATIVE { attendee->Partstat(vAttendee::PS_TENTATIVE); } ;

EV_ATT_RSVP: RSVP EQUALS BOOLEAN { attendee->Rsvp(boolval); };

EV_ATT_CN: CN EQUALS CN_VALUE;
CN_VALUE:  { 
	     char buffer[8192];
	     if (ScanTerm(buffer,8192,":")>0) {
	       attendee->Cn(buffer);
	     } else {
	       if (ScanTerm(buffer,8192,";")>0) {
		 attendee->Cn(buffer);
	       }
	     }
	   };
// STRING { attendee->Cn(string); } |

EVENT_DTSTART: DTSTART { date=NoDate(); } DT_OPTIONS COLON  TIMESTAMP {
  date.Day(ts.dd); date.Month(ts.mm); date.Year(ts.yy); date.Hour(ts.hh); date.Minute(ts.mi); date.Second(ts.ss);
  event->DTStart(date);
};

EVENT_DTEND:   DTEND { date=NoDate(); } DT_OPTIONS COLON TIMESTAMP { 
  date.Day(ts.dd); date.Month(ts.mm); date.Year(ts.yy); date.Hour(ts.hh); date.Minute(ts.mi); date.Second(ts.ss);
  event->DTEnd(date);
};

EVENT_DTSTAMP: DTSTAMP { date=NoDate(); } DT_OPTIONS COLON TIMESTAMP {
  date.Day(ts.dd); date.Month(ts.mm); date.Year(ts.yy); date.Hour(ts.hh); date.Minute(ts.mi); date.Second(ts.ss);
  event->DTStamp(date);
  Date d=event->LastModified();
  if (d==NoDate()) event->LastModified(date);
};

EVENT_MODIFIED: LAST_MODIFIED { date=NoDate(); } DT_OPTIONS COLON TIMESTAMP {
  date.Day(ts.dd); date.Month(ts.mm); date.Year(ts.yy); date.Hour(ts.hh); date.Minute(ts.mi); date.Second(ts.ss);
  event->LastModified(date);
  Date d=event->DTStamp();
  if (d==NoDate()) event->DTStamp(date);
};

EVENT_EXDATE: EXDATE { date=NoDate(); } DT_OPTIONS COLON TIMESTAMP {
  date.Day(ts.dd); date.Month(ts.mm); date.Year(ts.yy); date.Hour(ts.hh); date.Minute(ts.mi); date.Second(ts.ss);
  // event->Created(date);
  // what is exdate used for???
};

EVENT_CREATED: CREATED { date=NoDate(); } DT_OPTIONS COLON TIMESTAMP {
  date.Day(ts.dd); date.Month(ts.mm); date.Year(ts.yy); date.Hour(ts.hh); date.Minute(ts.mi); date.Second(ts.ss);
  event->Created(date);
};

DT_OPTIONS: | SEMICOLON DT_OPTION DT_OPTIONS;
DT_OPTION: TZ_OPTION | DV_OPTION;
TZ_OPTION: TZID EQUALS STRING { date.Timezone(string); } | TZID EQUALS IDENT { date.Timezone(string); };
DV_OPTION: VALUE EQUALS DT_VALUE;
DT_VALUE: DATE_TIME { date.Type(Date::DT_DATETIME); } | 
          DATE { date.Type(Date::DT_DATE); };

EVENT_UID: T_UID COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { event->UID(buffer); strcpy(PUID,buffer); } };

EVENT_SEQUENCE: SEQUENCE COLON NUMBER { int n; sscanf(number,"%d",&n); event->Sequence(n); };

EVENT_SUMMARY: SUMMARY COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { event->Subject(buffer); } };

EVENT_LOCATION: LOCATION COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { event->Location(buffer); } };
EVENT_BUSYSTATE: X_MICROSOFT_CDO_BUSYSTATUS COLON EV_BSTATE;
EV_BSTATE: BUSY { event->Busystate(vEvent::B_BUSY); } | 
           FREE { event->Busystate(vEvent::B_FREE); } | 
           OOF { event->Busystate(vEvent::B_OOF); } | 
           BUSY_UNAVAILABLE { event->Busystate(vEvent::B_UNAVAILABLE); } | 
           BUSY_TENTATIVE { event->Busystate(vEvent::B_TENTATIVE); } ;

EVENT_DESCRIPTION: DESCRIPTION COLON { 
  char buffer[8192]; 
  if (ScanEOL(buffer,8192)>0) { 
    event->Body(buffer); 
  } 
};
EVENT_PRIORITY: PRIORITY COLON NUMBER { int n; sscanf(number,"%d",&n); event->Priority(n); };

// RRULE:FREQ=YEARLY;BYMONTH=3;BYDAY=-1SU
// do parsing later
REPRULE: RDATE RDATE_OPTIONS COLON TIMESTAMPLIST | 
         EXDATE EXDATE_OPTIONS COLON TIMESTAMPLIST |
         RRULE COLON { rrule=new vRRule(); } FREQUENCY REPRULESET;

RDATE_OPTIONS: | SEMICOLON RDATE_OPTION RDATE_OPTIONS;
RDATE_OPTION: VALUE EQUALS TIMESTAMP | TZ_OPTION;

EXDATE_OPTIONS: | SEMICOLON EXDATE_OPTION EXDATE_OPTIONS;
EXDATE_OPTION: VALUE EQUALS TIMESTAMP | TZ_OPTION;

TIMESTAMPLIST: TIMESTAMP | TIMESTAMP COMMA TIMESTAMPLIST;

FREQUENCY: FREQ EQUALS FREQVAL;
FREQVAL: SECONDLY { rrule->Frequency(vRRule::F_SECONDLY); } |
         MINUTELY { rrule->Frequency(vRRule::F_MINUTELY); } | 
         HOURLY   { rrule->Frequency(vRRule::F_HOURLY); } | 
         DAILY    { rrule->Frequency(vRRule::F_DAILY); } | 
         WEEKLY   { rrule->Frequency(vRRule::F_WEEKLY); } | 
         MONTHLY  { rrule->Frequency(vRRule::F_MONTHLY); } | 
	 YEARLY   { rrule->Frequency(vRRule::F_YEARLY); } ;

REPRULESET: | SEMICOLON REPRULEITEM REPRULESET;
REPRULEITEM: REP_COUNT | REP_UNTIL | REP_INTERVAL | REP_BYMONTH | REP_BYDAY | 
             REP_WKST | REP_BYMONTHDAY | REP_BYYEARDAY | REP_BYWEEKNO |
             REP_BYSETPOS | REP_BYHOUR | REP_BYMINUTE | REP_BYSECOND;

REP_COUNT: COUNT EQUALS NUMBER { int c; sscanf(number,"%d",&c); rrule->Count(c); };
REP_UNTIL: UNTIL EQUALS TIMESTAMP { 
  rrule->Until(Date(ts.dd,ts.mm,ts.yy,ts.hh,ts.mi,ts.ss)); 
};
REP_INTERVAL: INTERVAL EQUALS NUMBER { int i; sscanf(number,"%d",&i); rrule->Interval(i); };

REP_BYMONTH: BYMONTH EQUALS LIST_OF_NUMBER { 
  for (int i=0;i<numberlist.Size();i++) rrule->AddMonth(numberlist.Get(i)); 
};

REP_BYYEARDAY: BYYEARDAY EQUALS LIST_OF_NUMBER { 
  for (int i=0;i<numberlist.Size();i++) rrule->AddYearDay(numberlist.Get(i)); 
};

REP_BYMONTHDAY: BYMONTHDAY EQUALS LIST_OF_NUMBER {
  for (int i=0;i<numberlist.Size();i++) rrule->AddMonthDay(numberlist.Get(i)); 
};

REP_BYWEEKNO: BYWEEKNO EQUALS LIST_OF_NUMBER { 
  for (int i=0;i<numberlist.Size();i++) rrule->AddWeekNo(numberlist.Get(i)); 
};

REP_BYDAY: BYDAY EQUALS ORDERED_LIST_OF_WEEKDAY {
  if (rrule->Frequency()==vRRule::F_MONTHLY) rrule->Frequency(vRRule::F_MONTHLYBYDAY);
  for (int i=0;i<weekdaylist.Size();i++) rrule->AddWeekday(weekdaylist.Get(i)); 
};

REP_BYHOUR: BYHOUR EQUALS LIST_OF_NUMBER { 
  for (int i=0;i<numberlist.Size();i++) rrule->AddHour(numberlist.Get(i)); 
};

REP_BYMINUTE: BYMINUTE EQUALS LIST_OF_NUMBER { 
  for (int i=0;i<numberlist.Size();i++) rrule->AddMinute(numberlist.Get(i)); 
};

REP_BYSECOND: BYSECOND EQUALS LIST_OF_NUMBER { 
  for (int i=0;i<numberlist.Size();i++) rrule->AddSecond(numberlist.Get(i)); 
};

REP_WKST: WKST EQUALS WEEKDAY { rrule->Wkst(wd); };

REP_BYSETPOS: BYSETPOS EQUALS LIST_OF_NUMBER { 
  for (int i=0;i<numberlist.Size();i++) rrule->AddPos(numberlist.Get(i)); 
};

ORDERED_LIST_OF_WEEKDAY: { weekdaylist.Clear(); } OWD OLW;
OLW: | COMMA OWD OLW;
OWD: { wdnum=0; } ORDNUMOPT WEEKDAY { weekdaylist.Add(OrderedWeekday(wdnum,wd)); };
ORDNUMOPT: | NUMBER { sscanf(number,"%d",&wdnum); };
WEEKDAY: MO { wd=W_MO; } | TU { wd=W_TU; } | WE { wd=W_WE; } | TH { wd=W_TH; } | 
         FR { wd=W_FR; } | SA { wd=W_SA; } | SU { wd=W_SU; } ;

LIST_OF_NUMBER: { numberlist.Clear(); } SCAN_LIST_OF_NUMBER;
SCAN_LIST_OF_NUMBER: NUMBER COMMA { int n; sscanf(number,"%d",&n); numberlist.Add(n); } SCAN_LIST_OF_NUMBER |
                     NUMBER { int n; sscanf(number,"%d",&n); numberlist.Add(n); };

EVENT_ALARM: ALARM_BEGIN { valarm=new vAlarm(); } ALARM_PROPS ALARM_END { event->Alarm(valarm); };
ALARM_PROPS: | ALARM_PROPERTY ALARM_PROPS;
ALARM_PROPERTY: ALARM_ACTION | ALARM_DESCRIPTION | ALARM_TRIGGER | ALARM_LAST_TRIGGERED | ALARM_WHEN | ALARM_REPEAT | 
                ALARM_DURATION |
                ALARM_ATTACH | EVENT_ATTENDEE { valarm->AddAttendee(attendee); } | MOZ_EVENT_LASTACK;

ALARM_ACTION: ACTION COLON ALARM_ACTIONS;
ALARM_ACTIONS: AUDIO { valarm->Action(vAlarm::A_AUDIO); } | 
               DISPLAY { valarm->Action(vAlarm::A_DISPLAY); } | 
               EMAIL { valarm->Action(vAlarm::A_EMAIL); } | 
               PROCEDURE { valarm->Action(vAlarm::A_PROCEDURE); } ;

ALARM_DESCRIPTION: DESCRIPTION COLON { 
  char buffer[8192]; 
  if (ScanEOL(buffer,8192)>0) { 
    valarm->Description(buffer); 
  } 
};

ALARM_TRIGGER: TRIGGER TRIGGER_OPTIONS TRIGGERS { 
  valarm->Trigger(date);
};
TRIGGER_OPTIONS: | TRIGGER_OPTION TRIGGER_OPTIONS;
TRIGGER_OPTION: TRIGGER_RELATED;
TRIGGER_RELATED: RELATED EQUALS START { valarm->TriggerRelated(1); } | 
                 RELATED EQUALS END { valarm->TriggerRelated(2); };
TRIGGERS:      { date=NullDate(); date.Relative(1); } STD_TRIGGER | 
               { date=NullDate(); date.Relative(0); } TIMED_TRIGGER |
               { date=NullDate(); date.Relative(1); } DUR_TRIGGER;
STD_TRIGGER:   COLON TIME_DURATION;
DUR_TRIGGER:   SEMICOLON VALUE EQUALS DURATION COLON TIME_DURATION;
TIMED_TRIGGER: SEMICOLON VALUE EQUALS DATE_TIME COLON TIMESTAMP {
  date=Date(ts.dd,ts.mm,ts.yy,ts.hh,ts.mi,ts.ss);
};

ALARM_LAST_TRIGGERED: LAST_TRIGGERED COLON TIMESTAMP;

TIME_DURATION: DURATION_SIGN DUR_P DURATION_VALUE;
DURATION_SIGN: | PLUS | MINUS { date.Relative(-1); };
DURATION_VALUE: DURATION_DATE | DURATION_TIME | DURATION_WEEK;
DURATION_DATE: DURATION_DAY DURATION_TIME | DURATION_DAY;
DURATION_TIME: DUR_T DURATION_TVALUE;
DURATION_TVALUE: DURATION_H_ | DURATION_M_ | DURATION_S;
DURATION_H_: DURATION_H DURATION_M_ | DURATION_H;
DURATION_M_: DURATION_M DURATION_S  | DURATION_M;
DURATION_H: NUMBER DUR_H { int n; sscanf(number,"%d",&n); date.Hour(n); };
DURATION_M: NUMBER DUR_M { int n; sscanf(number,"%d",&n); date.Minute(n); };
DURATION_S: NUMBER DUR_S { int n; sscanf(number,"%d",&n); date.Second(n); };
DURATION_DAY: NUMBER DUR_D { int n; sscanf(number,"%d",&n); date.Day(n); };
DURATION_WEEK: NUMBER DUR_W { int n; sscanf(number,"%d",&n); date.Week(n); };

ALARM_WHEN: X_MAPI_WHEN COLON TIMESTAMP { 
  valarm->MAPIWhen(Date(ts.dd,ts.mm,ts.yy,ts.hh,ts.mi,ts.ss)); 
};

ALARM_REPEAT: REPEAT COLON NUMBER { int n; sscanf(number,"%d",&n); valarm->Repeat(n); };

ALARM_DURATION: DURATION COLON TIME_DURATION { valarm->Duration(date); };

ALARM_ATTACH: ATTACH ATTACH_OPTIONS COLON {
  char buffer[8192]; 
  if (ScanEOL(buffer,8192)>0) {
    valarm->Attach(buffer);
  } 
}
ATTACH_OPTIONS: | SEMICOLON ATTACH_OPTION ATTACH_OPTIONS;
ATTACH_OPTION: ATTACH_ENCODING | ATTACH_BINARY | ATTACH_FMT;
ATTACH_ENCODING: ENCODING EQUALS BASE64 { valarm->Base64(1); };
ATTACH_BINARY: VALUE EQUALS BINARY { valarm->Binary(1); };
ATTACH_FMT: FMTTYPE EQUALS {
  char buffer[8192];
  if (ScanTerm(buffer,8192,":")>0) {
    valarm->AttachFormat(buffer);
  }
};

MOZ_EVENT_LASTACK: X_MOZ_LASTACK COLON TIMESTAMP {
  valarm->MOZLastAck(Date(ts.dd,ts.mm,ts.yy,ts.hh,ts.mi,ts.ss));
};

MOZ_EVENT_GENERATION: X_MOZ_GENERATION COLON NUMBER {
};

//--- VCF Card ----------------------------------------------------------------------------------------------------------------------------

VCF_CARD: CARD_BEGIN CARD_BODY CARD_END;

CARD_BODY: | CARD_PROPERTY CARD_BODY;

CARD_PROPERTY: CARD_VERSION | CARD_N | CARD_FN | CARD_NICKNAME | CARD_FILEAS | CARD_ORG | CARD_TITLE | CARD_NOTE | CARD_TEL | CARD_ADDRESS |
               CARD_LABEL | CARD_URL | CARD_EMAIL | CARD_ROLE | CARD_BDAY | CARD_FBURL | CARD_CLASS | CARD_PHOTO | CARD_UID | CARD_REV;

CARD_VERSION: VERSION COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };

CARD_N: N COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };

CARD_FN: FN COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };

CARD_NICKNAME: NICKNAME COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };

CARD_FILEAS: X_FILE_AS COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };

CARD_ORG: ORG COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };

CARD_TITLE: TITLE COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };

CARD_NOTE: NOTE CARD_NOTE_OPTIONS COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* version is text in buffer */ } };
CARD_NOTE_OPTIONS: | SEMICOLON CARD_NOTE_OPTION CARD_NOTE_OPTIONS;
CARD_NOTE_OPTION: ENCODING_OPTION;

ENCODING_OPTION: ENCODING EQUALS ENCODING_VALUE;
ENCODING_VALUE: EIGHT_BIT | BASE64 | QUOTED_PRINTABLE;

CARD_TEL: TEL CARD_TEL_PROPLIST COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* text in buffer */ } };

CARD_TEL_PROPLIST: | SEMICOLON CARD_TEL_PROP CARD_TEL_PROPLIST;

CARD_TEL_PROP: CARD_COMM_AREA | CARD_COMM_TYPE;

CARD_COMM_AREA: WORK | HOME | CELL | CAR | PAGER | PREF | AGENT | POSTAL;
CARD_COMM_TYPE: VOICE | FAX | TLX | ISDN | INTERNET;

CARD_ADDRESS: ADR CARD_ADDRESS_OPTIONS COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* text in buffer */ } };
CARD_ADDRESS_OPTIONS: | SEMICOLON CARD_ADDRESS_OPTION CARD_ADDRESS_OPTIONS;
CARD_ADDRESS_OPTION: CARD_COMM_AREA | ENCODING_OPTION;

CARD_LABEL: LABEL CARD_LABEL_OPTIONS COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* text in buffer */ } };
CARD_LABEL_OPTIONS: | SEMICOLON CARD_LABEL_OPTION CARD_LABEL_OPTIONS;
CARD_LABEL_OPTION: CARD_COMM_AREA | CARD_COMM_TYPE | ENCODING_OPTION;

CARD_URL: URL CARD_URL_OPTIONS COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* text in buffer */ } };
CARD_URL_OPTIONS: | SEMICOLON CARD_URL_OPTION CARD_URL_OPTIONS;
CARD_URL_OPTION: CARD_COMM_AREA | CARD_COMM_TYPE | ENCODING_OPTION;

CARD_EMAIL: EMAIL CARD_EMAIL_OPTIONS COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* text in buffer */ } };
CARD_EMAIL_OPTIONS: | SEMICOLON CARD_EMAIL_OPTION CARD_URL_OPTIONS;
CARD_EMAIL_OPTION: CARD_COMM_AREA | CARD_COMM_TYPE | ENCODING_OPTION;

CARD_ROLE: ROLE COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* text in buffer */ } };

CARD_BDAY: BDAY COLON TIMESTAMP;

CARD_FBURL: FBURL COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* text in buffer */ } };

CARD_CLASS: CLASS COLON CARD_CLASS_TYPE;
CARD_CLASS_TYPE: PUBLIC | PRIVATE;

CARD_PHOTO: PHOTO CARD_PHOTO_OPTIONS COLON { 
  /* scan a base64 code block */ 
  char *bc=ScanBase64BBlock();
  if (bc) {
    printf("******************************************\n");
    printf("%s\n",bc);
    printf("******************************************\n");
  }
};

CARD_PHOTO_OPTIONS: | SEMICOLON CARD_PHOTO_OPTION CARD_PHOTO_OPTIONS;
CARD_PHOTO_OPTION: ENCODING_OPTION;

CARD_UID: T_UID COLON { char buffer[8192]; if (ScanEOL(buffer,8192)>0) { /* UID is text in buffer */ } };
CARD_REV: REV COLON TIMESTAMP;

BOOLEAN: TRUE { boolval=1; } | FALSE { boolval=0; };

%%

int yylex();

void InitTokens() {
  if (!token_initialized) {
    newtoken("BEGIN",BEGIN);
    newtoken("END",END);

    newtoken("INDIVIDUAL",INDIVIDUAL);
    newtoken("VCALENDAR",VCALENDAR);
    
    newtoken("VTIMEZONE",VTIMEZONE);
    newtoken("TZID",TZID);
    newtoken("TZNAME",TZNAME);
    newtoken("STANDARD",STANDARD);
    newtoken("DAYLIGHT",DAYLIGHT);
    newtoken("TZOFFSETFROM",TZOFFSETFROM);
    newtoken("TZOFFSETTO",TZOFFSETTO);
    newtoken("BEGIN:VTIMEZONE",BEGIN_TZDEF);
    newtoken("END:VTIMEZONE",END_TZDEF);
    
    newtoken("RRULE",RRULE);
    newtoken("RDATE",RDATE);
    newtoken("EXDATE",EXDATE);
    
    newtoken("VFREEBUSY",VFREEBUSY);
    newtoken("RESOURCE",RESOURCE);
    newtoken("VJOURNAL",VJOURNAL);
    newtoken("VERSION",VERSION);
    newtoken("PUBLISH",PUBLISH);
    newtoken("PRODID",PRODID);
    newtoken("VEVENT",VEVENT);
    newtoken("VTODO",VTODO);
    newtoken("REV",REV);
    
    newtoken("METHOD",METHOD);
    newtoken("CN",CN);
    newtoken("CYTPYE",CYTPYE);
    newtoken("GROUP",GROUP);
    newtoken("ROOM",ROOM);
    newtoken("UNKOWN",UNKOWN);
  
    newtoken("BEGIN:VEVENT",EVENT_BEGIN);
    newtoken("END:VEVENT",EVENT_END);
    
    newtoken("UID",T_UID);
    newtoken("SEQUENCE",SEQUENCE);
    newtoken("CATEGORIES",CATEGORIES);
    newtoken("SUMMARY",SUMMARY);
    newtoken("LOCATION",LOCATION);
    newtoken("DESCRIPTION",DESCRIPTION);
    
    newtoken("VALUE",VALUE);
    newtoken("DATE",DATE);
    newtoken("DATE-TIME",DATE_TIME);
    
    newtoken("DTSTAMP",DTSTAMP);
    newtoken("DTSTART",DTSTART);
    newtoken("DTEND",DTEND);
    newtoken("CREATED",CREATED);
    newtoken("PRIORITY",PRIORITY);
    newtoken("LAST-MODIFIED",LAST_MODIFIED);
    newtoken("DELEGATED-FROM",DELEGATED_FROM);
    newtoken("DELEGATED-TO",DELEGATED_TO);
    
    newtoken("FREE",FREE);
    newtoken("BUSY",BUSY);
    newtoken("OOF",OOF);
    newtoken("UNAVAILABLE",BUSY_UNAVAILABLE);
    newtoken("TENTATIVE",BUSY_TENTATIVE);
    
    newtoken("TRUE",TRUE);
    newtoken("FALSE",FALSE);
    
    newtoken("X-MICROSOFT-CDO-BUSYSTATUS",X_MICROSOFT_CDO_BUSYSTATUS);
    newtoken("X-MICROSOFT-CDO-ALLDAYEVENT",X_MICROSOFT_CDO_ALLDAYEVENT);
    newtoken("X-KDE-LIBKCAL-ID",X_KDE_LIBKCAL);
    
    newtoken("CLASS",CLASS);
    newtoken("STATUS",STATUS);
    
    newtoken("BEGIN:VALARM",ALARM_BEGIN);
    newtoken("END:VALARM",ALARM_END);
    newtoken("ACTION",ACTION);
    newtoken("DISPLAY",DISPLAY);
    newtoken("AUDIO",AUDIO);
    newtoken("EMAIL",EMAIL);
    newtoken("PROCEDURE",PROCEDURE);
    newtoken("TRIGGER",TRIGGER);
    newtoken("LAST-TRIGGERED",LAST_TRIGGERED);
    newtoken("X-MAPI-WHEN",X_MAPI_WHEN);
  
    newtoken("ORGANIZER",ORGANIZER);
    newtoken("ATTENDEE",ATTENDEE);
    
    newtoken("ROLE",ROLE);
    newtoken("CHAIR",CHAIR);
    newtoken("REQ-PARTICIPANT",REQ_PARTICIPANT);
    newtoken("OPT-PARTICIPANT",OPT_PARTICIPANT);
    newtoken("NON-PARTICIPANT",NON_PARTICIPANT);
    
    newtoken("PARTSTAT",PARTSTAT);
    newtoken("NEEDS-ACTION",NEEDS_ACTION);
    newtoken("ACCEPTED",ACCEPTED);
    newtoken("DECLINED",DECLINED);
    newtoken("TENTATIVE",TENTATIVE);
    newtoken("DELEGATED",DELEGATED);
    
    newtoken("RSVP",RSVP);
    newtoken("TRANSP",TRANSP);
    newtoken("X-MAPI-LABEL",X_MAPI_LABEL);
    newtoken("X-MICROSOFT-CDO-IMPORTANCE",X_MICROSOFT_CDO_IMPORTANCE);
    
    newtoken("X-MOZ-LASTACK",X_MOZ_LASTACK);
    newtoken("X-MOZ-GENERATION",X_MOZ_GENERATION);
    
    newtoken("BEGIN:VCARD",CARD_BEGIN);
    newtoken("END:VCARD",CARD_END);
    newtoken("N",N);
    newtoken("FN",FN);
    newtoken("NICKNAME",NICKNAME);
    newtoken("X-FILE-AS",X_FILE_AS);
    newtoken("ORG",ORG);
    newtoken("TITLE",TITLE);
    newtoken("NOTE",NOTE);
    newtoken("ENCODING",ENCODING);
    newtoken("8BIT",EIGHT_BIT);
    newtoken("BASE64",BASE64);
    newtoken("BINARY",BINARY);
    newtoken("QUOTED-PRINTABLE",QUOTED_PRINTABLE);
    newtoken("TEL",TEL);
    newtoken("ADR",ADR);
    newtoken("LABEL",LABEL);
    newtoken("URL",URL);
    newtoken("WORK",WORK);
    newtoken("HOME",HOME);
    newtoken("CELL",CELL);
    newtoken("CAR",CAR);
    newtoken("PAGER",PAGER);
    newtoken("ISDN",ISDN);
    newtoken("PREF",PREF);
    newtoken("AGENT",AGENT);
    newtoken("VOICE",VOICE);
    newtoken("FAX",FAX);
    newtoken("POSTAL",POSTAL);
    newtoken("TLX",TLX);
    newtoken("INTERNET",INTERNET);
    
    newtoken("BDAY",BDAY);
    newtoken("FBURL",FBURL);
    newtoken("PHOTO",PHOTO);
    
    newtoken("PUBLIC",PUBLIC);
    newtoken("PRIVATE",PRIVATE);
    
    newtoken("FREQ",FREQ);
    newtoken("SECONDLY",SECONDLY);
    newtoken("MINUTELY",MINUTELY);
    newtoken("HOURLY",HOURLY);
    newtoken("DAILY",DAILY);
    newtoken("WEEKLY",WEEKLY);
    newtoken("MONTHLY",MONTHLY);
    newtoken("YEARLY",YEARLY);
    newtoken("COUNT",COUNT);
    newtoken("UNTIL",UNTIL);
    newtoken("INTERVAL",INTERVAL);
    newtoken("BYMONTH",BYMONTH);
    newtoken("BYYEARDAY",BYYEARDAY);
    newtoken("BYMONTHDAY",BYMONTHDAY);
    newtoken("BYWEEKNO",BYWEEKNO);
    newtoken("BYDAY",BYDAY);
    newtoken("BYHOUR",BYHOUR);
    newtoken("BYMINUTE",BYMINUTE);
    newtoken("BYSECOND",BYSECOND);
    newtoken("WKST",WKST);
    newtoken("BYSETPOS",BYSETPOS);
    newtoken("MO",MO);
    newtoken("TU",TU);
    newtoken("WE",WE);
    newtoken("TH",TH);
    newtoken("FR",FR);
    newtoken("SA",SA);
    newtoken("SU",SU);
    
    newtoken("TENTATIVE",TENTATIVE);
    newtoken("CONFIRMED",CONFIRMED);
    newtoken("CANCELLED",CANCELLED);
    
    newtoken("OPAQUE",OPAQUE);
    newtoken("TRANSPARENT",TRANSPARENT);
    
    newtoken("DURATION",DURATION);
    newtoken("RELATED",RELATED);
    newtoken("START",START);
    newtoken("P",DUR_P);
    newtoken("T",DUR_T);
    newtoken("H",DUR_H);
    newtoken("M",DUR_M);
    newtoken("S",DUR_S);
    newtoken("D",DUR_D);
    newtoken("W",DUR_W);
    newtoken("REPEAT",REPEAT);
    newtoken("ATTACH",ATTACH);
    newtoken("FMTTYPE",FMTTYPE);
    
    newtoken(":",COLON);
    newtoken(";",SEMICOLON);
    newtoken(",",COMMA);
    newtoken("=",EQUALS);
    newtoken("+",PLUS);   newterm(PLUS,'P');
    newtoken("-",MINUS);  newterm(MINUS,'P');

    token_initialized=1;
  }
}

#include "parser.code"

