#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

// debugging
#include <readline/readline.h>
#include <readline/history.h>

#define LOGFILE "cg.log"

#include "imap.h"

//---------------------------------------------------------------------------------------------------

static void Normalize(char *code) {
  if (code) {
    char *p=code;
    unsigned int t;
    float tt=(float)strlen(code);
    
    // remove 0x0d (makes parsing easier, the generater MUST DO CRLF !!!
    t=0;
    while (*p!=0) {
      char c1=*p;

      t++;
      // printf("\r   transforming CRLF to LF: %3d %%",(int)((float)t/tt*100.0));
      // fflush(stdout);

      if (c1==0x0d) {
	char *p2=p+1;

	// move content beginning from p2 to p
	unsigned int i=0;
	do {
	  *(p+i)=*p2;
	  i++;
	  p2++;
	} while (*p2!=0);
	*(p+i)=*p2;
      } else {
	p++;
      }
    }
    // printf("\n");

    // printf("   unfolding lines");
    // "unfold lines, line is continued on next line, if next line starts with a space
    p=code;
    t=0;
    while (*p!=0) {
      char c1=*p;
      char c2=*(p+1);

      t++;
      // printf("\r   unfolding lines: %3d %%",(int)((float)t/tt*100.0));
      // fflush(stdout);

      if ((c1==0x0a) && (c2==0x20)) {
	char *p2=p+2;

	// move content beginning from p2 to p
	unsigned int i=0;
	do {
	  *(p+i)=*p2;
	  i++;
	  p2++;
	} while (*p2!=0);
	*(p+i)=*p2;
      } else {
	p++;
      }
    }
    // printf("\n");
  }
}

IMAPItem::IMAPItem(const char* header_,const char *text_) {
  contentclass=NULL;
  boundary=NULL;
  header=NULL;
  part=NULL;

  if ((header_) && (strlen(header_)>0)) {
    header=new char [strlen(header_)+1];
    strcpy(header,header_);
    Normalize(header);
    char *pheader=strstr(header,"\n")+1;
    //    printf("--- header --------------------------------------\n");
    //    printf("%s\n",pheader);
    //    printf("-------------------------------------------------\n");
    if (strstr(pheader,"Content-Type: multipart")) {
      char *bbptr=strstr(pheader,"boundary=\"");
      if (bbptr) {
	char *ebptr=strstr(bbptr+10,"\"");
	if (ebptr) {
	  boundary=new char [ebptr-bbptr-10+1+2];
	  memset(boundary,0,ebptr-bbptr-10+1+2);
	  strncpy(boundary+2,bbptr+10,ebptr-bbptr-10);
	  boundary[0]='-';
	  boundary[1]='-';
	  // printf("BOUNDARY is [%s]\n",boundary);
	} else {
	  printf("*** multipart, but boundary not terminated correctly\n");
	}
      } else {
	printf("*** multipart, but boundary not found\n");
      }
      if (boundary) {
	numparts=0;
	part=NULL;
	char *ptext=strstr(text_,"\n")+1;
	char *ptr=ptext;
	//	printf("--- text ----------------------------------------\n");
	//	printf("%s\n",ptr);
	//	printf("-------------------------------------------------\n");
	while ((ptr) && (strstr(ptr,boundary)==ptr)) {
	  char *ptr1=strstr(ptr+1,boundary);
	  if (ptr1) {
	    char **newpart=new char* [numparts+1];
	    for (int i=0;i<numparts;i++) {
	      newpart[i]=part[i];
	    }
	    newpart[numparts]=new char[ptr1-ptr-strlen(boundary)+1-3];
	    memset(newpart[numparts],0,ptr1-ptr-strlen(boundary)+1-3);
	    strncpy(newpart[numparts],ptr+strlen(boundary)+1,ptr1-ptr-strlen(boundary)-3);
	    if (part) delete part;
	    part=newpart;
	    //	    printf("Part %d\n",numparts);
	    //	    printf("-----------------------------------------------\n");
	    //	    printf("%s\n",part[numparts]);
	    //	    printf("-----------------------------------------------\n");
	    numparts++;
	    ptr=ptr1;
	  } else {
	    ptr=NULL;
	  }
	}
      } else {
	multipartmessage=0;
	numparts=1;
	part=new char* [1];
	part[0]=new char [strlen(text_)+1];
	strcpy(part[0],text_);
      }
    } else {
      multipartmessage=0;
      numparts=1;
      part=new char* [1];
      part[0]=new char [strlen(text_)+1];
      strcpy(part[0],text_);
    }
  } else {
    multipartmessage=0;
    numparts=1;
    part=new char* [1];
    part[0]=new char [strlen(text_)+1];
    strcpy(part[0],text_);
  }
}

const char *IMAPItem::ContentClass() {
  if (contentclass)
    return contentclass;
  return "";
}

int IMAPItem::MultiPart() {
  return multipartmessage;
}

int IMAPItem::NumParts() {
  return numparts;
}

const char *IMAPItem::Part(int i) {
  if (part) {
    if ((0<=i) && (i<numparts))
      return part[i];
    else
      return "";
  }
  return "";
}

int IMAPItem::WriteToFile(int i,const char *aname) {
  if (part) {
    if ((0<=i) && (i<numparts)) {
      FILE *f=fopen(aname,"w");
      if (f) {
	fwrite(part[i],1,strlen(part[i]),f);
	fclose(f);
      }
      return -1;
    } else
      return -2;
  }
  return -3;
}

IMAPItem::~IMAPItem() {
  if (boundary) delete boundary;
  if (contentclass) delete contentclass;
  if (header) delete header;
  if ((numparts>0) && (part)) {
    for (int i=0;i<numparts;i++) if (part[i]) delete part[i];
    delete part;
  }
}

//---------------------------------------------------------------------------------------------------

unsigned int IMAP::sequence;
int IMAP::PPDI[2];
int IMAP::PPDO[2];
int IMAP::pid;
int IMAP::ppid=-1;
int IMAP::connected;
char IMAP::lbuffer[BSIZE];
char IMAP::HOST[1024];

int IMAP::Readln(char *buffer,int bsize) {
  int result,rr=0;
  unsigned int i=0,j;
  char c;
  int fid=PPDO[0];

  memset(buffer,0,bsize);
  do {
    result=read(fid,&c,1);
    if (result==1){
      buffer[i++]=c;
    }
  } while ((result>0) && (c!='\n'));

  return result;
}

char* IMAP::ScanLine(char *buffer,char *ptr) {
  char *ptr2;
  char *ptrn;
  
  if (ptr!=NULL) {
    ptr2=strstr(buffer,ptr);
    if (ptr2!=NULL) {
      while ((*ptr2!=0) && (*ptr2!='\n')) {
	ptr2++;
      }
      if (*ptr2!=0) ptr2++;
    }
  } else {
    ptr2=buffer;
  }
  
  ptrn=strstr(ptr2,"\n");
  memset(lbuffer,0,BSIZE);
  strncpy(lbuffer,ptr2,ptrn-ptr2);
  
  return lbuffer;
}

IMAP::IMAP() {
  sequence=0;
  connected=0;
  ppid=-1;
  strcpy(HOST,"");
}

IMAP::IMAP(const char *host,const char *username,const char *password) {
  strcpy(HOST,host);
  sequence=0;
  connected=Login(username,password);
  ppid=-1;
}

IMAP::~IMAP() {
  Logout();
}

int IMAP::SendCommand(const char *scmd) {
  char cmd[2*BSIZE];
  sequence++;
  sprintf(cmd,"%03d %s",sequence,scmd);
  // send: cmd
  printf("OUT: %s",cmd);
  write(PPDI[1],cmd,strlen(cmd));
  
  return 0;
}

int IMAP::SendData(const char *data,unsigned int dsize) {
  //  printf("*** IMAP-CMD *** send %4d ****************************\n",dsize);
  //  printf("%s",data);
  //  printf("******************************************************\n");

  write(PPDI[1],data,dsize);
  
  return 0;
}

unsigned int IMAP::GetReply(char *data,char *emsg) {
  char buffer[BSIZE];
  char term[BSIZE];
  
  if (emsg!=NULL) {
    sprintf(term,"%03d ",sequence);

    // receive: * ......
    // receive data
    // receive: <seq> OK completed
    //          <seq> NO ...
    //          <seq> BAD ...
    strcpy(data,"");
    do {
      Readln(buffer,BSIZE);
      //if (strstr(buffer,term)!=buffer) {
      strcat(data,buffer);
      //}
    } while (strstr(buffer,term)!=buffer);
    
    strcpy(emsg,buffer+strlen(term));
    printf("received: '%s'\n",data);
    return strlen(data);
  } else {
    Readln(data,BSIZE);
    return strlen(data);
  }
}

int IMAP::Login(const char *uname,const char *password) {
  char cmd[BSIZE];
  char data[BSIZE];
  char emsg[BSIZE];
  int rsize;

  printf("ppid==%d\n",ppid);
  
  // fork and invoke "telnet localhost 143"
  if (ppid==-1) {
    if ((pipe(PPDI)==0) && (pipe(PPDO)==0)) {
      if ((ppid=fork())==0) {
	printf("forked, invoking telnet %s 143\n",HOST);
	// close stdout
	close(1);
	// make ppd0[1] to stdout of child
	dup(PPDO[1]);
	close(PPDO[0]);
	close(PPDO[1]);

	close(0);
	dup(PPDI[0]);
	close(PPDI[1]);
	close(PPDI[0]);

	execl("/usr/bin/telnet","telnet",HOST,"143",NULL);

	exit(0);
      }
    }
  }

  sleep(1);

  printf("Trying to login\n");
  printf("Getting Header\n");
  Readln(data,BSIZE);
  Readln(data,BSIZE);
  Readln(data,BSIZE);
  Readln(data,BSIZE);

  sprintf(cmd,"LOGIN %s %s\n",uname,password);
  SendCommand(cmd);
  rsize=GetReply(data,emsg);
printf("Login: %s\n",emsg);  
  if (strstr(emsg,"OK")) {
    return 1;
  }
  return 0;
}

int IMAP::Logout() {
  char cmd[BSIZE];
  char data[BSIZE];
  char emsg[BSIZE];
  int rsize;
  int ws;

  sprintf(cmd,"LOGOUT\n");
  SendCommand(cmd);
  rsize=GetReply(data,emsg);

  wait(&ws);
  
  if (strstr(emsg,"OK ")) {
    return 1;
  }
  return 0;
}

unsigned int IMAP::Select(const char *mbox) {
  char cmd[BSIZE];
  char data[BSIZE];
  char emsg[BSIZE];
  unsigned int n;
  unsigned int n_recent;
  char *ptr;
  
  sprintf(cmd,"SELECT \"%s\"\n",mbox);
  SendCommand(cmd);
  GetReply(data,emsg);

  n=0;
  if (strstr(emsg,"OK")) {
    ptr=NULL;
    do {
      ptr=ScanLine(data,ptr);
      // printf("%x:%s\n",ptr,ptr);
      if (ptr) {
	if (strstr(ptr,"EXISTS")) {
           sscanf(ptr,"* %d EXISTS",&n);
           return n;
        }
      }
    } while ((!strstr(ptr,"EXISTS")) && (ptr!=NULL));
  }
  
  return 0;
}

unsigned int IMAP::Recent(const char *mbox) {
  char cmd[BSIZE];
  char data[BSIZE];
  char emsg[BSIZE];
  unsigned int n;
  unsigned int n_recent;
  char *ptr;
  
  sprintf(cmd,"SELECT \"%s\"\n",mbox);
  SendCommand(cmd);
  GetReply(data,emsg);

  n=0;
  if (strstr(emsg,"OK")) {
    ptr=NULL;
    do {
      ptr=ScanLine(data,ptr);
      if (ptr) {
	if (strstr(ptr,"RECENT")) {
           sscanf(ptr,"* %d RECENT",&n);
           return n;
        }
      }
    } while ((!strstr(ptr,"RECENT")) && (ptr!=NULL));
  }
  
  return 0;
}

int IMAP::Unseen(const char *mbox) {
  char cmd[BSIZE];
  char data[BSIZE];
  char emsg[BSIZE];
  unsigned int n;
  unsigned int n_recent;
  char *ptr;

  sprintf(cmd,"SELECT \"%s\"\n",mbox);
  SendCommand(cmd);
  GetReply(data,emsg);

  n=0;
  if ((strstr(emsg,"OK")) && (strstr(data,"UNSEEN"))) {
    ptr=NULL;
    do {
      ptr=ScanLine(data,ptr);
      if (ptr) {
        if (strstr(ptr,"UNSEEN")) return 1;
      }
    } while ((ptr!=NULL) && (strlen(ptr)>0) && (!strstr(ptr,"UNSEEN")));
  }

  return 0;
}

int IMAP::Fetch(unsigned int id,const char *part,char *data,char *emsg) {
  char cmd[BSIZE];
  
  sprintf(cmd,"FETCH %d:%d (BODY[%s])\n",id,id,part);
  SendCommand(cmd);

  return GetReply(data,emsg);
}

int IMAP::Expunge(char *emsg) {
  char cmd[BSIZE];
  char data[BSIZE];
  int r;

  sprintf(cmd,"EXPUNGE\n");
  SendCommand(cmd);
  r=GetReply(data,emsg);
  return r;
}

int IMAP::Delete(unsigned int id,char *emsg) {
  char cmd[BSIZE];
  char data[BSIZE];
  int r;

  sprintf(cmd,"STORE %d:%d +FLAGS (\\Deleted)\n",id,id);
  SendCommand(cmd);
  r=GetReply(data,emsg);

  if (strstr(emsg,"OK completed")==emsg) {
    sprintf(cmd,"EXPUNGE\n");
    SendCommand(cmd);
    return Expunge(emsg);
  }

  return r;
}

int IMAP::Send(const char *folder,const char *subject,const char *uid,CTYPE ctype,const char *body) {
  char rdata[BSIZE],emsg[BSIZE];
  char imapcmd[BSIZE];
  char head[BSIZE];
  char buffer[2*BSIZE];
  unsigned int msgid=time(NULL);
  unsigned int msgsize;
  unsigned int msgsize2;
  struct tm *ct;
  time_t t;

  t=time(NULL);
  ct=localtime(&t);
    
  strcpy(head,"");
  
  sprintf(buffer,"From: <%s@%s>\n","sync4j","fzd.de");
  strcat(head,buffer);
    
  sprintf(buffer,"Subject: %s\n",subject);        
  strcat(head,buffer);
  
  sprintf(buffer,"Date: %s",asctime(ct)); // Sun, 07 Aug 2005 13:01:22 +0200\n");
  strcat(head,buffer);
  
  sprintf(buffer,"Message-Id: <%u.%s>\n",time(NULL),uid);
  strcat(head,buffer);
  
  if (ctype==CTYPE_VCALENDAR) {
    strcat(head,"content-class: urn:content-classes:appointment\n");
    strcat(head,"MIME-Version: 1.0\n");
    strcat(head,"Content-Type: text/calendar;charset=\"utf-8\"; "
	          "method=\"PUBLISH\"; name=\"meeting.ics\"\n");
  }
  if (ctype==CTYPE_VCARD) {
    strcat(head,"content-class: urn:content-classes:person\n");
    strcat(head,"MIME-Version: 1.0\n");
    strcat(head,"Content-Type: text/x-vcard; charset=\"utf-8\"\n");
  }

  strcat(head,"Content-Transfer-Encoding: 8bit\n");
  strcat(head,"\n");
  
  strcpy(buffer,head);
  strcat(buffer,body);

  //  printf("-- HEAD: %6d ------------------\n",strlen(head));
  //  printf("%s",head);
  //  printf("-- BODY: %6d ------------------\n",strlen(body));
  //  printf("%s",body);

  int cnl=-1;
  for (int i=0;i<strlen(buffer);i++) if (buffer[i]=='\n') cnl++;

  // printf("--  SUM: buffer size %6d / buffer size with \\r\\n %6d / buffer-size + 24 %6d ------\n",
  // strlen(buffer),strlen(buffer)+cnl,strlen(buffer)+24);
  
  sprintf(imapcmd,"APPEND \"%s\" (\134Seen) {%d}\n",folder,strlen(buffer)+cnl);
  SendCommand(imapcmd);
  GetReply(rdata);

  SendData(buffer,strlen(buffer));

  GetReply(rdata,emsg);
  printf("reply on data [%s]\n",emsg);

  return 1;
}

int IMAP::GetFolder(const char* mailbox,char **folder) {
  char cmd[BSIZE];
  char data[BSIZE];
  char emsg[BSIZE];
  char token[1024];
  int r;
  int nf=0;

  sprintf(token,"~%s/",mailbox);
  sprintf(cmd,"LIST \"~%s\" \"*\"\n",mailbox);
  SendCommand(cmd);
  r=GetReply(data,emsg);

  if (strstr(emsg,"OK")) {
    char *ptr=NULL;
    do {
      ptr=ScanLine(data,ptr);
      if (ptr) {
	if (!(strstr(ptr,"OK completed"))) {
	  char line[1024];
	  strcpy(line,ptr);
	  char *p=strstr(line,token);
	  if (p) {
	    folder[nf]=new char [strlen(p)-strlen(token)+1];
	    strcpy(folder[nf],p+strlen(token));
            nf++; 
	  }
	}
      }
    } while ((!strstr(ptr,"OK completed")) && (ptr!=NULL));
  }

  return nf;
}

