#ifndef IMAP_
#define IMAP_

#define BSIZE 512*1024

class IMAPItem {
  int multipartmessage;
  char *boundary;
  int numparts;
  char* contentclass;
  char* header;
  char** part;
 public:
  IMAPItem(const char* header_,const char *text_);
  const char *ContentClass();
  int MultiPart();
  int NumParts();
  const char *Part(int i);
  int WriteToFile(int ia,const char *aname);
  ~IMAPItem();
};

class IMAP {
  static char HOST[1024];

  static unsigned int sequence;
  static int PPDI[2],PPDO[2];
  static int pid,ppid;
  static int connected;
  static char lbuffer[BSIZE];

  static int Readln(char *buffer,int bsize);
  static char* ScanLine(char *buffer,char *ptr);
 public:
  typedef enum CTYPE { CTYPE_VCALENDAR , CTYPE_VCARD};

  IMAP();
  IMAP(const char *host,const char *username,const char *password);
  ~IMAP();
  static int SendCommand(const char *scmd);
  static int SendData(const char *data,unsigned int dsize);

  static unsigned int GetReply(char *data,char *emsg=NULL);
  static int Login(const char *uname,const char *password);
  static int Logout();
  static unsigned int Select(const char *mbox);
  static unsigned int Recent(const char *mbox);
  static int Unseen(const char *mbox);
  // part is either HEADER or TEXT
  static int Fetch(unsigned int id,const char *part,char *data,char *emsg);
  static int Expunge(char *emsg);
  static int Delete(unsigned int id,char *emsg);
  static int Send(const char *folder,const char *subject,const char *uid,CTYPE ctype,const char *body);
  static int GetFolder(const char* mailbox,char **folder);
};

#endif
