#include <stdlib.h>

extern FILE *LOGFILE;

typedef enum { TOKEN, PRAGMA } TokenFlag;

typedef struct {
  char *s;
  int token;
  unsigned int nterm;
  char term[128];
  TokenFlag flag;
} tokenstruct;

typedef struct {
  char *errormessage;
  unsigned int errorcode;
} errorstruct;

static unsigned long line_number=1;
static int istoken=0;
static int isterm=0;
static tokenstruct token[1024];
static unsigned long ntoken=0;

static tokenstruct comment_open[10],comment_close[10];
static unsigned int nctoken=0;

static unsigned int nerrors=0;
static errorstruct errors[1024];

char *cptr,*code_,*lcode;
static int scanident;
static int slength;
static char string[1024],number[1024],codebuffer[2048];
static char* codeblock=NULL;

typedef struct timestamp {
  int yy,mm,dd,hh,mi,ss;
  time_t t;
  int localtz;
};

static timestamp ts;

static int errcode;
static int acctoken;
static int tokenindex;

/*--- some functions ---*/

void newerror(char *errmsg,unsigned int errcode) {
  errors[nerrors].errormessage=new char [strlen(errmsg)+1];
  strcpy(errors[nerrors].errormessage,errmsg);
  errors[nerrors].errorcode=errcode;
  nerrors++;
}

void newcommenttoken(char *opening,char *closing) {
  comment_open[nctoken].s=new char [strlen(opening)+1];
  strcpy(comment_open[nctoken].s,opening);
  comment_open[nctoken].token=0;
  comment_open[nctoken].nterm=0;

  comment_close[nctoken].s=new char [strlen(closing)+1];
  strcpy(comment_close[nctoken].s,closing);
  comment_close[nctoken].token=0;
  comment_close[nctoken].nterm=0;

  nctoken++;
}

void newtoken(const char *s,int tt,TokenFlag flg=TOKEN) {
  // make a list sorted by length of the string
  if (ntoken>0) {
    int ni=0;
    while ((ni<ntoken) && (strlen(s)<=strlen(token[ni].s))) ni++;
    // while ((ni<ntoken) && (strcmp(s,token[ni].s)<0)) ni++;

    if (ni==ntoken) {
      token[ntoken].s=new char [strlen(s)+1];
      strcpy(token[ntoken].s,s);
      token[ntoken].token=tt;
      token[ntoken].nterm=0;
      token[ntoken].flag=flg;
      
      ntoken++;
    } else {
      for (int i=ntoken;i>ni;i--) {
	token[i].s=new char [strlen(token[i-1].s)+1];
	strcpy(token[i].s,token[i-1].s);
	token[i].token=token[i-1].token;
	token[i].nterm=token[i-1].nterm;
	token[i].flag=token[i-1].flag;
	delete token[i-1].s;
      }

      token[ni].s=new char [strlen(s)+1];
      strcpy(token[ni].s,s);
      token[ni].token=tt;
      token[ni].nterm=0;
      token[ni].flag=flg;

      ntoken++;
    }
  } else {
    token[ntoken].s=new char [strlen(s)+1];
    strcpy(token[ntoken].s,s);
    token[ntoken].token=tt;
    token[ntoken].nterm=0;
    token[ntoken].flag=flg;

    ntoken++;
  }
}

void newterm(int tt,char c) {
  unsigned int i;

  for (i=0;i<ntoken;i++) {
    if (token[i].token==tt) {
      token[i].term[token[i].nterm++]=c;
    }
  }
}

extern int yyparse();
extern void lexmode(int scterm);

void yyerror(char *s);
int yylex();

#define EOT 0

int Readln(FILE *f,char *buffer,int bsize=1024) {
  int result;
  unsigned int i=0;
  char c;

  do {
    result=fscanf(f,"%c",&c);
    if ((result!=EOF) && (c!='\n')) {
      buffer[i++]=c;
      buffer[i]=0;
    }
  } while ((result!=EOF) && (c!='\n') && (i<bsize));
  return result;
}

extern int ScanEOL(char *buffer,int bsize);
extern int ScanTerm(char *buffer,int bsize,const char *term);
extern char* ScanBase64BBlock();
