#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

// debugging
#include <readline/readline.h>
#include <readline/history.h>

#include <date.h>

extern void Normalize(char *code);

FILE *LOGFILE;
int LOGLEVEL;

#include <imap.h>
#include <db-interface.h>

#include <funambol.h>

#include <vCalendar.h>
#include <vCard.h>

#include <cfgtools.h>

extern vCalendar* ParseCalendar(char *code);
extern vCard* ParseCard(char *code);
extern const char *ParsedUID();

Database *db;
static time_t BIAS;
// FILE *smsf;

//------------------------------------------------------------------------

void LTS() {
  time_t t=time(NULL);
  char buffer[512];
  strcpy(buffer,ctime(&t));
  if (buffer[strlen(buffer)-1]=='\n') buffer[strlen(buffer)-1]=0;
  fprintf(LOGFILE,"%s: ",buffer);
}

//------------------------------------------------------------------------
static int ecnt_new_server,ecnt_new_client,ecnt_update,ecnt_update_server,ecnt_delete,ecnt_delete_server;

typedef enum DB_CAL_COL { CAL_id,CAL_userid,CAL_last_update,CAL_status,CAL_type,CAL_all_day,CAL_body,CAL_busy_status,CAL_categories,
			  CAL_companies,CAL_birthday,CAL_duration,CAL_dstart,CAL_dend,CAL_folder,CAL_importance,CAL_location,
			  CAL_meeting_status,CAL_mileage,CAL_reminder_time,CAL_reminder,CAL_reminder_sound_file,CAL_reminder_options,
			  CAL_reminder_repeat_count,CAL_sensitivity,CAL_subject,CAL_rec_type,CAL_rec_interval,CAL_rec_month_of_year,
			  CAL_rec_day_of_month,CAL_rec_day_of_week_mask,CAL_rec_instance,CAL_rec_start_date_pattern,
			  CAL_rec_no_end_date,CAL_rec_end_date_pattern,CAL_rec_occurrences,CAL_dstart_tz,CAL_dend_tz,CAL_reminder_tz,
			  CAL_reply_time,CAL_completed,CAL_percent_complete,CAL_suid,CAL_stouch };
typedef enum DB_ATT_COL { ATT_partstat, ATT_role,ATT_rsvp,ATT_cn,ATT_attendee };

int AppendAttendees(Database *db,vEvent *event) {
  char sql[1024];
  sprintf(sql,"select partstat,role,rsvp,cn,attendee from fnbl_pim_calendar_attendees where eid=%ld",event->GUID());
  if (db->SQL(sql)) {
    if (db->NumRows()>0) {
      vAttendee *a;
      for (int i=0;i<db->NumRows();i++) {
	a=new vAttendee("");
	a->Partstat((vAttendee::PARTSTAT)db->GetInt(ATT_partstat));
	a->Role((vAttendee::ROLE)db->GetInt(ATT_role));
	a->Rsvp(db->GetInt(ATT_rsvp));
	a->Cn(db->GetString(ATT_cn));
	a->Name(db->GetString(ATT_attendee));
	if (a->Role()==vAttendee::R_ORGANIZER)
	  event->Organizer(a);
	else
	  event->AddAttendee(a);
	db->NextRow();
      }
    }
  } else {
  }
}

vEvent *EventFromDB(Database *db,long ne) {
  vEvent *event=NULL;
  char sql[1024];

  sprintf(sql,
	  "select "
	  "id,userid,last_update/1000,status,type,all_day,body,busy_status,categories,companies,birthday,duration,dstart,dend,folder,"
	  "importance,location,meeting_status,mileage,reminder_time,reminder,reminder_sound_file,reminder_options,"
	  "reminder_repeat_count,sensitivity,subject,rec_type,rec_interval,rec_month_of_year,rec_day_of_month,rec_day_of_week_mask,"
	  "rec_instance,rec_start_date_pattern,rec_no_end_date,rec_end_date_pattern,rec_occurrences,dstart_tz,dend_tz,reminder_tz,"
	  "reply_time,completed,percent_complete,"
	  "suid,stouch"
	  " from fnbl_pim_calendar where id=%ld",ne);
  if (db->SQL(sql)) {
    if (db->NumRows()>0) {
      long id;
      char rec_start_date_pattern[64];
      unsigned long reminder_options;

      event=new vEvent();
      
      //type,companies,birthday,duration,folder,"
      //importance,meeting_status,mileage
      //sensitivity,dstart_tz,dend_tz,reminder_tz,"
      //reply_time,completed,percent_complete,"

      id =db->GetLong(CAL_id);
      event->GUID(id);
      event->UID(db->GetString(CAL_suid));
      event->Subject(db->GetString(CAL_subject));
      event->Category(db->GetString(CAL_categories));
      event->Body(db->GetString(CAL_body));
      event->Location(db->GetString(CAL_location));
      
      event->DTStart(Date(db->GetString(CAL_dstart)));

      event->DTEnd(Date(db->GetString(CAL_dend)));
      event->AllDay(db->GetInt(CAL_all_day));

      event->LastModified(Date((time_t)(db->GetUnsignedLong(CAL_last_update))));
      event->Created((time_t)(db->GetUnsignedLong(CAL_last_update)));
      
      switch (db->GetInt(CAL_busy_status)) {
      case olBusy:        event->Busystate(vEvent::B_BUSY); break;
      case olFree:        event->Busystate(vEvent::B_FREE); break;
      case olTentative:   event->Busystate(vEvent::B_TENTATIVE); break;
      case olOutOfOffice: event->Busystate(vEvent::B_OOF); break;
      default: event->Busystate(vEvent::B_NO_BUSYSTATE);
      }

      switch (db->GetInt(CAL_meeting_status)) {
      case olNonMeeting: event->Status(vEvent::S_NO_STATE); break;
      case olMeeting: event->Status(vEvent::S_CONFIRMED); break;
      case olMeetingReceived: event->Status(vEvent::S_TENTATIVE); break;
      case olMeetingCanceled: event->Status(vEvent::S_CANCELLED); break;
      default: event->Status(vEvent::S_NO_STATE);
      }

      //reminder_time,reminder,reminder_sound_file,reminder_options,reminder_repeat_count
      if (db->GetChar(CAL_reminder)) {
	vAlarm *alarm=new vAlarm();
	
	// make the trigger relative !!!
	Date d1=Date(db->GetString(CAL_reminder_time));
	Date dt=d1-event->DTStart();

	alarm->Trigger(d1);
	alarm->Attach(db->GetString(CAL_reminder_sound_file));
	
	unsigned long reminder_options=db->GetUnsignedLong(CAL_reminder_options);
	unsigned long a=vAlarm::A_NO_ALARMACTION;
	if (reminder_options & olLED)     a=a | vAlarm::A_LED;
	if (reminder_options & olVibrate) a=a | vAlarm::A_VIBRATE;
	if (reminder_options & olDialog)  a=a | vAlarm::A_DISPLAY;
	if (reminder_options & olSound)   a=a | vAlarm::A_AUDIO;
	alarm->Action(a);
	alarm->Repeat(db->GetUnsignedLong(CAL_reminder_repeat_count));

	event->Alarm(alarm);
      }

      int rec_type =db->GetInt(CAL_rec_type);
      if (rec_type>=0) {
	vRRule *rrule=new vRRule();

	rrule->Interval(db->GetInt(CAL_rec_interval));

	int rec_month_of_year =db->GetInt(CAL_rec_month_of_year);
	int rec_day_of_month =db->GetInt(CAL_rec_day_of_month);
	unsigned int rec_day_of_week_mask=db->GetUInt(CAL_rec_day_of_week_mask);
	int rec_instance =db->GetInt(CAL_rec_instance);
	if (!db->GetInt(CAL_rec_no_end_date))
	  rrule->Until(Date(db->GetString(CAL_rec_end_date_pattern)));
	  
	if (db->GetInt(CAL_rec_occurrences)>=0)
	  rrule->Count(db->GetInt(CAL_rec_occurrences));
	strcpy(rec_start_date_pattern,db->GetString(CAL_rec_start_date_pattern));

	switch (rec_type)
	  {
	  case  0: rrule->Frequency( vRRule::F_DAILY ); break; // DAILY
	  case  1: rrule->Frequency( vRRule::F_WEEKLY ); break; // WEEKLY
	  case  2: rrule->Frequency( vRRule::F_MONTHLY ); break; // MONTHLY (BYMONTHDAY) 
	  case  3: rrule->Frequency( vRRule::F_MONTHLYBYDAY ); break; // MONTHLY (BYDAY)
	  case  5: rrule->Frequency( vRRule::F_YEARLY ); break; // YEARLY
	  default: rrule->Frequency( vRRule::F_NO_FREQUENCY ); // NOREC
	  }
	if (rec_month_of_year)  rrule->AddMonth(rec_month_of_year);
	if (((rec_type==2) || (rec_type=5)) && (rec_day_of_month))   rrule->AddMonthDay(rec_day_of_month);
	if ((rec_type==3) && (rec_day_of_week_mask>0)) {
	  if (rec_day_of_week_mask & olSunday)    { rrule->AddWeekday(OrderedWeekday(rec_instance,W_SU)); }
	  if (rec_day_of_week_mask & olMonday)    { rrule->AddWeekday(OrderedWeekday(rec_instance,W_MO)); }
	  if (rec_day_of_week_mask & olTuesday)   { rrule->AddWeekday(OrderedWeekday(rec_instance,W_TU)); }
	  if (rec_day_of_week_mask & olWednesday) { rrule->AddWeekday(OrderedWeekday(rec_instance,W_WE)); }
	  if (rec_day_of_week_mask & olThursday)  { rrule->AddWeekday(OrderedWeekday(rec_instance,W_TH)); }
	  if (rec_day_of_week_mask & olFriday)    { rrule->AddWeekday(OrderedWeekday(rec_instance,W_FR)); }
	  if (rec_day_of_week_mask & olSaturday)  { rrule->AddWeekday(OrderedWeekday(rec_instance,W_SA)); }
	}

	event->RRule(rrule);
      }
      AppendAttendees(db,event);
    } else {
      return event; // this is a new entry
    }
  } else {
    if (LOGLEVEL>=1) {
      LTS();
      fprintf(LOGFILE,"*** error getting event pim item by id: [%d] %s\n",ne,db->Error());
      fflush(LOGFILE);
    }
  }

  return event;
}

vCard *ContactFromDB(Database *db,long ne) {
  vCard *card=NULL;
  char sql[1024];

  sprintf(sql,
	  "select id,last_update/1000,importance,sensitivity,"
	  "subject,folder,anniversary,first_name,middle_name,last_name,display_name,"
	  "birthday,body,categories,children,hobbies,initials,languages,nickname,spouse,"
	  "suffix,title,gender,assistant,company,department,job_title,manager,mileage,"
	  "office_location,profession,companies,suid"
	  " from fnbl_pim_contact where id=%ld",ne);
  if (db->SQL(sql)) {
    if (db->NumRows()>0) {
      unsigned long id;
      time_t last_update;
      char suid[512];
      int i;
      char buffer[4096];

      if (LOGLEVEL>=3) {
	LTS();
	fprintf(LOGFILE,"new card\n");
	fflush(LOGFILE);
      }
      
      card=new vCard();

      id =db->GetLong(0);
      card->ID(id);
      last_update=(long)(db->GetUnsignedLong(1));
      card->LastUpdate(last_update);
      card->Importance(db->GetInt(2));
      card->Sensitivity(db->GetInt(3));
      card->Subject(db->GetString(4));
      card->Folder(db->GetString(5));
      card->Anniversary(db->GetString(6));
      card->Firstname(db->GetString(7));
      card->Middlename(db->GetString(8));
      card->Surename(db->GetString(9));
      card->Displayname(db->GetString(10));
      card->Birthday(db->GetString(11));
      card->Note(db->GetString(12));
      card->Category(db->GetString(13));
      card->Children(db->GetString(14));
      card->Hobby(db->GetString(15));
      card->Initials(db->GetString(16));
      card->Language(db->GetString(17));
      card->Nickname(db->GetString(18));
      card->Spouse(db->GetString(19));
      card->Suffix(db->GetString(20));
      card->Title(db->GetString(21));
      card->Gender(db->GetString(22));
      card->Assistant(db->GetString(23));
      card->Company(db->GetString(24));
      card->Department(db->GetString(25));
      card->JobTitle(db->GetString(26));
      card->Manager(db->GetString(27));
      card->Mileage(db->GetString(28));
      card->OfficeLocation(db->GetString(29));
      card->Profession(db->GetString(30));
      card->Companies(db->GetString(31));

      card->UID(db->GetString(32));
      card->Subject("");
      if (LOGLEVEL>=3) {
	LTS();
	fprintf(LOGFILE,"address for %ld\n",ne);
	fflush(LOGFILE);
      }
      sprintf(sql,"select type,street,city,state,postal_code,country,po_box "
	          "from fnbl_pim_address where contact=%ld",ne);
      if (db->SQL(sql)) {
	if (db->NumRows()>0) {
	  vAddress *adr=new vAddress();

	  adr->Type(db->GetInt(0));
	  adr->Street(db->GetString(1));
	  adr->City(db->GetString(2));
	  adr->State(db->GetString(3));
	  adr->ZIP(db->GetString(4));
	  adr->Country(db->GetString(5));

	  if (LOGLEVEL>=3) {
	    LTS();
	    fprintf(LOGFILE,"  add address\n");
	    fflush(LOGFILE);
	  }
	  card->AddAddress(adr);
	}
      }

      sprintf(sql,"select type,value from fnbl_pim_contact_item where contact=%ld",ne);
      if (db->SQL(sql)) {
	if (db->NumRows()>0) {
	  do {
	    vCommData *comm=new vCommData();
	    comm->Type(db->GetInt(0));
	    comm->Value(db->GetString(1));
	    if (LOGLEVEL>=3) {
	      LTS();
	      fprintf(LOGFILE,"  add contact\n");
	      fflush(LOGFILE);
	    }
	    card->AddCommData(comm);
	  } while (db->NextRow());
	}
      }
      if (LOGLEVEL>=3) {
	LTS();
	fprintf(LOGFILE,"return card\n");
	fflush(LOGFILE);
      }
      return card;
    } else {
      return card; // this is a new entry
    }
  } else {
    if (LOGLEVEL>=1) {
      LTS();
      fprintf(LOGFILE,"*** error getting event pim item by id: [%d] %s\n",ne,db->Error());
      fflush(LOGFILE);
    }
  }

  return card;
}

/*
 eid      | integer                |
 partstat | integer |
 role     | integer |
 rsvp     | integer |
 cn       | character varying(255) |
 attendee | character varying(255) |
*/

int UpdateAttendees(Database *db,char *uname,vEvent *event) {
  char sql[4096];

  sprintf(sql,"delete from fnbl_pim_calendar_attendees where eid=%d",event->GUID());
  if (db->SQL(sql)) {
    if (event->Organizer()) {
      sprintf(sql,
	      "insert into fnbl_pim_calendar_attendees "
	      "( eid, partstat,role,rsvp,cn, attendee ) values "
	      "( %d,%d,%d,%d,'%s','%s' )",
	      event->GUID(),event->Organizer()->Partstat(),event->Organizer()->Role(),event->Organizer()->Rsvp(),
	      event->Organizer()->Cn(),event->Organizer()->Name());
      if (!db->SQL(sql)) {
	if (LOGLEVEL>=1) {
	  LTS();
	  printf("SQL: %s\n",sql);
	  LTS();
	  fprintf(LOGFILE,"*** update attendees for event: [%s] failed\n",event->UID());
	  fprintf(LOGFILE,"                          %s\n",db->Error());
	  fflush(LOGFILE);
	}
      }
    }
    if (event->NumAttendees()>0) {
      for (int i=0;i<event->NumAttendees();i++) {
	sprintf(sql,
		"insert into fnbl_pim_calendar_attendees "
		"( eid, partstat,role,rsvp,cn, attendee ) values "
		"( %d,%d,%d,%d,'%s','%s' )",
		event->GUID(),event->Attendee(i)->Partstat(),event->Attendee(i)->Role(),event->Attendee(i)->Rsvp(),
		event->Attendee(i)->Cn(),event->Attendee(i)->Name());
	if (!db->SQL(sql)) {
	  if (LOGLEVEL>=1) {
	    LTS();
	    printf("SQL: %s\n",sql);
	    LTS();
	  fprintf(LOGFILE,"*** update attendees for event: [%s] failed\n",event->UID());
	  fprintf(LOGFILE,"                          %s\n",db->Error());
	  fflush(LOGFILE);
	  }
	}
      }
    }
  } else {
    if (LOGLEVEL>=1) {
      LTS();
      printf("SQL: %s\n",sql);
      LTS();
      fprintf(LOGFILE,"*** update attendees for event: [%s] failed\n",event->UID());
      fprintf(LOGFILE,"                          %s\n",db->Error());
      fflush(LOGFILE);
    }
  }
  
  return -1;
}

int EventToDB(Database *db,char *uname,vEvent *event,unsigned int folderid) {
  char sql[4096];

  char dbuffer[256],dtstart[256],dtend[256],rstart[256],rend[256],almtrigger[256];
  time_t modified;
  Date d;

  d=event->DTStart();
  strcpy(dtstart,d.ISOChar());
  d=event->DTEnd();
  strcpy(dtend,d.ISOChar());

  strcpy(dbuffer,event->REC_StartDate());
  if (strlen(dbuffer)==0)
    strcpy(rstart,"null");
  else
    sprintf(rstart,"'%s'",dbuffer);

  strcpy(dbuffer,event->REC_EndDate());
  if (strlen(dbuffer)==0)
    strcpy(rend,"null");
  else
    sprintf(rend,"'%s'",dbuffer);

  // should be sometime AFTER last sync 
  // (select end_sync from fnbl_principal_last_sync where sync_source='cal' and username=%s)
  //  d=event->LastModified();
  //  if (d!=NoDate())
  //    modified=d.Time();
  //  else
  modified=time(NULL);
  
  if (event->Alarm()) {
    strcpy(dbuffer,event->ALM_Time());
    sprintf(almtrigger,"'%s'",dbuffer);
  } else {
    strcpy(almtrigger,"null");
  }

  char esc_subject[1024];
  char esc_body[8192];
  char esc_category[1024];
  char esc_location[1024];
  strcpy(esc_subject,ESCAPE(event->Subject()));
  strcpy(esc_body,ESCAPE(event->Body()));
  strcpy(esc_category,ESCAPE(event->Category()));
  strcpy(esc_location,ESCAPE(event->Location()));

  int bstate;
  switch (event->Busystate()) {
  case vEvent::B_BUSY: bstate=olBusy; break;
  case vEvent::B_FREE: bstate=olFree; break;
  case vEvent::B_TENTATIVE: bstate=olTentative; break;
  case vEvent::B_OOF: bstate=olOutOfOffice; break;
  default: bstate=olFree; // how to code NULL ?
  }

  int mstate;
  switch (event->Status()) {
  case vEvent::S_NO_STATE:  mstate=olNonMeeting; break;
  case vEvent::S_TENTATIVE: mstate=olMeetingReceived; break;
  case vEvent::S_CONFIRMED: mstate=olMeeting; break;
  case vEvent::S_CANCELLED: mstate=olMeetingCanceled; break;
  default: mstate=olNonMeeting;
  }

  sprintf(sql,"select count(*) from fnbl_pim_calendar where id=%ld",event->GUID());
  if (db->SQL(sql)) {
    long cnt=db->GetLong();
    if (cnt==0) {
      // there might be a bug with PalmOS if subject == ""
      if (strlen(event->Subject())>0) {
	//--- insert database entry -------------------------------------------------------------
	/*
	  char: status, all_day, reminder, rec_no_end_date
	 */
	sprintf(sql,
		"insert into fnbl_pim_calendar ("
		" id,userid,last_update,status,type,sensitivity,"
		" all_day,dstart,dend,duration,"
		" subject,body,categories,location,meeting_status,mileage,busy_status,"
		" reminder_time,reminder,reminder_sound_file,reminder_options,reminder_repeat_count,"
		" rec_type,rec_interval,rec_month_of_year,rec_day_of_month,rec_day_of_week_mask,"
		" rec_instance,rec_start_date_pattern,"
		" rec_no_end_date,rec_end_date_pattern,rec_occurrences,"
		" dstart_tz,dend_tz,reminder_tz,"
		" suid,stouch,folderid"
		") values ("
		"%d,'%s',%u000,'N',1,0,"
		"'%d','%s','%s',0,"  // all_day
		"E'%s',E'%s',E'%s',E'%s',%d,null,%d,"
		"%s,'%d','%s',%d,%d," // reminder
		"%d,%d,"
		"%d,%d,%d,"
		"%d,%s,"      // rec_instance, rec_start_date_pattern
		"'%d',%s,%d," // rec_no_end_date, rec_end_date_pattern, rec_occurrences
		//		"'Europe/Berlin','Europe/Berlin','Europe/Berlin',"
		"'Europe/Berlin','Europe/Berlin','Europe/Berlin',"
		"'%s',1,%d "
		")",
		event->GUID(),uname,
		modified,
		event->AllDay(),dtstart,dtend,
		esc_subject,esc_body,esc_category,esc_location,mstate,bstate,
		almtrigger,event->ALM(),event->ALM_SoundFile(),event->ALM_Options(),event->ALM_Count(),
		event->REC_Type(),event->REC_Interval(),
		event->REC_Month(),event->REC_Day(),event->REC_DAYMASK(),
		event->REC_Instance(),rstart,
		event->REC_NoEndDate(),rend,-1,
		event->UID(),folderid
		);
	if (LOGLEVEL>=4) {
	  LTS();
	  fprintf(LOGFILE,"%s\n",sql);
	  fflush(LOGFILE);
	}
	if (db->SQL(sql)) {
	  ecnt_new_server++;
	  // insert attendees
	  UpdateAttendees(db,uname,event);
	  return event->GUID();
	} else {
	  if (LOGLEVEL>=1) {
	    LTS();
	    fprintf(LOGFILE,"*** insert event: [%s]\n",event->UID());
	    fprintf(LOGFILE,"                          %s\n",db->Error());
	    fprintf(LOGFILE,"                          --------------------------------------\n");
	    fprintf(LOGFILE,"                          USERID: %s\n",uname);
	    fprintf(LOGFILE,"                          LAST-MODIFIED: %u000\n",modified);
	    fprintf(LOGFILE,"                          DTSTART: %s\n",dtstart);
	    fprintf(LOGFILE,"                          DTEND: %s\n",dtend);
	    fprintf(LOGFILE,"                          SUBJECT: %s\n",event->Subject());
	    fprintf(LOGFILE,"                          BODY: %s\n",ESCAPE(event->Body()));
	    fprintf(LOGFILE,"                          CATEGORY: %s\n",ESCAPE(event->Category()));
	    fprintf(LOGFILE,"                          LOCATION: %s\n",event->Location());
	    fprintf(LOGFILE,"                          --------------------------------------\n");

	    fflush(LOGFILE);
	  }
	  return -1;
	}
      } else {
	if (LOGLEVEL>=1) {
	  LTS();
	  fprintf(LOGFILE,"*** insert event: [%s]: event has no subject! \n",event->UID());
	  fflush(LOGFILE);
	}
	return -1;
      }
    } else {
      // there might be a bug with PalmOS if subject == ""
      if (strlen(event->Subject())>0) {
	//--- update database entry -------------------------------------------------------------
	sprintf(sql,
		"update fnbl_pim_calendar set "
		" last_update=%u000,status='U',"
		" all_day=%d,dstart='%s',dend='%s',"
		" subject=E'%s',body=E'%s',categories=E'%s',location=E'%s',"
		" meeting_status=%d,busy_status=%d,"
		" rec_type=%d,rec_interval=%d,"
		" rec_month_of_year=%d,rec_day_of_month=%d,rec_day_of_week_mask=%d,"
		" rec_instance=%d,rec_start_date_pattern=%s,rec_no_end_date=%d,"
		" rec_end_date_pattern=%s,rec_occurrences=%d,stouch=1,"
		" reminder_time=%s,reminder='%d',reminder_sound_file='%s',reminder_options=%d,reminder_repeat_count=%d"
		" where suid='%s' and userid='%s'",
		modified,
		event->AllDay(),dtstart,dtend,
		esc_subject,esc_body,esc_category,esc_location,
		mstate,bstate,
		event->REC_Type(),event->REC_Interval(),
		event->REC_Month(),event->REC_Day(),event->REC_DAYMASK(),
		event->REC_Instance(),rstart,event->REC_NoEndDate(),
		rend,-1,
		almtrigger,event->ALM(),event->ALM_SoundFile(),event->ALM_Options(),event->ALM_Count(),
		event->UID(),uname
		);
	if (LOGLEVEL>=4) {
	  LTS();
	  fprintf(LOGFILE,"%s\n",sql);
	  fflush(LOGFILE);
	}
	if (db->SQL(sql)) {
	  ecnt_update++;
	  UpdateAttendees(db,uname,event);
	  // update attendees
	} else {
	  if (LOGLEVEL>=1) {
	    printf("SQL: %s\n",sql);
	    LTS();
	    fprintf(LOGFILE,"*** update client: [%s]\n",event->UID());
	    fprintf(LOGFILE,"                          %s\n",db->Error());
	    fprintf(LOGFILE,"                          --------------------------------------\n");
	    fprintf(LOGFILE,"                          USERID: %s\n",uname);
	    fprintf(LOGFILE,"                          LAST-MODIFIED: %u000\n",modified);
	    fprintf(LOGFILE,"                          DTSTART: %s\n",dtstart);
	    fprintf(LOGFILE,"                          DTEND: %s\n",dtend);
	    fprintf(LOGFILE,"                          SUBJECT: %s\n",event->Subject());
	    fprintf(LOGFILE,"                          BODY: %s\n",ESCAPE(event->Body()));
	    fprintf(LOGFILE,"                          CATEGORY: %s\n",ESCAPE(event->Category()));
	    fprintf(LOGFILE,"                          LOCATION: %s\n",event->Location());
	    fprintf(LOGFILE,"                          --------------------------------------\n");
	    fflush(LOGFILE);
	  }
	  return -1;
	}
      } else {
	if (LOGLEVEL>=1) {
	  LTS();
	  fprintf(LOGFILE,"*** update client event: [%s]: event has no subject! \n",event->UID());
	  fflush(LOGFILE);
	}
	return -1;
      }
    }
  }
  return -2;
}

// check, if item exists in SyncML and return ID (if exist)
int SyncMLLookupPIMItem(Database *db,const char *pname,const char *uid,const char *uname,unsigned long *id) {
  char sql[4096];

  sprintf(sql,"select count(*) from fnbl_pim_%s where suid='%s' and userid='%s'",pname,uid,uname);
  if (db->SQL(sql)) {
    if (db->GetInt()==1) {
      sprintf(sql,"select id from fnbl_pim_%s where suid='%s' and userid='%s'",pname,uid,uname);
      if (db->SQL(sql)) {
	*id=db->GetUnsignedLong();
	return 1;
      } else {
	if (LOGLEVEL>=1) {
	  LTS();
	  fprintf(LOGFILE,"*** error getting GUID of item %s-%s : %s\n",pname,uid,db->Error());
	  fflush(LOGFILE);
	}
	return -2;
      }
    }
  } else {
    if (LOGLEVEL>=1) {
      LTS();
      fprintf(LOGFILE,"*** error getting information on item %s-%s : %s\n",pname,uid,db->Error());
      fflush(LOGFILE);
    }
  }

  return -3;
}

// create new ID
long int SyncMLGenPIMItem(Database *db,const char *pname,const char *uid,const char *uname,unsigned long *nid) {
  char sql[4096];

  sprintf(sql,"select count(*) from fnbl_pim_%s where suid='%s' and userid='%s'",pname,uid,uname);
  if (db->SQL(sql)) {
    if (db->GetInt()==0) {
      sprintf(sql,"select counter,increment_by from fnbl_id where idspace='pim.id'");
      if (db->SQL(sql)) {
	long cid=db->GetLong(0);
	long inc=db->GetLong(1);
	*nid=cid+inc;
	sprintf(sql,"update fnbl_id set counter=%u where idspace='pim.id'",*nid);
	if (db->SQL(sql)) {
	  return 1;
	} else {
	  if (LOGLEVEL>=1) {
	    LTS();
	    fprintf(LOGFILE,"*** error updating GUID for namespace pim.id: %s\n",db->Error());
	    fflush(LOGFILE);
	  }
	  return -1;
	}
      } else {
	if (LOGLEVEL>=1) {
	  LTS();
	  fprintf(LOGFILE,"*** error getting GUID for namespace pim.id: %s\n",db->Error());
	  fflush(LOGFILE);
	}
	return -2;
      }
    }
  } else {
    if (LOGLEVEL>=1) {
      LTS();
      fprintf(LOGFILE,"*** error getting information on item %s-%s : %s\n",pname,uid,db->Error());
      fflush(LOGFILE);
    }
  }

  return -3;
}

long int SyncCalendarItem(IMAP *imap,char *folder,long mid,Database *db,char *uname,vEvent *event,int direction=1,const char stype[]="IMAP",
			  unsigned int folderid=0) {
  char sql[2048];

  // check, if item exists in SyncML
  // or return new item id

  if (LOGLEVEL>=2) {
    LTS();
    fprintf(LOGFILE,"sync %s:[%s]\n",uname,event->UID());
    fflush(LOGFILE);
  }
  unsigned long nid;
  int r=SyncMLLookupPIMItem(db,"calendar",event->UID(),uname,&nid);

  if (r>0) {
    // set SyncML ID
    event->GUID(nid);
    //--- this is an existing event --------------------------------------------------------
    //--- load item from DB
    vEvent *sevent=EventFromDB(db,nid);

    if (sevent!=NULL) { 
      unsigned long eid=sevent->GUID();
      Date d;

      time_t tsc;
      time_t tss;
      tss=time(NULL);
      tsc=tss;

      d=sevent->LastModified();
      if (d!=NoDate()) tsc= d.Time();                        // last change on client
      d=event->LastModified();
      if (d!=NoDate())tss= d.Time();                         // last change on server

      // forget about the bias
      tss=(time_t)(tss/BIAS)*BIAS;
      tsc=(time_t)(tsc/BIAS)*BIAS;

      // everything within BIAS range: client wins
      // check if there is a change anyway
      int echange=0;
      echange=event->Compare(sevent);

      int synced=0;
      // check, if item was almost synced to device (there must be a client mapping
      sprintf(sql,"select count(*) from fnbl_client_mapping where sync_source='cal' and guid='%ld'",event->GUID());
      if (db->SQL(sql)) {
	synced=db->GetInt(0);
      } else {
	LTS();
	fprintf(LOGFILE,"*** while checking, if event was almost synced: %s\n",db->Error());
	fflush(LOGFILE);
	synced=0;
      }

      if (LOGLEVEL>=4) {
	LTS();
	fprintf(LOGFILE,"synced: %d, last change server (%ld), client (%ld), diff (%ld)\n",synced,tss,tsc,tss-tsc);
	fflush(LOGFILE);
      }

      if (echange>0) {
	if (LOGLEVEL>=3) {
	  LTS();
	  fprintf(LOGFILE,"event update reason: %d\n",echange);
	  fflush(LOGFILE);
	  if (echange==1) {
	    LTS();
	    fprintf(LOGFILE,"SyncML Client DTSTART=%s - Calendar Server DTSTART=%s\n",sevent->DTStart().ToChar(),event->DTStart().ToChar());
	    fflush(LOGFILE);
	  }
	}
	if ((tss>tsc) || (!synced)) {
	  // last change was performed on server
	
	  // fprintf(smsf,"U:%s:%s\n",Time2ISO(event->DTStart()),event->Subject());
	  unsigned long r=0;

	  if (direction & 1)
	    r=EventToDB(db,uname,event,folderid);
	  else {
	    if (LOGLEVEL>=3) {
	      LTS();
	      fprintf(LOGFILE,"ACTION: would have send %s to client\n",event->UID());
	      fflush(LOGFILE);
	    }
	  }

	  if (sevent)
	    delete sevent;
	  else {
	    LTS();
	    fprintf(LOGFILE,"*** attempt to delete NULL pointer sevent at syncd.c:SyncCalendarItem.1\n");
	    fflush(LOGFILE);
	  }
	  return r;
	} else {
	  // last change was performed on client
	  sprintf(sql,"select status from fnbl_pim_calendar where suid='%s' and userid='%s'",event->UID(),uname);
	  if (db->SQL(sql)) {
	    char st[255];
	    strcpy(st,db->GetString(0));
	    //--- updated on device ----------------------------------------------------------------
	    if (!strcmp(st,"U")) {
	      if (LOGLEVEL>=3) {
		LTS();
		Date d=event->DTStart();
		fprintf(LOGFILE,"update on server: [%s] %s %s\n",event->UID(),d.ISOChar(),event->Subject());
		fflush(LOGFILE);
	      }
	      char emsg[2048];
	      int r;

	      if (imap) {
		if (direction & 2)
		  r=imap->Delete(mid,emsg);
		else {
		  strcpy(emsg,"OK");
		  if (LOGLEVEL>=3) {
		    LTS();
		    fprintf(LOGFILE,"ACTION: would have deleted %ul from server before update on server\n",mid);
		    fflush(LOGFILE);
		  }
		}
	      } else {
		strcpy(emsg,"OK");
		if (LOGLEVEL>=3) {
		  LTS();
		  fprintf(LOGFILE,"ACTION: update on server not supported for HTTP/ICS\n");
		  fflush(LOGFILE);
		}
	      }

	      if (strstr(emsg,"OK")==emsg) { 
		// write item again to IMAP server
		vCalendar *c=new vCalendar();
		c->AddEvent(sevent);
		
		char *sbuffer=NULL;
		c->Write(&sbuffer);
		char fname[1024];
		
		if (LOGLEVEL>=1) {
		  LTS();
		  fprintf(LOGFILE,"--- update event ------------\n");
		  fprintf(LOGFILE,"%s\n",sbuffer);
		  fprintf(LOGFILE,"-----------------------------\n");
		}
		
		sprintf(fname,"~%s/%s",uname,folder);

		if (imap) {
		  if (direction & 2) {
		    r=imap->Send(fname,sevent->Subject(),sevent->UID(),IMAP::CTYPE_VCALENDAR,sbuffer);
		  } else {
		    if (LOGLEVEL>=3) {
		      LTS();
		      fprintf(LOGFILE,"ACTION: would have send %s from Server\n",sevent->UID());
		      fflush(LOGFILE);
		    }
		  }
		} else {
		  // if (LOGLEVEL>=3) {
		  // LTS();
		  // fprintf(LOGFILE,"ACTION: update on server not supported for HTTP/ICS\n");
		  // fflush(LOGFILE);
		  // }
		}

		sprintf(sql,"update fnbl_pim_calendar set stouch=1 where suid='%s' and userid='%s'",event->UID(),uname);
		
		if (!db->SQL(sql)) {
		  if (LOGLEVEL>=1) {
		    LTS();
		    fprintf(LOGFILE,"uppdating suid [%s] for event %d : %s\n",event->UID(),eid,db->Error());
		    fflush(LOGFILE);
		  }
		}
		
		if (c)
		  delete c; // implicitely deletes sevent (as sevent is child of c)
		else {
		  LTS();
		  fprintf(LOGFILE,"*** attempt to delete NULL pointer c at syncd.c:SyncCalendarItem.3\n");
		  fflush(LOGFILE);
		}
		
		if (imap) {
		  if (direction & 2)
		    ecnt_update_server++;
		}
	      } else {
		if (LOGLEVEL>=1) {
		  LTS();
		  Date d=event->DTStart();
		  fprintf(LOGFILE,"*** %s: delete from server for update: [%s] %s %s\n",
			  emsg,event->UID(),d.ISOChar(),event->Subject());
		  fflush(LOGFILE);
		}
		
		// event handled anyway
		sprintf(sql,"update fnbl_pim_calendar set stouch=1 where suid='%s' and userid='%s'",event->UID(),uname);
		
		if (!db->SQL(sql)) {
		  if (LOGLEVEL>=1) {
		    LTS();
		    fprintf(LOGFILE,"uppdating suid [%s] for event %d : %s\n",event->UID(),eid,db->Error());
		    fflush(LOGFILE);
		  }
		}
		
		if (sevent)
		  delete sevent;
		else {
		  LTS();
		  fprintf(LOGFILE,"*** attempt to delete NULL pointer sevent at syncd.c:SyncCalendarItem.1\n");
		  fflush(LOGFILE);
		}
	      }
	      
	      return event->GUID();
	    }

	    //--- event was delete on device -------------------------------------------------------
	    if (!strcmp(st,"D")) {
	      if (LOGLEVEL>=2) {
		LTS();
		Date d=event->DTStart();
		fprintf(LOGFILE,"deleted on client -> delete from server: [%s] %s %s\n",
			event->UID(),d.ISOChar(),event->Subject());
		fflush(LOGFILE);
	      }
	      // IMAP commands:
	      char emsg[2048];
	      int r;

	      if (imap) {
		if (direction & 2) {
		  r=imap->Delete(mid,emsg);
		  ecnt_delete_server++;
		} else {
		  if (LOGLEVEL>=3) {
		    LTS();
		    fprintf(LOGFILE,"ACTION: would have deleted %s from Server\n",mid);
		    fflush(LOGFILE);
		  }
		}
	      } else {
		if (LOGLEVEL>=3) {
		  LTS();
		  fprintf(LOGFILE,"ACTION: delete on server not supported for HTTP/ICS\n");
		  fflush(LOGFILE);
		}
	      }

	      sprintf(sql,"update fnbl_pim_calendar set stouch=1 where suid='%s' and userid='%s'",event->UID(),uname);
	      db->SQL(sql);
	      
	      if (sevent)
		delete sevent;
	      else {
		LTS();
		fprintf(LOGFILE,"*** attempt to delete NULL pointer sevent at syncd.c:SyncCalendarItem.4\n");
		fflush(LOGFILE);
	      }
	      return event->GUID();
	    }
	  } else {
	    if (LOGLEVEL>=1) {
	      LTS();
	      fprintf(LOGFILE,"*** update status: [%s] %s\n",event->UID(),db->Error());
	      fflush(LOGFILE);
	    }
	  }
	}
      } else {
	if (LOGLEVEL>=3) {
	  LTS();
	  fprintf(LOGFILE,"event not really changed [%s]\n",event->UID());
	  fflush(LOGFILE);
	}
	sprintf(sql,"update fnbl_pim_calendar set stouch=1 where suid='%s' and userid='%s'",event->UID(),uname);
	db->SQL(sql);
	  
	if (sevent)
	  delete sevent;
	else {
	  LTS();
	  fprintf(LOGFILE,"*** attempt to delete NULL pointer sevent at syncd.c:SyncCalendarItem.2\n");
	  fflush(LOGFILE);
	}
	return event->GUID();
      }

      if (sevent)
	delete sevent;
      else {
	LTS();
	fprintf(LOGFILE,"*** attempt to delete NULL pointer sevent at syncd.c:SyncCalendarItem.5\n");
	fflush(LOGFILE);
      }

      return event->GUID();
    }
  } else {
    //--- the item does not exist in the syncml database, make a new entry -----------------------------
    unsigned long nid;
    int r;

    if (direction & 1)
      r=SyncMLGenPIMItem(db,"calendar",event->UID(),uname,&nid);
    else {
      r=0;
      if (LOGLEVEL>=3) {
	LTS();
	fprintf(LOGFILE,"ACTION: would have send %s to client\n",event->UID());
	fflush(LOGFILE);
      }
    }

    if (r>0) {
      if (LOGLEVEL>=3) {
	LTS();
	fprintf(LOGFILE,"    this is a new event (on the imap-server):%ld\n",nid);
	fflush(LOGFILE);
      }
      event->GUID(nid);
      return EventToDB(db,uname,event,folderid);
    } else {
      if (LOGLEVEL>=1) {
	LTS();
	fprintf(LOGFILE,"*** could not create entry in syncml db for new event from imap:%d\n",r);
	fflush(LOGFILE);
      }
    }
  }
  
  return nid;
}

int SyncCardItem(IMAP *imap,Database *db,vCard *card) {
}

int SyncIMAP(char *imaphost,char *imapuser,char *user,char *cname,int folderid,int direction,Database *db) {
  char pb[1024];
  unsigned int nmsg=0;

  IMAP *imap=new IMAP(imaphost,imapuser,"sync4j");
  sprintf(pb,"~%s/%s",user,cname);
	    
  nmsg=imap->Select(pb);
  {
    // delete marked items
    char emsg[1024];
    int r=imap->Expunge(emsg);
  }
	    
  nmsg=imap->Select(pb);
  if (LOGLEVEL>=1) {
    LTS();
    fprintf(LOGFILE,"IMAP-Folder ~%s/%s contains %d entries\n",user,cname,nmsg);
    fflush(LOGFILE);
  }
	    
  char header[BSIZE];
  char data[BSIZE];
  char emsg1[BSIZE];
  char emsg2[BSIZE];
  char boundary[BSIZE];
  
  //--- scan IMAP folder and process items ---------------------------
  for (int j=1;j<=nmsg;j++) {
    int r1=imap->Fetch(j,"HEADER",header,emsg1);
    int r2=imap->Fetch(j,"TEXT",data,emsg2);
    
    if (strstr(emsg2,"OK completed")==emsg2) {
      if (LOGLEVEL>=4) {
	LTS();
	fprintf(LOGFILE,header);
	fprintf(LOGFILE,data);
	fflush(LOGFILE);
      }
      
      IMAPItem *item=new IMAPItem(header,data);
      
      if ((strstr(data,"BEGIN:VCALENDAR")) && (strstr(data,"END:VCALENDAR"))) {
	for (int ii=0;ii<item->NumParts();ii++) {
	  //-------------------------------------------------------------------------------------
	  if (strstr(item->Part(ii),"BEGIN:VCALENDAR")) {
	    char *aptr=strstr(item->Part(ii),"BEGIN:VCALENDAR");
	    char *eptr=strstr(aptr,"END:VCALENDAR");
	    unsigned long eblength=eptr-aptr+strlen("END:VCALENDAR");
	    char *ebuffer=new char [eblength+1];
	    memset(ebuffer,0,eblength+1);
	    strncpy(ebuffer,aptr,eblength);
	    
	    vCalendar *v=ParseCalendar(ebuffer);
	    
	    if ((v) && (v->NumEvents()==1)) {
	      // store attachements to file-system
	      char abdname[8192];
	      if (GetValue("Application","Attachements"))
		strcpy(abdname,GetValue("Application","Attachements"));
	      else
		strcpy(abdname,".");
	      char abname[8192];
	      sprintf(abname,"%s/%s.attachament.",abdname,v->Event(0)->UID());
	      char cmd[8192];
	      sprintf(cmd,"rm -f %s*",abname);
	      system(cmd);
	      for (int ia=0;ia<item->NumParts();ia++) {
		if (ia!=ii) {
		  char aname[8192];
		  sprintf(aname,"%s%d",abname,ia);
		  item->WriteToFile(ia,aname);
		}
	      }
	      // sync item
	      long int r;
	      r=SyncCalendarItem(imap,cname,j,db,user,v->Event(0),direction,"IMAP",folderid);

	      //delete v;

	    } else {
	      if (LOGLEVEL>=1) {
		LTS();
		fprintf(LOGFILE,"parser failed on\n%s\n",ebuffer);
		fflush(LOGFILE);
	      }
	      
	      // what to do, if parser failed
	      char puid[2048];
	      strcpy(puid,ParsedUID());
	      if (strlen(puid)>0) {
		// but we found an ID, touch the appropriate record with stouch=-1
		// will indicate that the item is handled and should not be deleted!
		char sql[2048];
		sprintf(sql,"update fnbl_pim_calendar set stouch=-1 where suid='%s' and userid='%s'",puid,user);
		if (!db->SQL(sql)) {
		  if (LOGLEVEL>=1) {
		    LTS();
		    fprintf(LOGFILE,"failed to mark [%s] as handled\n   error is: %s\n",puid,db->Error());
		    fflush(LOGFILE);
		  }
		}
	      }
	    }
	    
	    delete ebuffer;
	  }
	}
      }
      
      if (strstr(data,"BEGIN:VCARD")) {
	for (int ii=0;ii<item->NumParts();ii++) {
	  //-------------------------------------------------------------------------------------
	  if (strstr(item->Part(ii),"BEGIN:VCARD")) {
	  }
	}
      }
      
      if (item) delete item;
    } // end of "OK completed"
    
  } // next message
  
  delete imap;

  return 1;
}

int SyncIMAP_NEW(char *imaphost,char *imapuser,char *user,char *cname,int folderid,int direction,
		 Database *db,Database *db1,Database *db2) {
  strcpy(imapuser,GetValue("IMAP","User"));
  IMAP *imap=new IMAP(imaphost,imapuser,"sync4j");
  char sql[2048];

  sprintf(sql,"select count(*) from fnbl_pim_calendar where userid='%s' and status='N' and suid is null",user);
  if (db->SQL(sql)) {
    int ne=db->GetInt();
    if (ne>0) {
      sprintf(sql,"select id from fnbl_pim_calendar where userid='%s' and status='N' and suid is null",user);
      if (db->SQL(sql)) {
	do {
	  long id;
	  id =db->GetLong(0);
	  if (LOGLEVEL>=3) {
	    LTS();
	    fprintf(LOGFILE,"new event from client: %ld\n",id);
	    fflush(LOGFILE);
	  }
	  
	  vEvent *event=EventFromDB(db2,id);
	  vCalendar *c=new vCalendar();
	  
	  time_t t;
	  char suid[256];
	  sprintf(suid,"%d.%s@funambol.%s",id,user,"fzd.de");
	  event->UID(suid);
	  c->AddEvent(event);
	  
	  char *sbuffer=NULL;
	  c->Write(&sbuffer);
	  char fname[1024];
	  sprintf(fname,"~%s/%s",user,cname);
	  int r;
	  
	  if (direction & 2) {
	    r=imap->Send(fname,event->Subject(),event->UID(),IMAP::CTYPE_VCALENDAR,sbuffer);
	  } else {
	    if (LOGLEVEL>=3) {
	      LTS();
	      fprintf(LOGFILE,"ACTION: would have send %s to server\n",event->UID());
	      fflush(LOGFILE);
	    }
	  }
	  
	  sprintf(sql,"update fnbl_pim_calendar set suid='%s',stouch=1,folderid=%d where id=%d",event->UID(),folderid,id);
	  if (!db1->SQL(sql)) {
	    if (LOGLEVEL>=1) {
	      LTS();
	      fprintf(LOGFILE,"uppdating suid [%s] for event %d : %s\n",event->UID(),id,db1->Error());
	      fflush(LOGFILE);
	    }
	  }
	  ecnt_new_client++;
	  
	  if (c)
	    delete c;
	  else {
	    LTS();
	    fprintf(LOGFILE,"*** attempt to delete NULL pointer c at syncd.c:main.4\n");
	    fflush(LOGFILE);
	  }
	  
	} while (db->NextRow());
      } else {
	if (LOGLEVEL>=1) {
	  LTS();
	  fprintf(LOGFILE,"getting data for new events: [%s] %s\n",user,db->Error());
	  fflush(LOGFILE);
	}
      }
    }
  } else {
    if (LOGLEVEL>=1) {
      LTS();
      fprintf(LOGFILE,"check new events: [%s] %s\n",user,db->Error());
      fflush(LOGFILE);
    }
  }

  delete imap;

  return 1;
}

int SyncICS_HTTP(char *imapuser,char *user,char *cname,int folderid,int direction,Database *db) {
  // download the full file via wget http://<host>/fname
  char dpath[4096];
  char cfname[4096];
  char cmd[4096];
  if (GetValue("Application","ICS"))
    strcpy(dpath,GetValue("Application","ICS"));
  else
    strcpy(dpath,".");
  
  // wget -q "URL" -O <...ics>
  
  sprintf(cfname,"%s/%d.ics",dpath,folderid);
  sprintf(cmd,"wget -q \"%s\" -O \"%s\"",cname,cfname);
  
  int r=system(cmd);
	    
  if (r==0) {
    // load the local file
    unsigned long cfs=FILESIZE(cfname);
    FILE* f=fopen(cfname,"r");
    if (f) {
      char *ebuffer=new char [cfs+1];
      memset(ebuffer,0,cfs+1);
      if (fread(ebuffer,1,cfs,f)==cfs) {
	printf("parsing\n-----------------------\n%s\n--------------------------\n",ebuffer);
	
	vCalendar *v=ParseCalendar(ebuffer);
	if ((v) && (v->NumEvents()>0)) {
	  printf("loaded %d events\n",v->NumEvents());
	  
	  long int rr;
	  for (unsigned int i=0;i<v->NumEvents();i++)
	    rr=SyncCalendarItem(NULL,cname,0,db,user,v->Event(i),direction,"HTTP/ICS",folderid);
	}
	// delete v;
      } else {
	if (LOGLEVEL>=1) {
	  LTS();
	  fprintf(LOGFILE,"failed to open load cached data for [%s] from [%s]\n",cname,cfname);
	  fflush(LOGFILE);
	}
      }
      delete ebuffer;
      fclose(f);
    } else {
      if (LOGLEVEL>=1) {
	LTS();
	fprintf(LOGFILE,"failed to open download [%s] for [%s]\n",cfname,cname);
	fflush(LOGFILE);
      }
    }
  } else {
    if (LOGLEVEL>=1) {
      LTS();
      fprintf(LOGFILE,"failed to download [%s], error code is %d\n",cname,r);
      fflush(LOGFILE);
    }
  }
}

int main(int argc,char* argv[]) {
  //  if (argc>=3) {
  ParseCommandline(argc,argv);

  if (!cfg())
    cfg("imap2syncml.ini");

  if (GetValue("Logging","Logfile")) {
    if (!strcmp(GetValue("Logging","Logfile"),"Console")) {
      LOGFILE=stdout;
    } else {
      LOGFILE=fopen(GetValue("Logging","Logfile"),"a");
    }
  } else
    LOGFILE=fopen("/var/log/imap2syncml.log","a");

  if (GetValue("Application","LOCALTZ")) {
    sscanf(GetValue("Application","LOCALTZ"),"%d",&LOCALTZ);
  } else
    LOCALTZ=0;

  
  LTS();
  fprintf(LOGFILE,"*** START SYNC\n");
  fflush(LOGFILE);
  
  FILE *f;
  
  if (GetValue("Application","Lockfile"))
    f=fopen(GetValue("Application","Lockfile"),"r");
  else
    f=fopen("/tmp/imap2syncml.lck","r");
  
  int ipid;
  if (f) {
    fscanf(f,"%d",&ipid);
    LTS();
    fprintf(LOGFILE,"imap2syncml.lck claims imap2syncml is still running with pid %d\n",ipid);
    fflush(LOGFILE);
    fclose(f);
    
    // check, if it really hangs ...
    char pfn[1024];
    sprintf(pfn,"/proc/%d/status",ipid);
    FILE *pf=fopen(pfn,"r");
    if (pf) {
      fclose(pf);
      LTS();
      fprintf(LOGFILE,"  /proc says yes, it is really running ...\n");
      fflush(LOGFILE);
      
      char cmd[1024];
      sprintf(cmd,"echo \"SyncML: syncd hangs\" | mail n.schmeisser@fzd.de");
      int r=system(cmd);
      
      LTS();
      fprintf(LOGFILE,"*** TERMINATE SYNC\n");
      fflush(LOGFILE);
      
      fclose(LOGFILE);
      
      return -1;
    } else {
      LTS();
      fprintf(LOGFILE,"  but /proc says no, it is not running ...\n");
      fflush(LOGFILE);
      unlink("/tmp/imap2syncml.lck");
      // continue normal operation
    }
  }
  
  if (GetValue("Logging","Loglevel"))
    sscanf(GetValue("Logging","Loglevel"),"%d",&LOGLEVEL);
  else
    LOGLEVEL=3;
  
  f=fopen("/tmp/imap2syncml.lck","w");
  ipid=getpid();
  time_t t=time(NULL);
  fprintf(f,"%d:%s\n",ipid,ctime(&t));
  fclose(f);
    
  IMAP *imap;
  
  char imaphost[2048];
  char imapuser[2048];
  
  if (GetValue("IMAP","Host") && GetValue("IMAP","User")) {
    strcpy(imaphost,GetValue("IMAP","Host"));
    strcpy(imapuser,GetValue("IMAP","User"));
    // imap=new IMAP(imaphost,imapuser,"sync4j");
  } else {
    LTS();
    fprintf(LOGFILE,"*** No IMAP connection specified, please set [IMAP] Host and User\n");
    LTS();
    fprintf(LOGFILE,"*** TERMINATE SYNC\n");
    fclose(LOGFILE);
    unlink("/tmp/imap2syncml.lck");
    return -1;
  }

  db=new Database();
  Database *db1=new Database();
  Database *db2=new Database();
  Database *db3=new Database();
  
  char dbhost[2048];
  char dbschema[2048];
  char dbuser[2048];
  
  if (GetValue("Database","Host") && GetValue("Database","Schema") && GetValue("Database","User")) {
    strcpy(dbhost,GetValue("Database","Host"));
    strcpy(dbschema,GetValue("Database","Schema"));
    strcpy(dbuser,GetValue("Database","User"));
  } else {
    LTS();
    fprintf(LOGFILE,"*** No database connection specified, please set [Database] Host, Schema and User\n");
    LTS();
    fprintf(LOGFILE,"*** TERMINATE SYNC\n");
    fclose(LOGFILE);
    unlink("/tmp/imap2syncml.lck");
    return -1;
  }

  db->Connect(dbhost,dbschema,dbuser,"funambol");
  db1->Connect(dbhost,dbschema,dbuser,"funambol");
  db2->Connect(dbhost,dbschema,dbuser,"funambol");
  db3->Connect(dbhost,dbschema,dbuser,"funambol");

  // db->DebugLevel(LOGLEVEL);
  
  if (db->SQL("select count(*) from fnbl_user where username<>'admin' and username<>'guest' and dosync=1")) {
    int nuser=db->GetInt();
    char **users=new char* [nuser];
    
    if (db->SQL("select username from fnbl_user where dosync=1 order by username")) {
      for (int i=0;i<nuser;i++) {
	users[i]=new char [255];
	strcpy(users[i],db->GetString());
	db->NextRow();
      }
    }
    
    if (GetValue("Application","BIAS"))
      sscanf(GetValue("Application","BIAS"),"%ld",&BIAS);
    else
      BIAS=600;

    for (int i=0;i<nuser;i++) {
      //--- check events on server ----------------------------------------------
      // untouched events are deleted from server !!!
      time_t t1,t2;
      
      char sql[4096];
      char emsg[4096];
      int direction=1;

      ecnt_new_server=0;
      ecnt_new_client=0;
      ecnt_update=0;
      ecnt_update_server=0;
      ecnt_delete=0;
      ecnt_delete_server=0;
      
      char cname[1024];
      char ctype[1024];
      char ftype[1024];
      int folderid;
      int isprimaryfolder;

      sprintf(sql,
	      "select fname,direction,type,host,id,isprimaryfolder,ftype from fnbl_user_folder "
	      "where direction>0 and username='%s' order by id",
	      users[i]);

      if (!(db3->SQL(sql))) {
	if (LOGLEVEL>=0) {
	  LTS();
	  fprintf(LOGFILE,"Fatal error: could not get data to access user folder for sync for user %s\n",users[i]);
	  fflush(LOGFILE);
	}
	exit(-1);
      }

      // loop through all folders
      do {
	
	strcpy(cname,db3->GetString(0));
	direction=db3->GetInt(1);
	strcpy(ctype,db3->GetString(2));
	strcpy(imaphost,db3->GetString(3));
	folderid=db3->GetInt(4);
	isprimaryfolder=db3->GetInt(5);
	strcpy(ftype,db3->GetString(6));
	strcpy(imapuser,GetValue("IMAP","User"));
	
	if (!strcmp(ftype,"events")) {
	  sprintf(sql,"update fnbl_pim_calendar set stouch=0 where userid='%s' and folderid=%d",users[i],folderid);
	
	  if (db->SQL(sql)) {
	    
	    time(&t1);
	    
	    if (!strcmp(ctype,"IMAP")) {
	      SyncIMAP(imaphost,imapuser,users[i],cname,folderid,direction,db);
	      SyncIMAP_NEW(imaphost,imapuser,users[i],cname,folderid,direction,db,db1,db2);
	    }
	    
	    if (!strcmp(ctype,"HTTP/ICS")) {
	      SyncICS_HTTP(imapuser,users[i],cname,folderid,direction,db);
	    } 
	    
	    //------------------------------------------------------------------
	    //--- find deleted items (items that have a suid and are not touched ----
	    sprintf(sql,"select count(*) from fnbl_pim_calendar where userid='%s' and suid is not null and stouch=0 and folderid=%d",
		    users[i],folderid);
	    if (db->SQL(sql)) {
	      ecnt_delete=db->GetInt();
	      if (ecnt_delete>0) {
		sprintf(sql,"select id from fnbl_pim_calendar where userid='%s' and suid is not null and stouch=0 and folderid=%d",
			users[i],folderid);
		if (db->SQL(sql)) {
		  do {
		    long id=db->GetLong(0);
		    vEvent *event=EventFromDB(db2,id);
		    // fprintf(smsf,"D:%s:%s\n",Time2ISO(event->DTStart()),event->Subject());
		    if (event)
		      delete event;
		    else {
		      LTS();
		      fprintf(LOGFILE,"*** attempt to delete NULL pointer event at syncd.c:main.3\n");
		      fflush(LOGFILE);
		    }
		  } while (db->NextRow());
		}
		
		sprintf(sql,"update fnbl_pim_calendar set last_update=%ld000,status='D',stouch=1 "
			"where userid='%s' and stouch=0 and folderid=%d",time(NULL),users[i],folderid);
		if (!(db->SQL(sql))) {
		  if (LOGLEVEL>=3) {
		    LTS();
		    fprintf(LOGFILE,"*** marking events for deletion: %s\n",emsg);
		    fflush(LOGFILE);
		  }
		}
	      }
	    }
	    
	  } else {
	    if (LOGLEVEL>=1) {
	      LTS();
	      fprintf(LOGFILE,"*** update event list failed with: %s\n",emsg);
	      fflush(LOGFILE);
	    }
	  }
	  
        } //--- end of calendar sync
	
	if ((!strcmp(ftype,"contacts")) && (!strcmp(ctype,"IMAP"))) {
	  IMAP *imap=new IMAP(imaphost,imapuser,"sync4j");
	  //--- check for new contacts on client ----------------------------------------------
	  if (isprimaryfolder) {
	    sprintf(sql,"select count(*) from fnbl_pim_contact where userid='%s' and status='N' and suid is null",users[i]);
	    if (db->SQL(sql)) {
	      int ne=db->GetInt();
	      if (ne>0) {
		sprintf(sql,"select id from fnbl_pim_contact where userid='%s' and status='N' and suid is null",users[i]);
		if (db->SQL(sql)) {
		  do {
		    long id;
		    id =db->GetLong(0);
		    if (LOGLEVEL>=3) {
		      LTS();
		      fprintf(LOGFILE,"new contact from client: %ld\n",id);
		      fflush(LOGFILE);
		    }
		    
		    vCard *card=ContactFromDB(db2,id);
		  
		    time_t t;
		    char suid[256];
		    sprintf(suid,"%d.%s@funambol.%s",id,users[i],"fzd.de");
		    card->UID(suid);
		    
		    char *sbuffer=NULL;
		    card->Write(&sbuffer);
		    char fname[1024];
		    sprintf(fname,"~%s/Kontakte",users[i]);
		    int r;
		    
		    if (direction & 2) {
		      r=imap->Send(fname,card->Subject(),card->UID(),IMAP::CTYPE_VCARD,sbuffer);
		    } else {
		      if (LOGLEVEL>=3) {
			LTS();
			fprintf(LOGFILE,"ACTION: would have send card %s to server\n",card->UID());
			fflush(LOGFILE);
		      }
		    }
		    
		    sprintf(sql,"update fnbl_pim_contact set suid='%s',stouch=1,folderid=%d where id=%ld",card->UID(),folderid,id);
		    if (!db1->SQL(sql)) {
		      if (LOGLEVEL>=1) {
			LTS();
			fprintf(LOGFILE,"uppdating suid [%s] for contact %d : %s\n",card->UID(),id,db1->Error());
			fflush(LOGFILE);
		      }
		    }
		    
		    ecnt_new_client++;
		    
		    if (card)
		      delete card;
		    else {
		      LTS();
		      fprintf(LOGFILE,"*** attempt to delete NULL pointer card at syncd.c:main.5\n");
		      fflush(LOGFILE);
		    }
		  
		  } while (db->NextRow());
		} else {
		  if (LOGLEVEL>=1) {
		    LTS();
		    fprintf(LOGFILE,"getting data for new contacts: [%s] %s\n",users[i],db->Error());
		    fflush(LOGFILE);
		  }
		}
	      }
	    } else {
	      if (LOGLEVEL>=1) {
	      LTS();
	      fprintf(LOGFILE,"check new contacts: [%s] %s\n",users[i],db->Error());
	      fflush(LOGFILE);
	      }
	    }
	  } //--- end of check for new contacts

	  if (imap) delete imap;

	} //--- end of contact sync
	
	
      } while (db3->NextRow());

      time(&t2);
      if (LOGLEVEL>=1) {
	LTS();
	fprintf(LOGFILE,"New events from server: %d\n",ecnt_new_server);
	LTS();
	fprintf(LOGFILE,"New events from client: %d\n",ecnt_new_client);
	LTS();
	fprintf(LOGFILE,"Updated events from server: %d\n",ecnt_update);
	LTS();
	fprintf(LOGFILE,"Updated events from client: %d\n",ecnt_update_server);
	LTS();
	fprintf(LOGFILE,"Deleted events: %d\n",ecnt_delete);
	LTS();
	fprintf(LOGFILE,"Elapsed time: %d s\n",(t2-t1));
	fflush(LOGFILE);
      }
      
      int new_mail=0;
      
      // fclose(smsf);
      
      // there is a bug with ecnt_delete
      ecnt_delete=0;
      
      // submit SMS to device (if it is a phone)
      if ((ecnt_new_server>0) || (ecnt_update>0) || (ecnt_delete>0) || (new_mail>0)) {
	sprintf(sql,"select device from fnbl_principal where username='%s'",users[i]);
	if (db->SQL(sql)) {
	  char did[256];
	  strcpy(did,db->GetString());
	  sprintf(sql,"select msisdn,push from fnbl_device where id='%s' and push>0",did);
	  if (db->SQL(sql)) {
	    char msisdn[256];
	    int push;
	    char *s=db->GetString(0);
	    if (s!=NULL)
	      strcpy(msisdn,s);
	    else
	      strcpy(msisdn,"");
	    push=db->GetInt(1);
	    if ((strlen(msisdn)>0) && ((push>=1) || ((new_mail>0) && (push>=2)))) {
	      char cmd[4096];
	      char rcpt[4096];
	      
	      if (GetValue("SMS","Gateway") && GetValue("SMS","Prefix")) {
		char smsgateway[2048];
		char smsprefix[2048];
		
		strcpy(smsgateway,GetValue("SMS","Gateway"));
		strcpy(smsprefix,GetValue("SMS","Prefix"));
		
		sprintf(rcpt,smsprefix,msisdn);
		strcat(rcpt,"@");
		strcat(rcpt,smsgateway);
		
		if (new_mail) {
		  sprintf(cmd,"echo \"SyncML: %d new events, %d updated events, %d deleted events, unseen e-mail\" "
			  "| mail %s",ecnt_new_server,ecnt_update,ecnt_delete,rcpt);
		} else {
		  sprintf(cmd,"echo \"SyncML: %d new events, %d updated events, %d deleted events\" "
				"| mail %s",ecnt_new_server,ecnt_update,ecnt_delete,rcpt);
		}
		int r=system(cmd);
		if (LOGLEVEL>=1) {
		  LTS();
		  fprintf(LOGFILE,"sent SMS notification to %s, status=%d\n",rcpt,r);
		  fflush(LOGFILE);
		}
	      } else {
		if (LOGLEVEL>=1) {
		  LTS();
		  fprintf(LOGFILE,"SMS push requested for %s, but no SMS data ([SMS] Gateway and Prefix) are specified\n",msisdn);
		  fflush(LOGFILE);
		}
	      }
	    }
	  }
	}
      }
    }
      
    if (users) {
      for (int i=0;i<nuser;i++) {
	if (users[i])
	  delete users[i];
	else {
	  LTS();
	  fprintf(LOGFILE,"*** attempt to delete NULL pointer users[%d] at syncd.c:main.6\n");
	  fflush(LOGFILE);
	}
      }
      
      delete users;
    } else {
      LTS();
      fprintf(LOGFILE,"*** attempt to delete NULL pointer users at syncd.c:main.6\n");
      fflush(LOGFILE);
    }
  }
    
  db->Disconnect();
  db1->Disconnect();
  
  unlink("/tmp/imap2syncml.lck");
  
  LTS();
  fprintf(LOGFILE,"syncd finished successfully\n");
  
  LTS();
  fprintf(LOGFILE,"*** STOP SYNC\n");
  fflush(LOGFILE);
  
  fclose(LOGFILE);
  
  return 0;
}

