#include <string.h>
#include <stdio.h>

#include <timetools.h>

int IsDT(char *d) {
  int yy,mm,dd,hh,mi,ss;
  if (sscanf(d,"%04d%02d%02dT%02d%02d%02d",&yy,&mm,&dd,&hh,&mi,&ss)==6)
    return 1;
  return 0;
}

int IsISO(char *d) {
  int yy,mm,dd,hh,mi,ss;
  if (sscanf(d,"%04d-%02d-%02d %02d:%02d:%02d",&yy,&mm,&dd,&hh,&mi,&ss)==6)
    return 1;
  return 0;
}

void Date2Timestamp(char *d1,char *d2) {
  int yy,mm,dd;
  int hh,mi,ss;

  //VALUE=DATE:20071011
  if (sscanf(d1,"%04d%02d%02dT%02d%02d%02d",&yy,&mm,&dd,&hh,&mi,&ss)==6)
    sprintf(d2,"%04d-%02d-%02d %02d:%02d:%02d+00",yy,mm,dd,hh,mi,ss); // should be GMT+TZOFFSET
  else {
    if (strstr(d1,"VALUE=")) {
      char *ptr=d1+6;
      if (strstr(ptr,"DATE:")) {
	sscanf(ptr,"DATE:%04d%02d%02d",&yy,&mm,&dd);
	sprintf(d2,"%04d-%02d-%02d",yy,mm,dd);
      } else {
	strcpy(d2,"");
      }
    } else {
      strcpy(d2,"");
    }
  }
}

void Time2DT(time_t t,char *ts,int localtz,int notime) {
  struct tm *tt;
  tzset();
  tt=gmtime(&t);

  if (notime) {
    sprintf(ts,"%04d%02d%02d",
	    tt->tm_year+1900,tt->tm_mon+1,tt->tm_mday);
  } else {
    sprintf(ts,"%04d%02d%02dT%02d%02d%02dZ",
	    tt->tm_year+1900,tt->tm_mon+1,tt->tm_mday,tt->tm_hour,tt->tm_min,tt->tm_sec);
  }
}

/*
Durch die Vereinheitlichung der unterschiedlichen Sommerzeitregelungen in der Europäischen Union wurde die Sommerzeit
1996 in Deutschland um einen Monat verlängert und gilt gemäß Richtlinie 2000/84/EG des Europäischen Parlamentes und des
Rates seitdem vom letzten Sonntag im März um 2:00 Uhr MEZ bis zum letzten Sonntag im Oktober um 3:00 Uhr MESZ.
*/

/*

Sommerzeit:

30.3.2008 2:00 - 26.10.2008 3:00
29.3.2008 2:00 - 25.10.2008 3:00

BEGIN DST 1206838800 : 2008-03-30 01:00:00+00   END DST 1224986400 : 2008-10-26 02:00:00+00
BEGIN DST 1238288400 : 2009-03-29 01:00:00+00   END DST 1256436000 : 2009-10-25 02:00:00+00
BEGIN DST 1269738000 : 2010-03-28 01:00:00+00   END DST 1288490400 : 2010-10-31 02:00:00+00
BEGIN DST 1301187600 : 2011-03-27 01:00:00+00   END DST 1319940000 : 2011-10-30 02:00:00+00
*/

time_t dst[][2]={
  {1206838800,1224986400}, // 2008
  {1238288400,1256436000}, // 2009
  {1269738000,1288490400}, // 2010
  {1301187600,1319940000}, // 2011
  {0,0}
};

int DST(time_t t) {
  int i=0;
  do {
    if ((dst[i][0]<=t) && (t<=dst[i][1])) return 1;
    i++;
  } while (dst[i][0]>0);

  return 0;
}

time_t UTC(time_t t,char *tz) {
  time_t r=t;
  if (
       (!strcmp(tz,"Europe/Central")) ||
       (strstr(tz,"GMT+01")) ||
       (strstr(tz,"Berlin"))
     ) 
  {
    r=r-3600;
    // printf("UTC: %s\n",tz);
  }
  if (DST(r)==1) {
    r=r-3600;
    // printf("UTC: apply daylight saving\n");
  }
  return r;
}

char *Time2ISO(time_t t) {  // returns GMT Time representation
  struct tm *tt;
  tzset();

  tt=gmtime(&t);

  static char ts[256];

  sprintf(ts,"%04d-%02d-%02d %02d:%02d:%02d+00",
	  tt->tm_year+1900,tt->tm_mon+1,tt->tm_mday,tt->tm_hour,tt->tm_min,tt->tm_sec);

  // should be GMT+TZOFFSET

  return ts;
}

time_t ISO2Time(char *ts,int localtz) {
  // lastmod looks like 'yyyy-mm-dd hh:mi:ss+ofs'
  int yy,mm,dd;
  int hh,mi,ss;
  int ofs;

  // printf("\nISO2Time: %s\n",ts);

  int r=sscanf(ts,"%04d-%02d-%02d %02d:%02d:%02d+%02d",&yy,&mm,&dd,&hh,&mi,&ss,&ofs);
  if (r==6) ofs=0;

  struct tm tt;

  tt.tm_sec=ss;
  tt.tm_min=mi; // +ofs;
  tt.tm_hour=hh;  // we are working with GMT only!!!
  tt.tm_mday=dd;
  tt.tm_mon=mm-1;
  tt.tm_year=yy-1900;

  time_t t;
  
  t=mktime(&tt); // returns UTC

  // printf("ISO2Time: UTC=%s\n",Time2ISO(t));

  return t;
}

time_t DT2Time(char *ts,int localtz) {
  int yy,mm,dd;
  int hh,mi,ss;
  struct tm tt;

  int r=sscanf(ts,"%04d%02d%02dT%02d%02d%02dZ",&yy,&mm,&dd,&hh,&mi,&ss);
  if (r==6) {
    tt.tm_sec=ss;
    tt.tm_min=mi;
    tt.tm_hour=hh+localtz; // TS given in timezone +localtz
    tt.tm_mday=dd;
    tt.tm_mon=mm-1;
    tt.tm_year=yy-1900;
  } else {
    r=sscanf(ts,"%04d%02d%02d",&yy,&mm,&dd);
    if (r==3) {
      tt.tm_sec=0;
      tt.tm_min=0;
      tt.tm_hour=0+localtz; // TS given in timezone +localtz
      tt.tm_mday=dd;
      tt.tm_mon=mm-1;
      tt.tm_year=yy-1900;
    } else {
      // error parsing date
      tt.tm_sec=0;
      tt.tm_min=0;
      tt.tm_hour=0+localtz; // TS given in timezone +localtz
      tt.tm_mday=1;
      tt.tm_mon=0;
      tt.tm_year=1900;
    }
  }

  time_t t;
  
  t=mktime(&tt);

  if (DST(t)) {  // daylightsaving active !!!
    t=t+3600;
  }

  return t;
}

time_t Date2Time(char *ts,int localtz) {
  char d[255];

  Date2Timestamp(ts,d);
  return ISO2Time(d,localtz);
}

char *ISO2DT(char *s,int localtz) {
  static char ts[256];
  time_t t=ISO2Time(s,localtz);
  Time2DT(t,ts,0);
  return ts;
}
