#include <vCalendar.h>
#include <string.h>
#include <stdlib.h>

#include <stdio.h>
#include <time.h>

#include <timetools.h>

static char *escape_string=NULL;

char *ESCAPE(char *s) {
  if (escape_string!=NULL) delete escape_string;
  int c=0;
  for (int i=0;i<strlen(s);i++) {
    if (s[i]=='\\') c+=2;
    if (s[i]==',') c+=2;
    if (s[i]=='\n') c+=2;
    if (s[i]=='\'') c++;
  }
  escape_string=new char [strlen(s)+c+1];
  strcpy(escape_string,"");
  for (int i=0;i<strlen(s);i++) {
    switch (s[i]) {
    case '\\': strcat(escape_string,"\\\\"); break;
    case '\'': strcat(escape_string,"''"); break;
    case ',': strcat(escape_string,"\\,"); break;
    case '\n': strcat(escape_string,"\\N"); break;
    default: strncat(escape_string,s+i,1);
    }

  }

  return escape_string;
}

static int depth=0;
const char* WEEKDAYS[]={ "", "MO", "TU", "WE", "TH", "FR", "SA", "SU" };

class daylight_saving_struct {
 public:
  Date dlbegin;
  Date dlend;
  daylight_saving_struct(int i=0) {
    dlbegin=NoDate();
    dlend=NoDate();
  }
};

static List<daylight_saving_struct> daylightsavings;

void AddDaylightsaving(Date d1,Date d2) {
  daylight_saving_struct ds;
  ds.dlbegin=d1;
  ds.dlend=d2;
  daylightsavings.Add(ds);
}

#define APPEND(buffer,X,V) \
if (strlen(V)>0) { \
    char *app_b; \
    if (buffer) { \
      app_b=new char [strlen(buffer)+strlen(X)+strlen(V)+1+1]; \
      strcpy(app_b,buffer); \
      delete buffer; \
    } else { \
      app_b=new char [strlen(X)+strlen(V)+1+1]; \
      strcpy(app_b,""); \
    } \
    buffer=app_b; \
    strcat(buffer,X); \
    strcat(buffer,V); \
    strcat(buffer,"\n"); \
}

#define SAPPEND(buffer,X) \
if (strlen(X)>0) { \
  char *app_b; \
  if (buffer) { \
    app_b=new char [strlen(buffer)+strlen(X)+1]; \
    strcpy(app_b,buffer); \
    delete buffer; \
  } else { \
    app_b=new char [strlen(X)+1]; \
    strcpy(app_b,""); \
  } \
  buffer=app_b; \
  strcat(buffer,X); \
}

#define IAPPEND(buffer,X) \
{ \
  char *app_b; \
  char XS[1024]; \
  sprintf(XS,"%d",X); \
  if (buffer) { \
    app_b=new char [strlen(buffer)+strlen(XS)+1]; \
    strcpy(app_b,buffer); \
    delete buffer; \
  } else { \
    app_b=new char [strlen(XS)+1]; \
    strcpy(app_b,""); \
  } \
  buffer=app_b; \
  strcat(buffer,XS); \
}

#define DTAPPEND(buffer,X,V) \
if (V>0) { \
   char ts[255]; \
   Time2DT(V,ts,0); \
   APPEND(buffer,X,ts) \
}

#define DTAPPENDNOTIME(buffer,X,V) \
if (V>0) { \
   char ts[255]; \
   Time2DT(V,ts,0,1); \
   APPEND(buffer,X,ts) \
}

long sign(long x) {
  if (x<0) return -1;
  return 1;
}

int ParseArgument(const char *ptr,const char *aname,char *value,const char *p_e) {
  char *p=strstr(ptr,aname);
  if ((p) && (((p_e) && (p<p_e)) || (!p_e))) {
    char *pp=p+strlen(aname);
    char *eol=strstr(pp,"\n");
    memset(value,0,2048);
    strncpy(value,pp,eol-pp);
    return 1;
  }
  return 0;
}

//----------------------------------------------------------------------------------------

OrderedWeekday::OrderedWeekday(int pos_,WEEKDAY w) {
  pos=pos_;
  weekday=w;
}

void OrderedWeekday::Pos(int pos_) { pos=pos_; }
int OrderedWeekday::Pos() { return pos; }
void OrderedWeekday::Weekday(WEEKDAY w) { weekday=w; }
WEEKDAY OrderedWeekday::Weekday() { return weekday; }

const char* vRRule::FTYPES[]={"", "SECONDLY" , "MINUTELY" , "HOURLY" , "DAILY" , "WEEKLY" , "MONTHLY" , "YEARLY" };

vRRule::vRRule() {
  frequency=F_NO_FREQUENCY;
  wkst=W_NO_WEEKDAY;
  count=-1;
  until=NoDate();
  interval=1;
}

void vRRule::Frequency(FTYPE x) {
  frequency=x;
}

vRRule::FTYPE vRRule::Frequency() {
  return frequency;
}

void vRRule::Wkst(WEEKDAY x) { 
  wkst=x;
}

void vRRule::Count(int x) {
  count=x;
}

void vRRule::Until(const Date &x) {
  until=x;
}

Date vRRule::Until() {
  return until;
}

void vRRule::Interval(int x) {
  interval=x;
}

int vRRule::Interval() {
  return interval;
}

#define RRULEMETHODS(MNAME,DTYPE,RTYPE,LIST)	    \
  void vRRule::Add##MNAME(DTYPE x) { LIST.Add(x); } \
  int vRRule::SizeOf##MNAME() { return LIST.Size(); } \
  RTYPE vRRule::Get##MNAME(int i) { return LIST.Get(i); }

RRULEMETHODS(Date,const Date &,Date,datelist)
RRULEMETHODS(YearDay,int,int,yeardaylist)
RRULEMETHODS(MonthDay,int,int,monthdaylist)
RRULEMETHODS(Month,int,int,monthlist)
RRULEMETHODS(WeekNo,int,int,weeknolist)
RRULEMETHODS(Hour,int,int,hourlist)
RRULEMETHODS(Minute,int,int,minutelist)
RRULEMETHODS(Second,int,int,secondlist)
RRULEMETHODS(Pos,int,int,poslist)
RRULEMETHODS(Weekday,const OrderedWeekday &,OrderedWeekday,weekdaylist)

#define BUILDLISTOUTPUT(L,S) \
  if (L.Size()>0) { \
    if (!isfirst) SAPPEND(*buffer,";"); \
    SAPPEND(*buffer,S); \
    SAPPEND(*buffer,"="); \
    for (int i=0;i<L.Size();i++) { \
      if (i>0) SAPPEND(*buffer,","); \
      IAPPEND(*buffer,L.Get(i)); \
    } \
    isfirst=0; \
  }
    
void vRRule::Write(char **buffer) {
  int isfirst=1;

  if (datelist.Size()>0) {
  } else { // end of RDATE, begin of RRULE
    SAPPEND(*buffer,"RRULE:");

    if (frequency!=F_NO_FREQUENCY) {
      if (!isfirst) SAPPEND(*buffer,";");
      SAPPEND(*buffer,"FREQ=");
      SAPPEND(*buffer,FTYPES[frequency]);
      isfirst=0;
    }
    
    if (wkst!=W_NO_WEEKDAY) {
      if (!isfirst) SAPPEND(*buffer,";");
      SAPPEND(*buffer,"WKST=");
      SAPPEND(*buffer,WEEKDAYS[wkst]);
      isfirst=0;
    }
    
    if (count>0) {
      if (!isfirst) SAPPEND(*buffer,";");
      SAPPEND(*buffer,"COUNT=");
      IAPPEND(*buffer,count);
      isfirst=0;
    }
    
    if (until!=NoDate()) {
      if (!isfirst) SAPPEND(*buffer,";");
      SAPPEND(*buffer,"UNTIL=");
      SAPPEND(*buffer,until.ToDT());
      isfirst=0;
    }
    
    if (interval>0) {
      if (!isfirst) SAPPEND(*buffer,";");
      SAPPEND(*buffer,"INTERVAL=");
      IAPPEND(*buffer,interval);
      isfirst=0;
    }
    
    BUILDLISTOUTPUT(yeardaylist,"BYYEARDAY");
    BUILDLISTOUTPUT(monthdaylist,"BYMONTHDAY");
    BUILDLISTOUTPUT(monthlist,"BYMONTH");
    BUILDLISTOUTPUT(weeknolist,"BYWEEKNO");
    BUILDLISTOUTPUT(hourlist,"BYHOUR");
    BUILDLISTOUTPUT(minutelist,"BYMINUTE");
    BUILDLISTOUTPUT(secondlist,"BYSECOND");
    BUILDLISTOUTPUT(poslist,"BYSETPOS");
    
    if (weekdaylist.Size()>0) {
      if (!isfirst) SAPPEND(*buffer,";");
      SAPPEND(*buffer,"BYDAY=");
      for (int i=0;i<weekdaylist.Size();i++) {
	if (i>0) SAPPEND(*buffer,",");
	OrderedWeekday w=weekdaylist.Get(i);
	if (w.Pos()) {
	  char wb[128];
	  sprintf(wb,"%d%s",w.Pos(),WEEKDAYS[w.Weekday()]);
	  SAPPEND(*buffer,wb);
	} else
	  SAPPEND(*buffer,WEEKDAYS[w.Weekday()]);
      }
      isfirst=0;
    }
    SAPPEND(*buffer,"\n");
  } // end of RRULE
}

//----------------------------------------------------------------------------------------

vAlarm::vAlarm() {
  description=NULL;
  attach=NULL;
  attachfmt=NULL;
  duration=NoDate();
  trigger=NoDate();
  trigger_related=-1;
  repeat=0;
  refdate=NoDate();
  when=NoDate();
  mapiwhen=NoDate();
  mozlastack=NoDate();
  base64=0;
  binary=0;
}

void vAlarm::Description(char *x) {
  if (description) {
    delete description;
    description=NULL;
  }
  if (strlen(x)>0) {
    description=new char [strlen(x)+1];
    strcpy(description,x);
  }
}

char* vAlarm::Description() { return description; }

unsigned long vAlarm::Action() { return action; }
void vAlarm::Action(unsigned long x) { action=x; }

char *vAlarm::Attach() { return attach; }
void vAlarm::Attach(char *x) {
  if (attach) {
    delete attach;
    attach=NULL;
  }
  if (strlen(x)>0) {
    attach=new char [strlen(x)+1];
    strcpy(attach,x);
  }
}

char *vAlarm::AttachFormat() { return attachfmt; }
void vAlarm::AttachFormat(char *x) {
  if (attachfmt) {
    delete attachfmt;
    attachfmt=NULL;
  }
  if (strlen(x)>0) {
    attachfmt=new char [strlen(x)+1];
    strcpy(attachfmt,x);
  }
}

int vAlarm::Base64() { return base64; }
void vAlarm::Base64(int x) { base64=x; }
int vAlarm::Binary() { return binary; }
void vAlarm::Binary(int x) { binary=x; }

void vAlarm::Repeat(int x) { repeat=x; }
int vAlarm::Repeat() { return repeat; }

void vAlarm::Trigger(Date x) { trigger=x; }
const Date& vAlarm::Trigger() { return trigger; }
void vAlarm::TriggerRelated(int x) { trigger_related=x; }
int vAlarm::TriggerRelated() { return trigger_related; }

void vAlarm::SetRefDate(Date x) {
  refdate=x;
}

const Date& vAlarm::RelativeTrigger() {
  return (trigger-refdate);
}

void vAlarm::MAPIWhen(Date x) { mapiwhen=x; }
const Date& vAlarm::MAPIWhen() { return mapiwhen; }

void vAlarm::MOZLastAck(Date x) { mozlastack=x; }
const Date& vAlarm::MOZLastAck() { return mozlastack; }

void vAlarm::Duration(Date x) { duration=x; }
const Date& vAlarm::Duration() { return duration; }

const List<vAttendee*> & vAlarm::Attendees() { return attendees; }
int vAlarm::NumAttendees() { return attendees.Size(); };
vAttendee* vAlarm::Attendee(int i) { return attendees.Get(i); }
void vAlarm::AddAttendee(vAttendee *x) { attendees.Add(x); };

const char* vAlarm::ALARMACTIONS[]={"","AUDIO","DISPLAY","EMAIL","PROCEDURE"};

void vAlarm::Write(char **buffer) {
  APPEND(*buffer,"","BEGIN:VALARM");
  if (repeat>0) {
    SAPPEND(*buffer,"REPEAT:");
    IAPPEND(*buffer,repeat);
    SAPPEND(*buffer,"\n");
  }
  if (duration!=NoDate()) {
    SAPPEND(*buffer,"DURATION:");
    SAPPEND(*buffer,duration.ToDT());
    SAPPEND(*buffer,"\n");
  }
  if (action!=A_NO_ALARMACTION) {
    APPEND(*buffer,"ACTION:",ALARMACTIONS[action]);
  }
  if (attach) {
    SAPPEND(*buffer,"ATTACH");
    if (attachfmt) {
      SAPPEND(*buffer,";FMTTYPE=");
      SAPPEND(*buffer,attachfmt);
    }
    SAPPEND(*buffer,":");
    SAPPEND(*buffer,attach);
    SAPPEND(*buffer,"\n");
  }
  if (description) {
    APPEND(*buffer,"DESCRIPTION:",description);
  }
  if (trigger!=NoDate()) {
    SAPPEND(*buffer,"TRIGGER");
    // if (trigger.Relative()==0) {
    // SAPPEND(*buffer,";VALUE=DATE-TIME");
    // }
    if (Base64()==1) SAPPEND(*buffer,";");
    if (Binary()==1) SAPPEND(*buffer,";");
    if (TriggerRelated()==1) SAPPEND(*buffer,";RELATED=START");
    if (TriggerRelated()==2) SAPPEND(*buffer,";RELATED=END");
    SAPPEND(*buffer,":");
    // SAPPEND(*buffer,trigger.ToDT());
    Date rd;
    if (trigger<refdate) {
      rd=refdate-trigger;

      SAPPEND(*buffer,"-");
      SAPPEND(*buffer,rd.ToDT());
    } else {
      rd=trigger-refdate;
      SAPPEND(*buffer,rd.ToDT());
    }
    SAPPEND(*buffer,"\n");
  }
  if (mapiwhen!=NoDate()) {
    SAPPEND(*buffer,"X-MAPI-WHEN:");
    SAPPEND(*buffer,mapiwhen.ToDT());
    SAPPEND(*buffer,"\n");
  }
  if (mozlastack!=NoDate()) {
    SAPPEND(*buffer,"X-MOZ-LASTACK:");
    SAPPEND(*buffer,mozlastack.ToDT());
    SAPPEND(*buffer,"\n");
  }

  // attendees

  APPEND(*buffer,"","END:VALARM")
}

#define SCOMP(X,Y) ((X) && (Y)) if (strcmp(X,Y))

int vAlarm::operator != (vAlarm &b) {
  // if (Action()!=b.Action()) return 1;
  //  if SCOMP(Description(),b.Description()) return 1;
  //  if SCOMP(Attach(),b.Attach()) return 1;
  //  if SCOMP(AttachFormat(),b.AttachFormat()) return 1;
  //  if (Repeat()  !=b.Repeat()) return 1;
  Date d1=Trigger();
  Date d2=b.Trigger();
  printf("Trigger: [%s] [%s]\n",d1.ToChar(),d2.ToChar());
  if (Trigger() !=b.Trigger()) return 1;
  // if (Duration()!=b.Duration()) return 1;
  

  return 0;
}

//-----------------------------------------------------------------------------

const char* vAttendee::ROLES[]={"","CHAIR","REQ-PARTICIPANT", "OPT-PARTICIPANT", "NON-PARTICIPANT" ,"ORGANIZER" };
const char* vAttendee::PARTSTATS[]={"", "NEEDS-ACTION", "ACCEPTED", "DECLINED", "TENTATIVE"};

vAttendee::vAttendee(const char *t) {
  strcpy(atype,t);
  strcpy(name,"");
  strcpy(cn,"");
  role=R_NO_ROLE;
  partstat=PS_NO_PARTSTAT;
  rsvp=0;
}
  
void vAttendee::Name(char *x) { strcpy(name,x); }
char* vAttendee::Name() { return name; }

void vAttendee::Cn(char *x) { strcpy(cn,x); }
char* vAttendee::Cn() { return cn; }

void vAttendee::Role(ROLE x) { role=x; }
vAttendee::ROLE vAttendee::Role() { return role; }

void vAttendee::Partstat(PARTSTAT x) { partstat=x; }
vAttendee::PARTSTAT vAttendee::Partstat() { return partstat; }

void vAttendee::Rsvp(int x) { rsvp=x; }
int vAttendee::Rsvp() { return rsvp; }

void vAttendee::Write(char **buffer) {
  SAPPEND(*buffer,atype);
  if (role==R_ORGANIZER) {
    SAPPEND(*buffer,"ORGANIZER");
  } else {
    SAPPEND(*buffer,"ATTENDEE");
    if (role!=R_NO_ROLE) { 
      SAPPEND(*buffer,";ROLE="); 
      SAPPEND(*buffer,ROLES[role]); 
    }
  }
  if (partstat!=PS_NO_PARTSTAT) {
    SAPPEND(*buffer,";PARTSTAT=");
    SAPPEND(*buffer,PARTSTATS[partstat]);
  }
  if (rsvp) {
    SAPPEND(*buffer,";RSVP=TRUE");
  } else {
    SAPPEND(*buffer,";RSVP=FALSE");
  }

  if (strlen(cn)>0) {
    SAPPEND(*buffer,";CN=");
    SAPPEND(*buffer,cn);
    SAPPEND(*buffer,"");
  }
  SAPPEND(*buffer,":");
  SAPPEND(*buffer,name);
  SAPPEND(*buffer,"\n");
}

//-----------------------------------------------------------------------------

void vEvent::RRule(vRRule *x) { rrule=x; }
vRRule* vEvent::RRule() { return rrule; }

int vEvent::ALM() {
  if (alarm) {
    return 1;
  }

  return 0;
}

const char* vEvent::ALM_Time() {
  if (alarm) {
    Date d=alarm->Trigger();
    if (d.Relative()) {
      d=DTStart()+alarm->Trigger();
      //      printf("ALM REL TRIGGER:[%s]\n",d.ISOChar());
      //      printf("     DTSTART is:[%s]\n",DTStart().ISOChar());
    } else {
      d=alarm->Trigger();
      //      printf("ALM ABS TRIGGER:[%s]\n",d.ISOChar());
    }

    /*
    // convert to GMT
    if (tz) {
      // check, if daylightsaving is on for this dtstart
      int i=0;
      while (tz[i]!=NULL) {
	if (!strcmp(ESCAPE(d.Timezone()),tz[i]->Name())) {
	  Date dd=tz[i]->GMT(d);
	  return dd.ISOChar();
	}
	i++;
      }
    }
    */

    return d.ISOChar();
  }

  return "";
}

const char* vEvent::ALM_SoundFile() {
  if (alarm) {
    char *s=alarm->Attach();
    if (s) return s;
  }

  return "";
}

int vEvent::ALM_Options() {
  if (alarm) {
  }

  return 0;
}

int vEvent::ALM_Count() {
  if (alarm) {
  }

  return 0;
}

int vEvent::REC_Type() {
  if (rrule) {
    switch (rrule->Frequency()) {
    case vRRule::F_DAILY  :      return 0; break;
    case vRRule::F_WEEKLY :      return 1; break;
    case vRRule::F_MONTHLY:      return 2; break;
    case vRRule::F_MONTHLYBYDAY: return 3; break;
    case vRRule::F_YEARLY:       return 5; break;
    }
  }

  return -1;
}

void vEvent::REC_Type(int x) {
  if (!rrule) rrule=new vRRule();
  switch (x) {
  case 0: rrule->Frequency(vRRule::F_DAILY); break;
  case 1: rrule->Frequency(vRRule::F_WEEKLY); break;
  case 2: rrule->Frequency(vRRule::F_MONTHLY); break;
  case 3: rrule->Frequency(vRRule::F_MONTHLYBYDAY); break;
  case 5: rrule->Frequency(vRRule::F_YEARLY); break;
  default: rrule->Frequency(vRRule::F_NO_FREQUENCY);
  }
}

int vEvent::REC_Interval() {
  if (rrule) return rrule->Interval();
  return 0;
}

void vEvent::REC_Interval(int x) {
}

int vEvent::REC_Month() {
  if (rrule) {
    if (rrule->SizeOfMonth()==1) {
      return rrule->GetMonth(0);
    }
  }
  return 0;
}

void vEvent::REC_Month(int x) {
}

int vEvent::REC_Day() {
  if (rrule) {
    if (rrule->SizeOfMonthDay()==1) {
      return rrule->GetMonthDay(0);
    }
  }
  return 0;
}

void vEvent::REC_Day(int x) {
}

int vEvent::REC_DAYMASK() {
  return rec_day_of_week_mask;
}

void vEvent::REC_DAYMASK(int x) {
}

int vEvent::REC_Instance() {
  return rec_instance;
}

void vEvent::REC_Instance(int x) {
}

int vEvent::REC_NoEndDate() {
  if (rrule) {
    if (rrule->Until()!=NoDate()) return 0;
  }
  return 1;
}

void vEvent::REC_NoEndDate(int x) {
}

const char *vEvent::REC_StartDate() {
  if (rrule) {
    Date d=DTStart();
    if (AllDay()) {
      d.Hour(-1);
    }
    // return d.ISOChar();
    return d.ToDT();
  }
  return "";
}

const char *vEvent::REC_EndDate() {
  if (rrule) {
    if (rrule->Until()!=NoDate()) return rrule->Until().ToDT();
  }
  return "";
}

void vEvent::REC_EndDate(char *x) {
}

unsigned long vEvent::GUID() {
  return guid;
}

void vEvent::GUID(unsigned long x) {
  guid=x;
}

const char* vEvent::STATES[]={"","TENTATIVE","CONFIRMED","CANCELLED"};
const char* vEvent::TRANSPARENTS[]={"","OPAQUE","TRANSPARENT"};
const char* vEvent::BUSYSTATES[]={"","BUSY","FREE","TENTATIVE","OOF","UNAVAILABLE"};

vEvent::vEvent(char *s,vTimezone **tz_) {
  guid=0;
  
  dtstart=NoDate();
  dtend=NoDate();
  dtstamp=NoDate();
  created=NoDate();
  last_modified=NoDate();
  
  strcpy(summary,"");
  strcpy(description,"");
  strcpy(location,"");
  // strcpy(rrule,"");
  rrule=NULL;
  organizer=NULL;
  busystate=B_NO_BUSYSTATE;
  state=S_NO_STATE;
  transparency=T_NO_TRANSPARENCY;
  strcpy(category,"");
  strcpy(event_class,"");
  strcpy(uid,"");
  mapilabel=sequence=priority=-1;
  allday=0;
  buffer=NULL;

  rec_type=-1;
  rec_interval=1;
  rec_month_of_year=rec_day_of_month=rec_day_of_week_mask=0;
  rec_instance=rec_no_end_date=rec_occurences=0;
  strcpy(rec_start_date_pattern,"");
  strcpy(rec_end_date_pattern,"");

  mapilabel=-1;
  cdo_importance=-1;

  tz=tz_;

  buffer=NULL;
  dtstamp=time(NULL);

  alarm=NULL;
}

vEvent::~vEvent() {
  if (buffer) delete buffer;

  if (alarm) delete alarm;

  if (tz) {
  }

  if (rrule) delete rrule;
}

void vEvent::Write(char **buffer) {
  char b[255];

  APPEND(*buffer,"BEGIN:","VEVENT");
  APPEND(*buffer,"UID:",uid);
  APPEND(*buffer,"DTSTAMP:",dtstamp.ToDT());
  APPEND(*buffer,"LAST-MODIFIED:",last_modified.ToDT());
  APPEND(*buffer,"CREATED:",created.ToDT());

  SAPPEND(*buffer,"DTSTART");
  if ((strlen(dtstart.Timezone())>0) && (strcmp(dtstart.Timezone(),"UTC"))) {
    SAPPEND(*buffer,";TZID=\"");
    SAPPEND(*buffer,dtstart.Timezone());
    SAPPEND(*buffer,"\"");
  }
  if (dtstart.Type()==Date::DT_DATE) {
    SAPPEND(*buffer,";VALUE=DATE");
  }
  APPEND(*buffer,":",dtstart.ToDT());

  SAPPEND(*buffer,"DTEND");
  if ((strlen(dtend.Timezone())>0) && (strcmp(dtend.Timezone(),"UTC"))) {
    SAPPEND(*buffer,";TZID=\"");
    SAPPEND(*buffer,dtend.Timezone());
    SAPPEND(*buffer,"\"");
  }
  if (dtend.Type()==Date::DT_DATE) {
    SAPPEND(*buffer,";VALUE=DATE");
  }
  APPEND(*buffer,":",dtend.ToDT());

  if (allday>=0) {
    if (allday) {
      APPEND(*buffer,"X-MICROSOFT-CDO-ALLDAYEVENT:","TRUE");
    } else {
      APPEND(*buffer,"X-MICROSOFT-CDO-ALLDAYEVENT:","FALSE");
    }
  }

  if (Status()!=S_NO_STATE) APPEND(*buffer,"STATUS:",STATES[Status()]);

  if (rrule) rrule->Write(buffer);

  if (strlen(summary)>0) APPEND(*buffer,"SUMMARY:",summary);
  if (strlen(description)>0) APPEND(*buffer,"DESCRIPTION:",description);
  if (strlen(location)>0) APPEND(*buffer,"LOCATION:",location);

  if (organizer) organizer->Write(buffer);
  if (Busystate()!=B_NO_BUSYSTATE) APPEND(*buffer,"X-MICROSOFT-CDO-BUSYSTATUS:",BUSYSTATES[Busystate()]);
  if (Transparency()!=T_NO_TRANSPARENCY) APPEND(*buffer,"TRANSP:",TRANSPARENTS[Transparency()]);
  
  if (Sequence()>=0) {
    SAPPEND(*buffer,"SEQUENCE:");
    IAPPEND(*buffer,Sequence());
    SAPPEND(*buffer,"\n");
  }
  if (Priority()>=0) {
    SAPPEND(*buffer,"PRIORITY:");
    IAPPEND(*buffer,Priority());
    SAPPEND(*buffer,"\n");
  }
  if (CDOImportance()>=0) {
    SAPPEND(*buffer,"X-MICROSOFT-CDO-IMPORTANCE:");
    IAPPEND(*buffer,CDOImportance());
    SAPPEND(*buffer,"\n");
  }
  if (MAPILabel()>=0) {
    SAPPEND(*buffer,"X-MAPI-LABEL:");
    IAPPEND(*buffer,MAPILabel());
    SAPPEND(*buffer,"\n");
  }
  
  if (strlen(category)>0) {
    APPEND(*buffer,"CATEGORIES:",category);
  }

  if (strlen(event_class)>0) {
    APPEND(*buffer,"CLASS:",event_class);
  }

  for (int i=0;i<attendees.Size();i++)
    attendees.Get(i)->Write(buffer);

  // valarm

  if (alarm) {
    alarm->SetRefDate(DTStart());
    alarm->Write(buffer);
  }

  APPEND(*buffer,"END:","VEVENT");
}

//-----------------------------------------------------------------------------

char *vEvent::UID() {
  return uid;
}

void vEvent::UID(char *x) {
  strcpy(uid,x);
}

int vEvent::Sequence() { return sequence; }
void vEvent::Sequence(int x) { sequence=x; }

int vEvent::Priority() { return priority; }
void vEvent::Priority(int x) { priority=x; }

int vEvent::AllDay() {
  return allday;
}

void vEvent::AllDay(int x) {
  allday=x;
}

int vEvent::MAPILabel() {
  return mapilabel;
}

void vEvent::MAPILabel(int x) {
  mapilabel=x;
}

int vEvent::CDOImportance() {
  return cdo_importance;
}

void vEvent::CDOImportance(int x) {
  cdo_importance=x;
}

void vEvent::Created(Date x) { created=x; }
const Date& vEvent::Created() { return created; }

void vEvent::LastModified(Date x) { last_modified=x; }
const Date& vEvent::LastModified() { return last_modified; }

void vEvent::DTStamp(Date x) { dtstamp=x; }
const Date& vEvent::DTStamp() { return dtstamp; }

void vEvent::DTStart(Date x) { 
  dtstart=x;
}

Date vEvent::DTStart() { 
  if (AllDay()) {
    Date d=dtstart;
    d.Hour(0); d.Minute(0); d.Second(0);
    return d;
  }
  
  // convert to GMT
  if (tz) {
    // printf("*** checking for TZ conversion\n");
    // printf("DTStart:%s\n",dtstart.ISOChar());
    // check, if daylightsaving is on for this dtstart
    int i=0;
    while (tz[i]!=NULL) {
      // printf("*** TZ=[%s]\n",tz[i]->Name());
      // printf("*** DZ=[%s]\n",ESCAPE(dtstart.Timezone()));
      if (!strcmp(ESCAPE(dtstart.Timezone()),tz[i]->Name())) {
	// printf("*** %x->GMT\n",tz[i]);
	return tz[i]->GMT(dtstart);
      }
      i++;
    }
  }

  return dtstart; 
}

void vEvent::DTEnd(Date x) { 
  dtend=x; 
}

Date vEvent::DTEnd() { 
  if (AllDay()) {
    Date d=dtstart;
    d.Hour(23); d.Minute(59); d.Second(0);
    return d;
  }

  // convert to GMT
  if (tz) {
    // check, if daylightsaving is on for this dtstart
    int i=0;
    while (tz[i]!=NULL) {
      if (!strcmp(ESCAPE(dtend.Timezone()),tz[i]->Name())) {
	return tz[i]->GMT(dtend);
      }
      i++;
    }
  }

  return dtend; 
}

char *vEvent::Subject() {
  return summary;
}

void vEvent::Subject(char *x) {
  strcpy(summary,x);
}

char *vEvent::Category() {
  return category;
}

void vEvent::Category(char *x) {
  strcpy(category,x);
}

char *vEvent::Class() {
  return event_class;
}

void vEvent::Class(char *x) {
  strcpy(event_class,x);
}

vEvent::STATE vEvent::Status() {
  return state;
}

void vEvent::Status(STATE x) {
  state=x;
}

vEvent::TRANSPARENCY vEvent::Transparency() {
  return transparency;
}

void vEvent::Transparency(TRANSPARENCY x) {
  transparency=x;
}

void vEvent::Busystate(BUSYSTATE x) { busystate=x; }
vEvent::BUSYSTATE vEvent::Busystate() { return busystate; }

void vEvent::Alarm(vAlarm *x) { 
  if (alarm) delete alarm; 
  alarm=x;
}
vAlarm *vEvent::Alarm() { return alarm; }

vAttendee *vEvent::Organizer() {
  return organizer;
}

void vEvent::Organizer(vAttendee *x) {
  organizer=x;
}

const List<vAttendee*> & vEvent::Attendees() { return attendees; }
int vEvent::NumAttendees() { return attendees.Size(); };
vAttendee* vEvent::Attendee(int i) { return attendees.Get(i); }
void vEvent::AddAttendee(vAttendee *x) { attendees.Add(x); };

char *vEvent::Body() {
  return description;
}

void vEvent::Body(char *x) {
  strcpy(description,x);
}

char *vEvent::Location() {
  return location;
}

void vEvent::Location(char *x) {
  strcpy(location,x);
}

int vEvent::Compare(vEvent *ev) {
  if (DTStart()!=ev->DTStart()) return 1;
  if (DTEnd()  !=ev->DTEnd())   return 2;
  // dtstamp,created,last_modified;
  if (strcmp(Subject(),ev->Subject())) return 3;
  if (strcmp(Body(),ev->Body())) return 4;
  if (strcmp(Location(),ev->Location())) return 5;
  if (strcmp(Category(),ev->Category())) return 6;
  
  if ((Organizer()) && (ev->Organizer())) {
    if (strcmp(Organizer()->Name(),ev->Organizer()->Name())) return 7;
  }

  if (
      ((Organizer()) && (!ev->Organizer())) ||
      ((!Organizer()) && (ev->Organizer()))
      ) return 8;
  
  // if (Busystate()!=ev->Busystate()) return 9;           // busy-state not known to palm, but to BlackBerry !!!
  if (Status()!=ev->Status()) return 10;
  // if (Transparency()!=ev->Transparency()) return 11;

  // rrule[1024];
  // int sequence,priority
  if (AllDay()!=ev->AllDay()) return 12;

  if (REC_Type()!=ev->REC_Type()) return 13;
  if (REC_Type()>=0) {
    if (REC_Interval()!=ev->REC_Interval()) return 14;
    if (REC_Month()!=ev->REC_Month()) return 15;
    if (REC_Day()!=ev->REC_Day()) return 16;
    if (REC_DAYMASK()!=ev->REC_DAYMASK()) return 17;
    if (REC_Instance()!=ev->REC_Instance()) return 18;
    if (REC_NoEndDate()!=ev->REC_NoEndDate()) return 19;
    if (!REC_NoEndDate()) {
      printf("rec end date [%s] [%s]\n",REC_EndDate(),ev->REC_EndDate());
      if (strcmp(REC_EndDate(),ev->REC_EndDate())) return 20;
    }

    if (((Alarm()) && (!ev->Alarm())) || ((!Alarm()) && (!ev->Alarm()))) return 21;
    if ((Alarm()) && (ev->Alarm())) {
      if ((*Alarm())!=(*(ev->Alarm()))) return 22;
    }
  }

  return 0;
}

//-----------------------------------------------------------------------------

TZ_DAYLIGHT::TZ_DAYLIGHT(char *s) {
  if (s) {
    int r=Parse(s);
  }
  buffer=NULL;
}

long TZ_DAYLIGHT::OffsetTo() {
  return (offset_to/100)*60+sign(offset_to)*(offset_to % 100);
}

long TZ_DAYLIGHT::OffsetFrom() {
  return (offset_from/100)*60+sign(offset_from)*(offset_from % 100);
}

//    BEGIN:DAYLIGHT
//      DTSTART:20031025T160000
//      TZOFFSETTO:+1100
//      TZOFFSETFROM:+0000
//      TZNAME:EST
//    END:DAYLIGHT
int TZ_DAYLIGHT::Parse(char *s) {
  char *p_b=strstr(s,"BEGIN:DAYLIGHT");
  char *p_e=strstr(s,"END:DAYLIGHT");
  if ((p_b) && (p_e)) {
    char *ptr=p_b+15;
    char value[2048];

    if (ParseArgument(ptr,"TZNAME:",value,p_e))
      strcpy(tzname,value);
    else
      strcpy(tzname,"");

    if (ParseArgument(ptr,"DTSTART:",value,p_e))
      dtstart=DT2Time(value,LOCALTZ);
    else
      dtstart=0;
    
    if (ParseArgument(ptr,"TZOFFSETTO:",value,p_e))
      sscanf(value,"%d",&offset_to);
    else
      offset_to=0;

    if (ParseArgument(ptr,"TZOFFSETFROM:",value,p_e))
      sscanf(value,"%d",&offset_from);
    else
      offset_from=12;

    return 1;
  }

  return -1;
}

void TZ_DAYLIGHT::Write(char **buffer) {
  char b[256];

  APPEND(*buffer,"","BEGIN:DAYLIGHT")
  DTAPPEND(*buffer,"DTSTART:",dtstart)
  sprintf(b,"%+05d",offset_to);
  APPEND(*buffer,"TZOFFSETTO:",b)
  sprintf(b,"%+05d",offset_from);
  APPEND(*buffer,"TZOFFSETFROM:",b)
  APPEND(*buffer,"TZNAME:",tzname)
  APPEND(*buffer,"","END:DAYLIGHT")
}

//-----------------------------------------------------------------------------

TZ_STANDARD::TZ_STANDARD(char *s) {
  if (s) {
    int r=Parse(s);
  }
  buffer=NULL;
}

long TZ_STANDARD::OffsetTo() {
  return (offset_to/100)*60+sign(offset_to)*(offset_to % 100);
}

long TZ_STANDARD::OffsetFrom() {
  return (offset_from/100)*60+sign(offset_from)*(offset_from % 100);
}

//    BEGIN:STANDARD
//      DTSTART:20040328T030000
//      TZOFFSETTO:+1000
//      TZOFFSETFROM:+1100
//      TZNAME:EST
//    END:STANDARD
int TZ_STANDARD::Parse(char *s) {
  char *p_b=strstr(s,"BEGIN:STANDARD");
  char *p_e=strstr(s,"END:STANDARD");
  if ((p_b) && (p_e)) {
    char *ptr=p_b+15;
    char value[2048];

    if (ParseArgument(ptr,"TZNAME:",value,p_e))
      strcpy(tzname,value);
    else
      strcpy(tzname,"");

    if (ParseArgument(ptr,"DTSTART:",value,p_e))
      dtstart=DT2Time(value,LOCALTZ);
    else
      dtstart=0;
    
    if (ParseArgument(ptr,"TZOFFSETTO:",value,p_e))
      sscanf(value,"%+04d",&offset_to);
    else
      offset_to=0;

    if (ParseArgument(ptr,"TZOFFSETFROM:",value,p_e))
      sscanf(value,"%+04d",&offset_from);
    else
      offset_from=12;

    return 1;
  }

  return -1;
}

void TZ_STANDARD::Write(char **buffer) {
  char b[255];

  APPEND(*buffer,"","BEGIN:STANDARD");
  DTAPPEND(*buffer,"DTSTART:",dtstart);
  sprintf(b,"%+04d",offset_to);
  APPEND(*buffer,"TZOFFSETTO:",b);
  sprintf(b,"%+04d",offset_from);
  APPEND(*buffer,"TZOFFSETFROM:",b);
  APPEND(*buffer,"TZNAME",tzname);
  APPEND(*buffer,"","END:STANDARD");
}

//-----------------------------------------------------------------------------

TZ_DEF::TZ_DEF(const char *n) {
  dtstart=NoDate();
  offset_to=0;
  offset_from=0;
  strcpy(tzname,"");
  rrule=NULL;
}

void TZ_DEF::DTStart(const Date &d) {
  dtstart=d;
}

void TZ_DEF::OffsetTo(long o) { offset_to=o; }
long TZ_DEF::OffsetTo() { return offset_to; }
void TZ_DEF::OffsetFrom(long o) { offset_from=o; }
long TZ_DEF::OffsetFrom() { return offset_from; }

void TZ_DEF::RRule(vRRule *x) { rrule=x; }
vRRule *TZ_DEF::RRule() { return rrule; }

void TZ_DEF::Write(char **buffer) {
  char ds[512];
  APPEND(*buffer,"DTSTART:",dtstart.ToDT());

  SAPPEND(*buffer,"TZOFFSETFROM:");
  sprintf(ds,"%+04d",OffsetFrom());
  SAPPEND(*buffer,ds);
  SAPPEND(*buffer,"\n");

  SAPPEND(*buffer,"TZOFFSETTO:");
  sprintf(ds,"%+04d",OffsetTo());
  SAPPEND(*buffer,ds);
  SAPPEND(*buffer,"\n");

  if (rrule) rrule->Write(buffer);
}

vTimezone::vTimezone(char *s) {
  tz_standard=NULL;
  tz_daylight=NULL;

  buffer=NULL;
}

void vTimezone::Name(char* s) {
  strcpy(tzname,s);
}

char *vTimezone::Name() {
  return tzname;
}

void vTimezone::SetStandard(TZ_DEF *t) {
  if (tz_standard) delete tz_standard;
  tz_standard=t;
}

void vTimezone::SetDaylight(TZ_DEF *t) {
  if (tz_daylight) delete tz_daylight;
  tz_daylight=t;
}

long vTimezone::OffsetTo(int daylight) {
  if (daylight) {
    if (tz_daylight) return tz_daylight->OffsetTo();
  } else {
    if (tz_standard) return tz_standard->OffsetTo();
  }

  return 0;
}

long vTimezone::OffsetFrom(int daylight) {
  if (daylight) {
    if (tz_daylight) return tz_daylight->OffsetFrom();
  } else {
    if (tz_standard) return tz_standard->OffsetFrom();
  }
  return 0;
}

const Date &vTimezone::GMT(const Date &date) {
  Date d=date;
  Date dd;

  if (tz_daylight) {
    for (int i=0;i<daylightsavings.Size();i++) {
      // check, if in range
      if ((daylightsavings.Get(i).dlbegin.MkTime()<=date.MkTime()) &&
	  (date.MkTime()<=daylightsavings.Get(i).dlend.MkTime())) {
	// printf("*** DL GMT(%s)-%d\n",d.ISOChar(),OffsetTo(1));
	return d-OffsetTo(1);
      }
    }
  }

  if (tz_standard) {
    // printf("*** ST GMT(%s)-%d\n",d.ISOChar(),OffsetTo(0));
    // dd=d-OffsetTo(0);
    // printf("***    GMT=%s\n",dd.ISOChar());
    return d-OffsetTo(0);
  }

  return date;
}

void vTimezone::Write(char **buffer) {
  char tb[255];

  APPEND(*buffer,"","BEGIN:VTIMEZONE");
  APPEND(*buffer,"TZID:",tzname);
  DTAPPEND(*buffer,"LAST-MODIFIED:",modified);

  if (tz_standard) {
    APPEND(*buffer,"BEGIN:","STANDARD");
    tz_standard->Write(buffer);
    APPEND(*buffer,"END:","STANDARD");
  }
  if (tz_daylight) {
    APPEND(*buffer,"BEGIN:","DAYLIGHT");
    tz_daylight->Write(buffer);
    APPEND(*buffer,"END:","DAYLIGHT");
  }

  APPEND(*buffer,"","END:VTIMEZONE");
}

//-----------------------------------------------------------------------------

vCalendar::vCalendar(char *s) {
  nevent=0;
  buffer=NULL;
  vtimezone=NULL;
  ntz=1;
  vtimezones=new vTimezone* [1];
  vtimezones[0]=NULL;

  strcpy(VERSION,"2.0");
  strcpy(PRODID,"IMAP2SYNC4J");
  strcpy(METHOD,"PUBLISH");

  //if (s!=NULL) {
  //int r=Parse(s);
  //}
}

void vCalendar::AddEvent(vEvent *ev) {
  if (nevent<MAXEVENTS) {
    event[nevent]=ev;
    nevent++;
  }
}

//BEGIN:vCalendar
//  VERSION:2.0
//  X-WR-CALNAME:Australian_TV_Melbourne
//  PRODID:-//Apple Computer\, Inc//iCal 1.5//EN
//  X-WR-RELCALID:C5CAE0F0-766E-11D8-8884-000A27E11D90-CALP
//  X-WR-TIMEZONE:Australia/Melbourne
//  CALSCALE:GREGORIAN
//  METHOD:PUBLISH

//  vTimezone

//  vEvent
//END:vCalendar

void vCalendar::Write(char **buffer) {
  APPEND(*buffer,"","BEGIN:VCALENDAR");
  APPEND(*buffer,"METHOD:",METHOD);
  APPEND(*buffer,"PRODID:",PRODID);
  APPEND(*buffer,"VERSION:",VERSION);

  // APPEND(*buffer,"X-WR-CALNAME:",X_WR_CALNAME)
  // APPEND(*buffer,"X-WR-RELCALID:",X_WR_RELCALID)
  // APPEND(*buffer,"X-WR-TIMEZONE:",X_WR_TIMEZONE)
  // APPEND(*buffer,"CALSCALE:",CALSCALE)
    
  //  vTimezone
  if (vtimezones) {
    int i=0;
    while (vtimezones[i]) {
      vtimezones[i]->Write(buffer);
      i++;
    }
  }

  for (int i=0;i<nevent;i++) {
    //  vEvent
    event[i]->Write(buffer);
  }

  APPEND(*buffer,"","END:VCALENDAR")
}

vTimezone **vCalendar::Timezones() {
  return vtimezones;
}

unsigned int vCalendar::NumEvents() {
  return nevent;
}

vEvent *vCalendar::Event(int i) {
  if ((nevent>0) && (0<=i) && (i<nevent))
    return event[i];

  return NULL;
}

void vCalendar::Timezone(vTimezone *tz) { 
  if (vtimezone) delete vtimezone;
  vtimezone=tz;
}

void vCalendar::AddTimezone(vTimezone *tz) {
  vTimezone **tzz=vtimezones;
  vtimezones=new vTimezone* [ntz+1];
  if (tzz) {
    for (unsigned int i=0;i<ntz;i++) vtimezones[i]=tzz[i];
    delete tzz;
  }
  vtimezones[ntz-1]=tz;
  vtimezones[ntz]=NULL;
  ntz++;
}

void vCalendar::ProductID(char *s) { strcpy(PRODID,s); }
void vCalendar::Version(char *s)   { strcpy(VERSION,s);   }

vCalendar::~vCalendar() {
  if (vtimezone) delete vtimezone;
  if (vtimezones) {
    for (unsigned int i=0;i<ntz;i++) if (vtimezones[i]) delete vtimezones[i];
    delete vtimezones;
  }
  for (int i=0;i<nevent;i++) delete event[i];
}

//------------------------------------------------------------------------------------------------
