#ifndef ICAL_VCALENDAR_
#define ICAL_VCALENDAR_

#include <stdlib.h>
#include <date.h>
#include <List.h>

extern char *ESCAPE(char *s);

typedef enum WEEKDAY { W_NO_WEEKDAY, W_MO, W_TU, W_WE, W_TH, W_FR, W_SA, W_SU };
extern const char* WEEKDAYS[];

class OrderedWeekday {
  int pos;
  WEEKDAY weekday;
 public:
  OrderedWeekday(int pos_=0,WEEKDAY w=W_NO_WEEKDAY);
  void Pos(int pos_);
  int Pos();
  void Weekday(WEEKDAY w);
  WEEKDAY Weekday();
};

class vRRule {
 public:
  enum FTYPE { F_NO_FREQUENCY  , F_SECONDLY , F_MINUTELY , F_HOURLY , F_DAILY , F_WEEKLY , F_MONTHLY , F_MONTHLYBYDAY, F_YEARLY };
 private:
  static const char* FTYPES[];

  FTYPE frequency; // SECONDLY 1 | MINUTELY 2 | HOURLY 3 | DAILY 4 | WEEKLY 5 | MONTHLY 6 | YEARLY 7;
  WEEKDAY wkst;    // start of workweek 1=MO, 2=TU, ..., 7=SU

  int count;     // <0 not define, =0 for ever >0 number of repetitions
  Date until;    // alternative to count, ends on

  int interval;  // time distance (depends on frequency) between repetitions
                 // alternatives
  List<Date> datelist;    // list of repetition dates, (implies use of RDATE instead of RRULE

  List<int> yeardaylist;
  List<int> monthdaylist; // list of days in month (maybe <0 (count from end of month) or >0 (count from start)
  List<int> monthlist;
  List<int> weeknolist;
  List<int> hourlist;
  List<int> minutelist;
  List<int> secondlist;
  List<int> poslist;
  List<OrderedWeekday> weekdaylist;
 public:
  vRRule();

  void Frequency(FTYPE x);
  FTYPE Frequency();
  void Wkst(WEEKDAY x);
  void Count(int x);
  void Until(const Date &x);
  Date Until();
  void Interval(int x);
  int Interval();

  void AddDate(const Date &x);
  int SizeOfDate();
  Date GetDate(int i);

  void AddYearDay(int x);
  int SizeOfYearDay();
  int GetYearDay(int i);

  void AddMonthDay(int x);
  int SizeOfMonthDay();
  int GetMonthDay(int i);

  void AddMonth(int x);
  int SizeOfMonth();
  int GetMonth(int i);

  void AddWeekNo(int x);
  int SizeOfWeekNo();
  int GetWeekNo(int i);

  void AddHour(int x);
  int SizeOfHour();
  int GetHour(int i);

  void AddMinute(int x);
  int SizeOfMinute();
  int GetMinute(int i);

  void AddSecond(int x);
  int SizeOfSecond();
  int GetSecond(int i);

  void AddPos(int x);
  int SizeOfPos();
  int GetPos(int i);

  void AddWeekday(const OrderedWeekday &x);
  int SizeOfWeekday();
  OrderedWeekday GetWeekday(int i);

  void Write(char **buffer);
};

class TZ_DEF {
  Date dtstart;
  long offset_to,offset_from;
  char tzname[255];
  vRRule *rrule;
 public:
  TZ_DEF(const char *n="");
  void Write(char **buffer);

  void DTStart(const Date &d);

  void OffsetTo(long o);
  long OffsetTo();
  void OffsetFrom(long o);
  long OffsetFrom();

  void RRule(vRRule *x);
  vRRule *RRule();
};

class TZ_DAYLIGHT {
  char *buffer;
  time_t dtstart;
  long offset_to,offset_from;
  char tzname[255];
 public:
  TZ_DAYLIGHT(char *s);
  int Parse(char *s);
  void Write(char **buffer);

  long OffsetTo();
  long OffsetFrom();
};

class TZ_STANDARD {
  char *buffer;
  time_t dtstart;
  long offset_to,offset_from;
  char tzname[255];
 public:
  TZ_STANDARD(char *s=NULL);
  int Parse(char *s);
  void Write(char **buffer);

  long OffsetTo();
  long OffsetFrom();
};

class vTimezone {
  char tzname[255];
  time_t modified;
  
  char *buffer;
  //  TZ_DAYLIGHT *tz_dl;
  //  TZ_STANDARD *tz_std;

  TZ_DEF *tz_standard;
  TZ_DEF *tz_daylight;
  Date cdate;
 public:
  vTimezone(char *s=NULL);
  // int Parse(char *s);
  void Write(char **buffer);

  void Name(char *s);
  char *Name();
  long OffsetTo(int daylight=0);
  long OffsetFrom(int daylight=0);

  void SetStandard(TZ_DEF *t);
  void SetDaylight(TZ_DEF *t);

  const Date &GMT(const Date &date);
};

#define ACCMETHOD(M,T) \
  void M(T x); \
  T M();

class vAttendee {
 public:
  enum ROLE { R_NO_ROLE, R_CHAIR, R_REQ_PARTICIPANT, R_OPT_PARTICIPANT, R_NON_PARTICIPANT, R_ORGANIZER };
  enum PARTSTAT { PS_NO_PARTSTAT, PS_NEEDS_ACTION, PS_ACCEPTED, PS_DECLINED, PS_TENTATIVE };
 private:
  static const char* ROLES[];
  static const char* PARTSTATS[];

  char atype[255];
  char name[1024];
  char cn[1024];
  ROLE role;
  PARTSTAT partstat;
  int rsvp;
 public:
  vAttendee(const char* t);
  
  ACCMETHOD(Name,char*);
  ACCMETHOD(Cn,char*);
  ACCMETHOD(Role,ROLE);
  ACCMETHOD(Partstat,PARTSTAT);
  ACCMETHOD(Rsvp,int);

  void Write(char **buffer);
};

class vAlarm {
 public:
  enum ALARMACTION { A_NO_ALARMACTION=0, A_AUDIO=1, A_VIBRATE=2, A_LED=4, A_DISPLAY=8, A_EMAIL=16, A_PROCEDURE=32 };
 private:
  static const char* ALARMACTIONS[];
  unsigned long action;
  char *description;
  char *attach;
  char *attachfmt;
  int base64;
  int binary;
  Date trigger;  // absolute time of trigger
  int trigger_related;
  Date duration; // - seconds BEFORE event, + seconds AFTER event
  int repeat;
  Date refdate;  // start of event
  Date when;     // trigger seconds before/after dtstart, computed as trigger-refdate

  Date mapiwhen;
  Date mozlastack;

  List<vAttendee*> attendees;
 public:
  vAlarm();
  
  unsigned long Action();
  void Action(unsigned long x);

  void Description(char *x);
  char* Description();

  char *Attach();
  void Attach(char *x);
  char *AttachFormat();
  void AttachFormat(char *x);
  int Base64();
  void Base64(int x);
  int Binary();
  void Binary(int x);

  void Repeat(int x);
  int Repeat();
  
  void Trigger(Date x);
  const Date& Trigger();
  void TriggerRelated(int x);
  int TriggerRelated();

  void SetRefDate(Date x);
  const Date& RelativeTrigger();

  void MAPIWhen(Date x);
  const Date& MAPIWhen();

  void MOZLastAck(Date x);
  const Date& MOZLastAck();

  void Duration(Date x);
  const Date& Duration();

  const List<vAttendee*> & Attendees();
  int NumAttendees();
  vAttendee* Attendee(int i);
  void AddAttendee(vAttendee *x);

  void Write(char **buffer);

  int operator != (vAlarm &b);
};


class vEvent {
 public:
  enum STATE { S_NO_STATE  , S_TENTATIVE, S_CONFIRMED, S_CANCELLED };
  enum TRANSPARENCY { T_NO_TRANSPARENCY, T_OPAQUE, T_TRANSPARENT };
  enum BUSYSTATE { B_NO_BUSYSTATE, B_BUSY, B_FREE, B_TENTATIVE, B_OOF, B_UNAVAILABLE };
 private:
  static const char* STATES[];
  static const char* TRANSPARENTS[];
  static const char* BUSYSTATES[];

  unsigned long guid;
  Date dtstart,dtend,dtstamp,created,last_modified;

  char summary[255],description[4096],location[255];
  vAttendee *organizer;
  BUSYSTATE busystate;
  char category[255];
  int sequence,priority,allday;
  char uid[255];
  char *buffer;

  int rec_type;                   // -1: NOREC,FREQ { 0:DAILY 1:WEEKLY 2:MONTHLY (BYMONTHDAY) 3:MONTHLY (BYDAY) 5:YEARLY }
  int rec_interval;               // INTERVAL
  int rec_month_of_year;          // BYMONTH
  int rec_day_of_month;           // BYMONTHDAY:
  int rec_day_of_week_mask;       // BYDAY=  1:SU 2:MO 4:TU 8:WE 16:TH 32:FR 64:SA
  int rec_instance;               // BYDAY=<instance><mask>, e.g. 3 2 = 3MO  3rd monday
  char rec_start_date_pattern[64];// DTSTART
  int rec_no_end_date;            // 0: end date exist, 1: no end date
  char rec_end_date_pattern[64];  // UNTIL
  int rec_occurences;             // -1 ???

  int mapilabel;
  int cdo_importance;
  char event_class[256];
  STATE state;
  TRANSPARENCY transparency;

  List<vAttendee*> attendees;

  vAlarm *alarm;
  vTimezone **tz;

  vRRule *rrule;
 public:
  vEvent(char *s=NULL,vTimezone **tz_=NULL);
  ~vEvent();

  void Write(char **buffer);

  // the data

  unsigned long GUID();
  void GUID(unsigned long x);

  char *UID();
  void UID(char *x);

  int Sequence();
  void Sequence(int x);

  int Priority();
  void Priority(int x);

  int AllDay();
  void AllDay(int x);

  int MAPILabel();
  void MAPILabel(int x);

  int CDOImportance();
  void CDOImportance(int x);

  void Created(Date x);
  const Date& Created();

  void LastModified(Date x);
  const Date& LastModified();

  void DTStamp(Date x);
  const Date& DTStamp();

  void DTStart(Date x);
  Date DTStart();

  void DTEnd(Date x);
  Date DTEnd();

  char *Subject();
  void Subject(char *x);

  char *Body();
  void Body(char *x);

  char *Location();
  void Location(char *x);

  char *Category();
  void Category(char *x);

  char *Class();
  void Class(char *x);

  STATE Status();
  void Status(STATE x);

  TRANSPARENCY Transparency();
  void Transparency(TRANSPARENCY x);

  vAttendee *Organizer();
  void Organizer(vAttendee *x);

  const List<vAttendee*> & Attendees();
  int NumAttendees();
  vAttendee* Attendee(int i);
  void AddAttendee(vAttendee *x);

  BUSYSTATE Busystate();
  void Busystate(BUSYSTATE x);

  void Alarm(vAlarm *x);
  vAlarm *Alarm();

  int ALM();
  const char* ALM_Time();
  const char* ALM_SoundFile();
  int ALM_Options();
  int ALM_Count();

  void RRule(vRRule *x);
  vRRule* RRule();

  int REC_Type();
  void REC_Type(int x);

  int REC_Interval();
  void REC_Interval(int x);

  int REC_Month();
  void REC_Month(int x);

  int REC_Day();
  void REC_Day(int x);

  int REC_DAYMASK();
  void REC_DAYMASK(int x);

  int REC_Instance();
  void REC_Instance(int x);

  int REC_NoEndDate();
  void REC_NoEndDate(int x);

  const char *REC_StartDate();
  const char *REC_EndDate();
  void REC_EndDate(char *x);

  int Compare(vEvent *ev);
};

#define MAXEVENTS 1024

class vCalendar {
  char *buffer;

  vTimezone *vtimezone;
  unsigned int ntz;
  vTimezone **vtimezones;
  unsigned int nevent;
  vEvent *event[MAXEVENTS];

  char VERSION[255];
  char PRODID[255];
  char X_WR_CALNAME[255];
  char X_WR_RELCALID[255];
  char X_WR_TIMEZONE[255];
  char CALSCALE[255];
  char METHOD[255];

  int pstate;
 public:
  vCalendar(char *s=NULL);
  ~vCalendar();

  void Write(char **buffer);

  void Timezone(vTimezone *tz);
  void AddTimezone(vTimezone *tz);
  vTimezone **Timezones();

  unsigned int NumEvents();
  vEvent *Event(int i=0);
  void AddEvent(vEvent *ev);

  void ProductID(char *s);
  void Version(char *s);
};

#endif

